/* 
 * Copyright (c) 2010
 * Texas Instruments
 *
 *  All rights reserved.  Property of Texas Instruments
 *  Restricted rights to use, duplicate or disclose this code are
 *  granted through contract.
 * 
 * */
/*
 *  ======== DriverTable.xdc ========
 *
 *
 */

import xdc.runtime.Error;
import xdc.runtime.Assert;
import xdc.runtime.IGateProvider;
import ti.sdo.utils.NameServer;

/*! 
 *  DriverTable module
 *
 *  This module maintains a table of {@link IDriver#Handle}. Other modules can
 *  lookup these handles by name and open channels for IO. The {@link Stream}
 *  module will open channels using this table. All names in this table have to
 *  be unique.
 *
 *  This module allows addition of IDriver handles into the table at 
 *  configuration time as well as at runtime. However, the total number of 
 *  runtime entries have to be decided at configuration time. There is no limit
 *  to entries added statically.
 *
 *  This module uses {@link ti.sdo.utils.NameServer} to maintain its table
 */
 
module DriverTable {
    /*! 
     *  Max runtime entries that can be added.
     *
     *  Currently this module requires total number of entries that need
     *  to be added at runtime to be identified at configuration time.
     */
    config UInt maxRuntimeEntries = 0;
    
    /*! 
     *  Gate used to make the table thread safe.
     */
    config IGateProvider.Handle gate = null;
    
    /*! 
     *  Length, in MAUs, of the name field in the table.
     */
    config UInt maxNameLen = 16;

    /*! 
     *  Section name is used to place the IDriver table.
     */
    metaonly config String tableSection = null;
        
    /*!
     *  ======== add ========
     *  Add IDriver handle to the table at runtime.
     *
     *  Will raise an error when name already exists in table.
     *  This API is not thread safe. Set {@link #gate} parameter
     *  to protect table if called from multiple threads.
     *
     *  @param(name)            name of entry
     *  @param(driverHandle)    {@link IDriver#Handle}
     *  @param(eb)              Error Block
     */
    Void add(String name, IDriver.Handle driverHandle, Error.Block *eb);

    /*!
     *  ======== addMeta ========
     *  Add to IDriver handle to the table at configuration time.
     *
     *  @param(name)            name of entry
     *  @param(driverHandle)    {@link IDriver#Handle}
     */
    metaonly Void addMeta(String name, IDriver.Handle driverHandle);

    /*!
     *  ======== remove ========
     *  Remove entry from IDriver table at runtime.
     *
     *  Will raise an error when name not found in table
     *  This API is not thread safe. Set {@link #gate} parameter
     *  to protect table if called from multiple threads.
     *
     *  @param(name)    name of entry
     *  @param(eb)      error block
     */
    Void remove(String name, Error.Block *eb);

    /*!
     *  ======== match ========
     *  @_nodoc
     *  match entry in IDriver table.
     *
     *  This function matches the name with table entries and writes to
     *  the handle if it finds a match. If entry is not found it sets
     *  handle to null. It returns length matched.
     *  This API is not thread safe. Set {@link #gate} parameter
     *  to protect table if called from multiple threads.
     *
     *  @param(name)    name of entry
     *  @param(handle)  pointer to IDriver handle.
     *  @param(eb)      error block
     *  @b(returns)     length matched
     */
    Int match(String name, IDriver.Handle *handle, Error.Block *eb);

internal:

    /*!
     * Structure of entry in metaonly table
     */
    struct Entry {
        String  name;
        Ptr     handle;
    };

    /*! 
     *  Array for all statically configured table entries 
     */
    metaonly config Entry staticEntries[];
        
    struct Module_State {
        NameServer.Handle   drvTable;
    };
}
/*
 *  @(#) ti.sdo.io; 1, 0, 0, 0,301; 4-9-2010 18:37:34; /db/vtree/library/trees/ipc/ipc-d23x/src/
 */

