/* 
 * Copyright (c) 2010
 * Texas Instruments
 *
 *  All rights reserved.  Property of Texas Instruments
 *  Restricted rights to use, duplicate or disclose this code are
 *  granted through contract.
 * 
 * */
/*
 *  ======== IConverter.xdc ========
 *
 */

import xdc.runtime.Error;
 
/*!
 *  Interface defining an IConverter
 *
 *  This interfaces allows applications to "stack" functionality on top
 *  of a driver. Modules that implement this interface can manipulate data
 *  coming to and from a driver. Simple scaling, fixed to float or float to
 *  fixed transformations can be done using IConverters without major changes
 *  in the application.
 *
 *  IConverters can only be used along with the {@link Stream} module.
 *  
 *  Stream maintains a name table of {@link IConverter} handles. 
 *  This table is used by Stream to create an IO stack. The name passed to
 *  {@link Stream#create} is usually of the form "/scale/uart". This name may
 *  correspond to the following IO stack.
 *
 *         Stream Instance
 *
 *                 |

 *                 V
 *
 *         IConverter Instance (/scale)
 *
 *                 |
 *
 *                 V
 *
 *         IDriver Instance (/uart)
 *
 *  In this case the Stream requires "/scale" to be in its IConverter table 
 *  and "/uart" to be in {@link DriverTable}. The IConverter table associates
 *  a name with an IConverter Handle. Note that these names have to be of the
 *  form "/name1".
 *
 *  There may be several other IConverters such as a 
 *  {@link ti.sdo.io.converters.Transformer} instance in the stack.
 *
 *  IConverter implementation follows a simple asynchronous issue/reclaim
 *  model. Once an instance of an IConverter is created it accepts IO
 *  packets through the {@link #issue} function. Issue ALWAYS results in a 
 *  callback when IO completes or an error occurs.
 *
 *  The IConverter device above it in the stack or the {@link Stream} 
 *  module will call {@link #reclaim} to get the packet back.
 *
 *  {@link ti.sdo.io.DriverTypes#ControlCmd} are sent to the IConverters or the 
 *  underlying drivers using {@link #control} function.
 *
 *  Only  packets with {@link ti.sdo.io.DriverTypes#READ} and 
 *  {@link ti.sdo.io.DriverTypes#WRITE} are operated on by IConverter. Other 
 *  commands are passed down.
 */

 
interface IConverter
{
    /*!
     *  ======== Q_TERMINATING ========
     *  Terminating quality.
     *
     *  Implementations with this "quality" can be at the bottom of the IO
     *  stack
     */
    const Int Q_TERMINATING = 1;
    
    /*!
     *  Typedef for callback function.
     *
     *  The IConverter instance lower in the stack will invoke this callback 
     *  whenever an I/O operation completes.
     */
    typedef Void (*DoneFxn)(UArg);

instance:
    
    /*! ======== open ========
     *  Opens the IConverter Instance.
     *
     *  This is called at runtime after the IConverter instance has been
     *  created. This function opens the IConverter instance lower in the
     *  stack and gives its callback function and arg.
     *
     *  @param(name)        remaining name
     *  @param(mode)        DriverTypes_INPUT/OUTPUT
     *  @param(chanParams)  channel params for driver at the bottom of stack
     *  @param(cbFxn)       callback function
     *  @param(cbArg)       callback function arg
     *  @param(eb)          error block
     */
    Void open(String name, UInt mode, UArg chanParams, 
        DoneFxn cbFxn, UArg cbArg, Error.Block *eb);

    /*! ======== close ========
     *  Close an IConverter Instance.
     *
     *  @param(eb)       error block
     */
    Void close( Error.Block *eb);
           
    /*! ======== issue ========
     *  Issue a packet for IO.
     *
     *  The IConverter might work on the buffer of data if the mode is
     *  {@link ti.sdo.io.DriverTypes#OUTPUT} and call the issue function for the
     *  IConverter lower in the stack. Some IConverters may be the last in
     *  the IO stack. issue() always results in a callback.
     *
     *  @param(packet)      IO packet
     *  @param(eb)          Error Block
     */
    Void issue(DriverTypes.Packet *packet, Error.Block *eb);

    /*! ======== reclaim ========
     *  Reclaim a previously issued packet.
     *
     *  The IConverter will call the reclaim function for the
     *  IConverter lower in the stack. It may work on the buffer of data 
     *  returned if the mode is {@link ti.sdo.io.DriverTypes#INPUT}.
     *
     *  @param(packetp)   pointer to returned packet
     *  @param(eb)        Error Block
     */
    Void reclaim(DriverTypes.Packet **packetp, Error.Block *eb);

    /*! ======== control ========
     *  Send a control command.
     *
     *  The IConverter will respond to command meant for it and pass down all
     *  others.
     *
     *  @param(cmd)    control cmd
     *  @param(cmdArg) control cmd arg
     *  @param(eb)     error block
     */
    Void control(DriverTypes.ControlCmd cmd, UArg cmdArg, Error.Block *eb);
    
    /*!
     * ======== query ========
     * Query for qualities supported.
     *
     *  @param(qual)    quality to be tested     
     */
    Bool query(Int qual);
}
/*
 *  @(#) ti.sdo.io; 1, 0, 0, 0,301; 4-9-2010 18:37:35; /db/vtree/library/trees/ipc/ipc-d23x/src/
 */

