/* 
 * Copyright (c) 2010
 * Texas Instruments
 *
 *  All rights reserved.  Property of Texas Instruments
 *  Restricted rights to use, duplicate or disclose this code are
 *  granted through contract.
 * 
 * */
/*  ======== Generator.c ========
 */

#include <xdc/std.h>
#include <xdc/runtime/Error.h>
#include <xdc/runtime/Gate.h>
 
#include <ti/sdo/utils/List.h> 
#include <ti/sdo/io/IDriver.h> 
#include <ti/sdo/io/DriverTypes.h> 

#include <package/internal/Generator.xdc.h>

/*
 *************************************************************************
 *                      Module functions
 *************************************************************************
 */

/*
 *  ======== Generator_simulateIsr ========
 */
Void Generator_simulateIsr(UArg arg)
{
    Generator_ChanObj *chan;
    DriverTypes_Packet *packet;
    List_Elem  *elem = NULL;
    List_Handle chanQ;
    
    chanQ = Generator_Module_State_chanList();

    /* loop through all channels and process a packet per channel */
    while ((elem = List_next(chanQ, elem)) != NULL) {
        chan = (Generator_ChanObj *)elem;
        
        if (chan->returnPending) {               
            packet = List_get(chan->pendList);
            //process one packet per isr
            if (packet != NULL) {
                if (chan->userFxn) {
                    chan->userFxn(packet->addr, 
                    packet->origSize, chan->userArg);
                }
                packet->error = NULL;
                packet->size = packet->origSize;
                chan->cbFxn(chan->cbArg, packet);
            }
        }
    }
}


/*
 *  ======== open ========
 */
Ptr Generator_open(Generator_Object *obj, String name, UInt mode, \
    UArg chanParams, DriverTypes_DoneFxn cbFxn, UArg cbArg, Error_Block *eb)
{
    Generator_ChanObj *chan = NULL;
    Generator_ChanParams *prms = (Generator_ChanParams *) chanParams;
    UInt key;
    List_Handle chanQ;
    
    chanQ = Generator_Module_State_chanList();
    
    key = Gate_enterSystem();
    if (mode == DriverTypes_INPUT) {
        if ( !obj->chans[0].inUse) {
            chan = &(obj->chans[DriverTypes_INPUT]);
            chan->inUse = TRUE;  
    }
    }

    if (mode == DriverTypes_OUTPUT) {
        if (!obj->chans[1].inUse) {
            chan = &(obj->chans[DriverTypes_OUTPUT]);
            chan->inUse = TRUE;  
    }
    }
    Gate_leaveSystem(key);

    if (chan == NULL) {
        Error_raise(eb, DriverTypes_EINUSE, 0, 0);
        return (NULL);
    }

    chan->cbFxn = cbFxn;
    chan->cbArg = cbArg;
    chan->userFxn = prms->userFxn;
    chan->userArg = prms->userArg;
    chan->returnPending = prms->returnPending;

    List_put(chanQ, (List_Elem *)chan);
        
    return (chan);
}

/*
 *  ======== submit ========
 */
UInt Generator_submit(Generator_Object *obj, Ptr chanHandle, 
    DriverTypes_Packet *packet, Error_Block *eb)
{
    Generator_ChanObj *chan = (Generator_ChanObj *)chanHandle;

    if (chan->returnPending) {
        List_put(chan->pendList, (List_Elem *)packet);
        return (DriverTypes_PENDING);
    }
    else {
        /* Call the users defined function to send or receive buffer's data  */
        if (chan->userFxn) {
            chan->userFxn(packet->addr, packet->origSize, chan->userArg);
        }
        return (DriverTypes_COMPLETED);
    }
}

/*
 *  ======== control ========
 */
Void Generator_control(Generator_Object *obj, Ptr chanHandle, 
    DriverTypes_ControlCmd cmd, UArg cmdArg, Error_Block *eb)
{
    UInt i = 0;
    Generator_ChanObj *chan = (Generator_ChanObj *)chanHandle;
    DriverTypes_Packet *packet;
        
    if (cmd == DriverTypes_CHAN_ABORT) {
        while(!List_empty(chan->pendList)) {
            i++;
            packet = List_get(chan->pendList);
            packet->error = DriverTypes_EABORTED;
            packet->size = 0;
            chan->cbFxn(chan->cbArg, packet);
        }
        /* return number of packets aborted */
        if (cmdArg) {
            *(UInt *)cmdArg = i;
        }
    }
    else {
        Error_raise(eb, DriverTypes_ENOTIMPL, 0, 0 );
    }
}

/*
 *  ======== close ========
 */
Void Generator_close(Generator_Object *obj, Ptr chanHandle, Error_Block *eb)
{
    Generator_ChanObj *chan;
    UInt key;
    List_Handle chanQ;
    
    chanQ = Generator_Module_State_chanList();
    
    chan = (Generator_ChanObj *)chanHandle;

    if (!List_empty(chan->pendList)) {
        Error_raise (eb, DriverTypes_EINUSE, 0, 0);
        return;
    }
    if (!chan->inUse) {
        Error_raise (eb, DriverTypes_EBADARGS, 0, 0);
        return;
    }
 
    key = Gate_enterSystem();
    List_remove(chanQ, (List_Elem *)chan);
    chan->inUse = FALSE;
    Gate_leaveSystem(key);
}

/*
 *  ======== DriverAdapter_Instance_init ========
 */
Void Generator_Instance_init(Generator_Object *obj, 
    const Generator_Params *params)
{
    Int i;
    Generator_ChanObj *chan;
    
    for (i = 0; i < Generator_NUMCHANS; i++) {
        chan = &obj->chans[i];
               
        List_elemClear(&chan->elem);
        chan->inUse = FALSE;
        chan->pendList = List_create(NULL, NULL);
    }
}
/*
 *  @(#) ti.sdo.io.drivers; 1, 0, 0, 0,299; 4-9-2010 18:37:45; /db/vtree/library/trees/ipc/ipc-d23x/src/
 */

