/* 
 * Copyright (c) 2010
 * Texas Instruments
 *
 *  All rights reserved.  Property of Texas Instruments
 *  Restricted rights to use, duplicate or disclose this code are
 *  granted through contract.
 * 
 * */
/*
 *  ======== Generator.xdc ========
 *
 *
 */

import xdc.rov.ViewInfo;

import xdc.runtime.IHeap;
import xdc.runtime.Error;
import ti.sdo.io.DriverTypes;
import ti.sdo.utils.List;

/*! 
 *  Generator module
 *
 *  Using this module clients can simulate a device with an input channel and
 *  an output channel.
 *  Generator channels are used to generate sequences of constants, sine waves, 
 *  random noise, or other streams of data defined by a user function.
 *
 *  When a channel is opened, the user gets to specify a function to simulate
 *  the input channel and a function to simulate the output channel 
 *  characteristics.
 *
 *  The Generator module can be configured to process the io just like
 *  a real driver, where a {@link #submit} call will return pending and
 *  io will be completed in the context of an isr. This mode is called
 *  returnPending.
 *  However the user has to call {@link #simulateIsrFxn} in an isr, Swi or a 
 *  Task to support this mode.  In simulateIsr, one pending IO Packet for both 
 *  channel for all Generator instances is processed.
 */
module Generator inherits ti.sdo.io.IDriver {  

    /*! typedef for user specified I/O generators
     *
     *  Functions of this type get passed the buffer, buffer size and a 
     *  function specific argument
     */
    typedef Void (*GenFunc)(Ptr, SizeT, UArg);

    /*! Number of channels per generator device. 
     *  
     *  one input and one output channel
     */
    const Int NUMCHANS = 2;

    /*! Channel parameters used along with open  */
    struct ChanParams {
        GenFunc userFxn;        /* user defined function           */
        UArg    userArg;        /* argument for user defined function     */
        Bool    returnPending;  /* TRUE to simulate driver which postpones 
                                   io processing */
    };
    
    metaonly struct BasicView {
        String label;
    };
    
    metaonly struct GeneratorView {
        String                  mode;
        Bool                    inUse;
        Bool                    returnPending;
        String                  callbackFxn[];
        UArg                    callbackArg;
        String                  userFxn[];
        UArg                    userArg;
    }
    
    @Facet
    metaonly config ViewInfo.Instance rovViewInfo = 
        ViewInfo.create({
            viewMap: [
                ['Basic', {type: ViewInfo.INSTANCE,      viewInitFxn: 'viewInitBasic', structName: 'BasicView'}],
                ['Data',  {type: ViewInfo.INSTANCE_DATA, viewInitFxn: 'viewInitData',  structName: 'GeneratorView'}],                   
            ]
        });
    
    /*!
     *  This function is used to give energy to the Generator 
     *  driver to process its IO packets.  It simulates real ISR.
     *
     *  The application needs to call this function within a hwi, swi or
     *  task thread if any channels are opened in {@link #returnPending}
     *  mode.
     *
     *  The Generator module will process all channels with returnPending set 
     *  to true within this function. One packet per channel for all
     *  generator instances gets processed during a single call to this 
     *  function.
     */
    Void simulateIsr(UArg arg);

internal:

    struct ChanObj {
        List.Elem               elem;
        Bool                    inUse;         /* is channel is use? */
        Bool                    returnPending; /* pending mode */ 
        List.Handle             pendList;     /* queue io packets */
        DriverTypes.DoneFxn     cbFxn;        /* callback fxn */
        UArg                    cbArg;        /* callback arg */
        GenFunc                 userFxn;      /* user gen fxn */
        UArg                    userArg;      /* user gen arg */
    };

    struct Instance_State{
        ChanObj chans[NUMCHANS];/* two chans per device */
    };
    
    struct Module_State {
        List.Object chanList;
    }
}
/*
 *  @(#) ti.sdo.io.drivers; 1, 0, 0, 0,299; 4-9-2010 18:37:46; /db/vtree/library/trees/ipc/ipc-d23x/src/
 */

