/* 
 * Copyright (c) 2010
 * Texas Instruments
 *
 *  All rights reserved.  Property of Texas Instruments
 *  Restricted rights to use, duplicate or disclose this code are
 *  granted through contract.
 *
 * */
/*
 *  ======== GateMP.c ========
 */

#include <xdc/std.h>
#include <string.h>
#include <xdc/runtime/Error.h>
#include <xdc/runtime/Assert.h>
#include <xdc/runtime/System.h>
#include <xdc/runtime/IGateProvider.h>
#include <xdc/runtime/IHeap.h>
#include <xdc/runtime/Memory.h>
#include <xdc/runtime/Gate.h>
#include <xdc/runtime/Log.h>

#include <ti/sdo/ipc/interfaces/IGateMPSupport.h>

#include <ti/sysbios/gates/GateMutexPri.h>
#include <ti/sysbios/gates/GateSwi.h>
#include <ti/sysbios/gates/GateHwi.h>
#include <ti/sysbios/hal/Cache.h>

#include "package/internal/GateMP.xdc.h"

#include <ti/sdo/ipc/_Ipc.h>
#include <ti/sdo/utils/_MultiProc.h>
#include <ti/sdo/ipc/_GateMP.h>
#include <ti/sdo/utils/_NameServer.h>
#include <ti/sdo/ipc/_SharedRegion.h>

/* Helper macros */
#define GETREMOTE(mask) ((ti_sdo_ipc_GateMP_RemoteProtect)(mask >> 8))
#define GETLOCAL(mask)  ((ti_sdo_ipc_GateMP_LocalProtect)(mask & 0xFF))
#define SETMASK(remoteProtect, localProtect) \
                        ((Bits32)(remoteProtect << 8 | localProtect))

#ifdef __ti__
    #pragma FUNC_EXT_CALLED(GateMP_Params_init);
    #pragma FUNC_EXT_CALLED(GateMP_create);
    #pragma FUNC_EXT_CALLED(GateMP_close);
    #pragma FUNC_EXT_CALLED(GateMP_delete);
    #pragma FUNC_EXT_CALLED(GateMP_enter);
    #pragma FUNC_EXT_CALLED(GateMP_getDefaultRemote);
    #pragma FUNC_EXT_CALLED(GateMP_getLocalProtect);
    #pragma FUNC_EXT_CALLED(GateMP_getRemoteProtect);
    #pragma FUNC_EXT_CALLED(GateMP_leave);
    #pragma FUNC_EXT_CALLED(GateMP_open);
    #pragma FUNC_EXT_CALLED(GateMP_openByAddr);
    #pragma FUNC_EXT_CALLED(GateMP_sharedMemReq);
#endif

/*
 *  ======== GateMP_getSharedParams ========
 */
static Void GateMP_getSharedParams(GateMP_Params *sparams,
    const ti_sdo_ipc_GateMP_Params *params)
{
    sparams->name             = params->name;
    sparams->regionId         = params->regionId;
    sparams->sharedAddr       = params->sharedAddr;
    sparams->localProtect     = (GateMP_LocalProtect)
        params->localProtect;
    sparams->remoteProtect    = (GateMP_RemoteProtect)
        params->remoteProtect;
}

/*
 *  ======== GateMP_getRTSCParams ========
 */
static Void GateMP_getRTSCParams(ti_sdo_ipc_GateMP_Params *params,
    const GateMP_Params *sparams)
{

    ti_sdo_ipc_GateMP_Params_init(params);

    params->name             = sparams->name;
    params->regionId         = sparams->regionId;
    params->sharedAddr       = sparams->sharedAddr;
    params->localProtect     = (ti_sdo_ipc_GateMP_LocalProtect)
        sparams->localProtect;
    params->remoteProtect    = (ti_sdo_ipc_GateMP_RemoteProtect)
        sparams->remoteProtect;
}

/*
 *  ======== GateMP_Params_init ========
 */
Void GateMP_Params_init(GateMP_Params *sparams)
{
    ti_sdo_ipc_GateMP_Params params;

    ti_sdo_ipc_GateMP_Params_init(&params);
    GateMP_getSharedParams(sparams, &params);
}

/*
 *  ======== GateMP_create ========
 */
GateMP_Handle GateMP_create(const GateMP_Params *sparams)
{
    ti_sdo_ipc_GateMP_Params params;
    GateMP_Object *obj;
    Error_Block eb;

    GateMP_getRTSCParams(&params, (Ptr)sparams);

    Error_init(&eb);

    /* call the module create */
    obj = ti_sdo_ipc_GateMP_create(&params, &eb);

    return ((GateMP_Handle)obj);
}

/*
 *  ======== GateMP_delete ========
 */
Int GateMP_delete(GateMP_Handle *handlePtr)
{
    ti_sdo_ipc_GateMP_delete(handlePtr);

    return (GateMP_S_SUCCESS);
}

/*
 *************************************************************************
 *                       Instance functions
 *************************************************************************
 */

/*
 *  ======== GateMP_enter ========
 */
IArg GateMP_enter(GateMP_Object *obj)
{
    return (ti_sdo_ipc_GateMP_enter__F(obj));
}

/*
 *  ======== GateMP_leave ========
 */
Void GateMP_leave(GateMP_Object *obj, IArg key)
{
    ti_sdo_ipc_GateMP_leave__F(obj, key);
}

/*
 *  ======== ti_sdo_ipc_GateMP_Instance_init__F ========
 */
Int ti_sdo_ipc_GateMP_Instance_init__F(GateMP_Object *obj,
    const ti_sdo_ipc_GateMP_Params *params,
        Error_Block *eb)
{
    IArg key;
    IGateMPSupport_Handle remoteHandle;
    GateMP_RemoteSystemProxy_Params systemParams;
    GateMP_RemoteCustom1Proxy_Params custom1Params;
    GateMP_RemoteCustom2Proxy_Params custom2Params;
    SizeT minAlign, offset;
    SharedRegion_SRPtr sharedShmBase;
    GateMP_Params sparams;
    UInt32 nsValue[2];
    IHeap_Handle regionHeap;
    
    /* Initialize resourceId to an invalid value */
    obj->resourceId = (UInt)-1;

    /* Open GateMP instance */
    if (params->openFlag == TRUE) {
        /* all open work done here except for remote gateHandle */
        obj->localProtect   = params->localProtect;
        obj->remoteProtect  = params->remoteProtect;
        obj->nsKey          = 0;
        obj->numOpens       = 1;
        obj->attrs          = (GateMP_Attrs *)params->sharedAddr;
        obj->regionId       = SharedRegion_getId((Ptr)obj->attrs);
        obj->cacheEnabled   = SharedRegion_isCacheEnabled(obj->regionId);

        /* Assert that the buffer is in a valid shared region */
        Assert_isTrue(obj->regionId != SharedRegion_INVALIDREGIONID,
                      Ipc_A_addrNotInSharedRegion);

        obj->allocSize      = 0;
        obj->objType        = Ipc_ObjType_OPENDYNAMIC;

        minAlign = Memory_getMaxDefaultTypeAlign();

        if (SharedRegion_getCacheLineSize(obj->regionId) > minAlign) {
            minAlign = SharedRegion_getCacheLineSize(obj->regionId);
        }

        offset = _Ipc_roundup(sizeof(GateMP_Attrs), minAlign);

        obj->proxyAttrs = (Ptr)((UInt32)obj->attrs + offset);
    }
    /* Create GateMP instance */
    else {
        obj->localProtect  = params->localProtect;
        obj->remoteProtect = params->remoteProtect;
        obj->nsKey         = 0;
        obj->numOpens      = 0;

        if (obj->remoteProtect == GateMP_RemoteProtect_NONE) {
            /* Creating a local gate (Attrs is in local memory) */
            /* all work done here and return */
            obj->cacheEnabled  = FALSE; /* local */
            obj->objType = Ipc_ObjType_LOCAL;
            obj->gateHandle = GateMP_createLocal(obj->localProtect);

            /* Using default target alignment */
            obj->attrs = Memory_alloc(NULL, sizeof(GateMP_Attrs), 0, eb);
            if (obj->attrs == NULL) {
                return (2);
            }

            obj->attrs->arg = (Bits32)obj;
            obj->attrs->mask = SETMASK(obj->remoteProtect, obj->localProtect);
            obj->attrs->creatorProcId = MultiProc_self();
            obj->attrs->status = GateMP_CREATED;
            if (params->name) {
                nsValue[0] = (UInt32)obj->attrs;
                nsValue[1] = MultiProc_self() << 16;
                obj->nsKey = NameServer_add(GateMP_module->nameServer,
                    params->name, &nsValue, 2 * sizeof(UInt32));
                if (obj->nsKey == NULL) {
                    return (3);
                }
            }
            
            /* Nothing else to do for local gates */
            return (0);
        }

        if (params->sharedAddr == NULL) {
            /* Need to allocate from heap */
            obj->objType        = Ipc_ObjType_CREATEDYNAMIC_REGION;
            obj->regionId       = params->regionId;
            GateMP_getSharedParams(&sparams, params);
            obj->allocSize      = GateMP_sharedMemReq(&sparams);
            obj->cacheEnabled   = SharedRegion_isCacheEnabled(obj->regionId);

            /* The region heap will do the alignment */
            regionHeap = SharedRegion_getHeap(obj->regionId);
            Assert_isTrue(regionHeap != NULL, SharedRegion_A_noHeap);
            obj->attrs = Memory_alloc(regionHeap, obj->allocSize, 0, eb);
            if (obj->attrs == NULL) {
                return (3);
            }

            minAlign = Memory_getMaxDefaultTypeAlign();

            if (SharedRegion_getCacheLineSize(obj->regionId) > minAlign) {
                minAlign = SharedRegion_getCacheLineSize(obj->regionId);
            }

            offset = _Ipc_roundup(sizeof(GateMP_Attrs), minAlign);

            obj->proxyAttrs = (Ptr)((UInt32)obj->attrs + offset);
        }
        else {
            /* creating using sharedAddr */
            obj->regionId = SharedRegion_getId(params->sharedAddr);
            /* Assert that the buffer is in a valid shared region */
            Assert_isTrue(obj->regionId != SharedRegion_INVALIDREGIONID,
                          Ipc_A_addrNotInSharedRegion);

            /* Assert that sharedAddr is cache aligned */
            Assert_isTrue(((UInt32)params->sharedAddr %
                SharedRegion_getCacheLineSize(obj->regionId) == 0),
                Ipc_A_addrNotCacheAligned);

            obj->attrs = (GateMP_Attrs *)params->sharedAddr;
            obj->cacheEnabled  = SharedRegion_isCacheEnabled(obj->regionId);
            obj->objType = Ipc_ObjType_CREATEDYNAMIC;

            minAlign = Memory_getMaxDefaultTypeAlign();

            if (SharedRegion_getCacheLineSize(obj->regionId) > minAlign) {
                minAlign = SharedRegion_getCacheLineSize(obj->regionId);
            }

            offset = _Ipc_roundup(sizeof(GateMP_Attrs), minAlign);

            obj->proxyAttrs = (Ptr)((UInt32)obj->attrs + offset);
        }
    }

    /* Proxy work for open and create done here */
    switch (obj->remoteProtect) {
        case GateMP_RemoteProtect_SYSTEM:
            if (obj->objType != Ipc_ObjType_OPENDYNAMIC) {
                /* Created Instance */
                obj->resourceId = GateMP_getFreeResource(
                    GateMP_module->remoteSystemInUse,
                             GateMP_module->numRemoteSystem, eb);
                if (Error_check(eb) == TRUE) {
                    return (4);
                }
            }
            else {
                /* resourceId set by open call */
                obj->resourceId = params->resourceId;
            }

            /* Create the proxy object */
            GateMP_RemoteSystemProxy_Params_init(&systemParams);
            systemParams.resourceId = obj->resourceId;
            systemParams.openFlag = (obj->objType == Ipc_ObjType_OPENDYNAMIC);
            systemParams.sharedAddr = obj->proxyAttrs;
            remoteHandle = GateMP_RemoteSystemProxy_create(
                        (IGateMPSupport_LocalProtect)obj->localProtect,
                        &systemParams, eb);

            if (remoteHandle == NULL) {
                return (5);
            }

            /* Finish filling in the object */
            obj->gateHandle = IGateMPSupport_Handle_upCast(remoteHandle);

            /* Fill in the local array because it is cooked */
            key = Gate_enterSystem();
            GateMP_module->remoteSystemGates[obj->resourceId] = obj;
            Gate_leaveSystem(key);
            break;

        case GateMP_RemoteProtect_CUSTOM1:
            if (obj->objType != Ipc_ObjType_OPENDYNAMIC) {
                obj->resourceId = GateMP_getFreeResource(
                             GateMP_module->remoteCustom1InUse,
                             GateMP_module->numRemoteCustom1, eb);
                if (Error_check(eb) == TRUE) {
                    return (4);
                }
            }
            else {
                /* resourceId set by open call */
                obj->resourceId = params->resourceId;
            }

            /* Create the proxy object */
            GateMP_RemoteCustom1Proxy_Params_init(&custom1Params);
            custom1Params.resourceId = obj->resourceId;
            custom1Params.openFlag = (obj->objType == Ipc_ObjType_OPENDYNAMIC);
            custom1Params.sharedAddr = obj->proxyAttrs;
            remoteHandle = GateMP_RemoteCustom1Proxy_create(
                (IGateMPSupport_LocalProtect)obj->localProtect,
                &custom1Params, eb);
            if (remoteHandle == NULL) {
                return (5);
            }

            /* Fill in the local array because it is cooked */
            key = Gate_enterSystem();
            GateMP_module->remoteCustom1Gates[obj->resourceId] = obj;
            Gate_leaveSystem(key);
            break;

        case GateMP_RemoteProtect_CUSTOM2:
            if (obj->objType != Ipc_ObjType_OPENDYNAMIC) {
                obj->resourceId = GateMP_getFreeResource(
                             GateMP_module->remoteCustom2InUse,
                             GateMP_module->numRemoteCustom2, eb);
                if (Error_check(eb) == TRUE) {
                    return (4);
                }
            }
            else {
                /* resourceId set by open call */
                obj->resourceId = params->resourceId;
            }

            /* Create the proxy object */
            GateMP_RemoteCustom2Proxy_Params_init(&custom2Params);
            custom2Params.resourceId = obj->resourceId;
            custom2Params.openFlag = (obj->objType == Ipc_ObjType_OPENDYNAMIC);
            custom2Params.sharedAddr = obj->proxyAttrs;
            remoteHandle = GateMP_RemoteCustom2Proxy_create(
                        (IGateMPSupport_LocalProtect)obj->localProtect,
                        &custom2Params, eb);
            if (remoteHandle == NULL) {
                return (5);
            }

            /* Fill in the local array because it is cooked */
            key = Gate_enterSystem();
            GateMP_module->remoteCustom2Gates[obj->resourceId] = obj;
            Gate_leaveSystem(key);
            break;

        default:
            /* Should never be here */
            Assert_isTrue(FALSE, Ipc_A_internal);
            break;
    }
    
    obj->gateHandle = IGateMPSupport_Handle_upCast(remoteHandle);

    /* Place Name/Attrs into NameServer table */
    if (obj->objType != Ipc_ObjType_OPENDYNAMIC) {
        /* Fill in the attrs */
        obj->attrs->arg = obj->resourceId;
        obj->attrs->mask = SETMASK(obj->remoteProtect, obj->localProtect);
        obj->attrs->creatorProcId = MultiProc_self();
        obj->attrs->status = GateMP_CREATED;

        if (obj->cacheEnabled) {
            Cache_wbInv(obj->attrs, sizeof(GateMP_Attrs), Cache_Type_ALL, TRUE);
        }

        if (params->name != NULL) {
            sharedShmBase = SharedRegion_getSRPtr(obj->attrs, obj->regionId);
            nsValue[0] = (UInt32)sharedShmBase;
            nsValue[1] = MultiProc_self() << 16 | 1;
            obj->nsKey = NameServer_add(GateMP_module->nameServer,
                                params->name, &nsValue, 2 * sizeof(UInt32));
            if (obj->nsKey == NULL) {
                return (6);
            }
        }
    }

    if (obj->objType != Ipc_ObjType_OPENDYNAMIC) {
        Log_write2(GateMP_LM_create,
                (UArg)obj->remoteProtect, (UArg)obj->resourceId);
    }
    else {
        Log_write2(GateMP_LM_open,
            (UArg)obj->remoteProtect, (UArg)obj->resourceId);
    }

    return (0);
}

/*
 *  ======== GateMP_Instance_finalize ========
 */
Void GateMP_Instance_finalize(GateMP_Object *obj, Int status)
{
    IArg systemKey;
    GateMP_RemoteSystemProxy_Handle systemHandle;
    GateMP_RemoteCustom1Proxy_Handle custom1Handle;
    GateMP_RemoteCustom2Proxy_Handle custom2Handle;
    
    GateMP_Handle *remoteHandles;
    UInt8 *inUseArray;
    UInt numResources;

    /* Cannot call when numOpen is non-zero. */
    Assert_isTrue(obj->numOpens == 0, GateMP_A_invalidDelete);

    /* Removed from NameServer */
    if (obj->nsKey != 0) {
        NameServer_removeEntry(GateMP_module->nameServer, obj->nsKey);
    }

    /* Set the status to 0 */
    if (obj->objType != Ipc_ObjType_OPENDYNAMIC) {
        obj->attrs->status = 0;
        if (obj->cacheEnabled) {
            Cache_wbInv(obj->attrs, sizeof(GateMP_Attrs), Cache_Type_ALL, TRUE);
        }
    }

    if (obj->remoteProtect == GateMP_RemoteProtect_NONE) {
        if (obj->attrs != NULL) {
            Memory_free(NULL, obj->attrs, sizeof(GateMP_Attrs));
        }
        
        return;
    }
    
    switch (obj->remoteProtect) {
        case GateMP_RemoteProtect_SYSTEM:
            if (obj->gateHandle) {
                systemHandle =
                GateMP_RemoteSystemProxy_Handle_downCast2(obj->gateHandle);
                GateMP_RemoteSystemProxy_delete(&systemHandle);
            }
            
            inUseArray = GateMP_module->remoteSystemInUse;
            remoteHandles = GateMP_module->remoteSystemGates;
            numResources = GateMP_module->numRemoteSystem;
            break;

        case GateMP_RemoteProtect_CUSTOM1:
            if (obj->gateHandle) {
                custom1Handle =
                   GateMP_RemoteCustom1Proxy_Handle_downCast2(obj->gateHandle);
                GateMP_RemoteCustom1Proxy_delete(&custom1Handle);
            }
            
            inUseArray = GateMP_module->remoteCustom1InUse;
            remoteHandles = GateMP_module->remoteCustom1Gates;
            numResources = GateMP_module->numRemoteCustom1;
            break;

        case GateMP_RemoteProtect_CUSTOM2:
            if (obj->gateHandle) {
                custom2Handle =
                GateMP_RemoteCustom2Proxy_Handle_downCast2(obj->gateHandle);
                GateMP_RemoteCustom2Proxy_delete(&custom2Handle);
            }
            
            inUseArray = GateMP_module->remoteCustom2InUse;
            remoteHandles = GateMP_module->remoteCustom2Gates;
            numResources = GateMP_module->numRemoteCustom2;
            break;

        default:
            /* Should never be here */
            Assert_isTrue(FALSE, Ipc_A_internal);
            break;
    }
    /* Clear the handle array entry in local memory */
    if (obj->resourceId != (UInt)-1) {
        remoteHandles[obj->resourceId] = NULL;
    }
    
    if (obj->objType != Ipc_ObjType_OPENDYNAMIC 
            && obj->resourceId != (UInt)-1) {
        /* Clear the resource used flag in shared memory */
        systemKey = GateMP_enter(GateMP_module->defaultGate);
        inUseArray[obj->resourceId] = FALSE;
        if (obj->cacheEnabled) {
            Cache_wbInv(inUseArray, numResources, Cache_Type_ALL, TRUE);
        }
        GateMP_leave(GateMP_module->defaultGate, systemKey);
    }
    
    if (obj->objType == Ipc_ObjType_CREATEDYNAMIC_REGION) {
        /* Free memory allocated from the region heap */
        if (obj->attrs) {
            Memory_free(SharedRegion_getHeap(obj->regionId), obj->attrs,
                obj->allocSize);
        }
    }
    
    Log_write2(GateMP_LM_delete, (UArg)obj->remoteProtect,
            (UArg)obj->resourceId);
}

/*
 *  ======== GateMP_enter ========
 */
IArg ti_sdo_ipc_GateMP_enter__F(GateMP_Object *obj)
{
    IArg key;

    Assert_isTrue(obj != NULL, Ipc_A_nullArgument);

    key = IGateProvider_enter(obj->gateHandle);

    Log_write3(GateMP_LM_enter,(UArg)obj->remoteProtect,
            (UArg)obj->resourceId, key);

    return (key);
}

/*
 *  ======== GateMP_leave ========
 */
Void ti_sdo_ipc_GateMP_leave__F(GateMP_Object *obj, IArg key)
{
    Assert_isTrue(obj != NULL, Ipc_A_nullArgument);

    IGateProvider_leave(obj->gateHandle, key);

    Log_write3(GateMP_LM_leave, (UArg)obj->remoteProtect,
        (UArg)obj->resourceId, key);
}

/*
 *************************************************************************
 *                       Module functions
 *************************************************************************
 */

/*
 *  ======== GateMP_open ========
 */
Int GateMP_open(String name, GateMP_Handle *handlePtr)
{
    SharedRegion_SRPtr sharedShmBase;
    Int status;
    UInt32 len;
    Ptr sharedAddr;
    UInt32 nsValue[2];

    /* Assert that a pointer has been supplied */
    Assert_isTrue(handlePtr != NULL, Ipc_A_nullArgument);

    len = sizeof(nsValue);

    /*
     *  Get the Attrs out of the NameServer instance.
     *  Search all processors.
     */
    status = NameServer_get(GateMP_module->nameServer, name,
                        &nsValue, &len,
                        NULL);

    if (status < 0) {
        *handlePtr = NULL;
        return (GateMP_E_NOTFOUND);
    }

    /* If trying to open a local GateMP remotely return fail */
    if (!(nsValue[1] & 0x1) && ((nsValue[1] >> 16) != MultiProc_self())) {
        *handlePtr = NULL;
        return (GateMP_E_FAIL);
    }

    sharedShmBase = (SharedRegion_SRPtr)nsValue[0];

    sharedAddr = SharedRegion_getPtr(sharedShmBase);

    status = GateMP_openByAddr(sharedAddr, handlePtr);

    return (status);
}

/*
 *  ======== GateMP_openByAddr ========
 */
Int GateMP_openByAddr(Ptr sharedAddr, GateMP_Handle *handlePtr)
{
    Int status = GateMP_S_SUCCESS;
    IArg key;
    GateMP_Object *obj;
    ti_sdo_ipc_GateMP_Params params;
    GateMP_Attrs *attrs;
    UInt16 regionId;
    Error_Block eb;

    Error_init(&eb);

    attrs = (GateMP_Attrs *)sharedAddr;

    /* get the region id and invalidate attrs is needed */
    regionId = SharedRegion_getId(sharedAddr);
    if (SharedRegion_isCacheEnabled(regionId)) {
        Cache_inv(attrs, sizeof(GateMP_Attrs), Cache_Type_ALL, TRUE);
    }

    if (attrs->status != GateMP_CREATED) {
        *handlePtr = NULL;
        status = GateMP_E_NOTFOUND;
    }
    else {
        /* Local gate */
        if (GETREMOTE(attrs->mask) == GateMP_RemoteProtect_NONE) {
            if (attrs->creatorProcId != MultiProc_self()) {
                Error_raise(&eb, GateMP_E_localGate, 0, 0);
            }

            key = Gate_enterSystem();
            obj = (GateMP_Handle)attrs->arg;
            *handlePtr = obj;
            obj->numOpens++;
            Gate_leaveSystem(key);

            return (GateMP_S_SUCCESS);
        }

        /* Remote case */
        switch (GETREMOTE(attrs->mask)) {
            case GateMP_RemoteProtect_SYSTEM:
                obj = GateMP_module->remoteSystemGates[attrs->arg];
                break;

            case GateMP_RemoteProtect_CUSTOM1:
                obj = GateMP_module->remoteCustom1Gates[attrs->arg];
                break;

            case GateMP_RemoteProtect_CUSTOM2:
                obj = GateMP_module->remoteCustom2Gates[attrs->arg];
                break;

            default:
                break;
        }

        /*
         *  If the object is NULL, then it must have been created on a remote
         *  processor. Need to create a local object. This is accomplished
         *  by setting the openFlag to TRUE.
         */
        if (obj == NULL) {
            /* Create a GateMP object with the openFlag set to true */
            ti_sdo_ipc_GateMP_Params_init(&params);
            params.openFlag = TRUE;
            params.sharedAddr = sharedAddr;
            params.resourceId = attrs->arg;
            params.localProtect = GETLOCAL(attrs->mask);
            params.remoteProtect = GETREMOTE(attrs->mask);

            obj = ti_sdo_ipc_GateMP_create(&params, &eb);
            if (obj == NULL) {
                status = GateMP_E_FAIL;
            }
        }
        else {
            obj->numOpens++;
        }

        /* Return the "opened" GateMP instance  */
        *handlePtr = obj;
    }

    return (status);
}

/*
 *  ======== GateMP_close ========
 */
Int GateMP_close(GateMP_Handle *handlePtr)
{
    IArg key;
    GateMP_Handle gateHandle = *handlePtr;
    Int count;

    /*
     *  Cannot call with the numOpens equal to zero.  This is either
     *  a created handle or been closed already.
     */
    Assert_isTrue((gateHandle->numOpens != 0), GateMP_A_invalidClose);

    key = Gate_enterSystem();
    count = --gateHandle->numOpens;
    Gate_leaveSystem(key);

    /* If the gate is remote, call the close function */
    if (gateHandle->remoteProtect != GateMP_RemoteProtect_NONE) {
        /*
         *  if the count is zero and the gate is opened, then this
         *  object was created in the open (i.e. the create happened
         *  on a remote processor.
         */
        if ((count == 0) && (gateHandle->objType == Ipc_ObjType_OPENDYNAMIC)){
            GateMP_delete(&gateHandle);
        }
    }
    *handlePtr = NULL;

    Log_write2(GateMP_LM_close,
            (UArg)gateHandle->remoteProtect, (UArg)gateHandle->resourceId);

    return (GateMP_S_SUCCESS);
}

/*
 *  ======== GateMP_getSharedAddr ========
 */
SharedRegion_SRPtr GateMP_getSharedAddr(GateMP_Object *obj)
{
    SharedRegion_SRPtr srPtr;

    srPtr = SharedRegion_getSRPtr(obj->attrs, obj->regionId);

    return (srPtr);
}

/*
 *  ======== GateMP_query ========
 */
Bool GateMP_query(Int qual)
{
    return (FALSE);
}

/*
 *  ======== GateMP_getDefaultRemote ========
 */
GateMP_Handle GateMP_getDefaultRemote()
{
    return (GateMP_module->defaultGate);
}

/*
 *  ======== GateMP_getLocalProtect ========
 */
GateMP_LocalProtect GateMP_getLocalProtect(GateMP_Object *obj)
{
    Assert_isTrue(obj != NULL, Ipc_A_nullArgument);
    
    return ((GateMP_LocalProtect)obj->localProtect);
}

/*
 *  ======== GateMP_getRemoteProtect ========
 */
GateMP_RemoteProtect GateMP_getRemoteProtect(GateMP_Object *obj)
{
    Assert_isTrue(obj != NULL, Ipc_A_nullArgument);
    
    return ((GateMP_RemoteProtect)obj->remoteProtect);
}

/*
 *  ======== GateMP_createLocal ========
 */
IGateProvider_Handle GateMP_createLocal(GateMP_LocalProtect localProtect)
{
    IGateProvider_Handle gateHandle;

    /* Create the local gate. */
    switch (localProtect) {
        case GateMP_LocalProtect_NONE:
            /* Plug with the GateNull singleton */
            gateHandle = GateMP_module->gateNull;
            break;

        case GateMP_LocalProtect_INTERRUPT:
            /* Plug with the GateHwi singleton */
            gateHandle = GateMP_module->gateHwi;
            break;

        case GateMP_LocalProtect_TASKLET:
            /* Plug with the GateSwi singleton */
            gateHandle = GateMP_module->gateSwi;
            break;

        case GateMP_LocalProtect_THREAD:
        case GateMP_LocalProtect_PROCESS:
            /* Plug with the GateMutexPri singleton */
            gateHandle = GateMP_module->gateMutexPri;
            break;

        default:
            /* Invalid local protection level encountered */
            Assert_isTrue(FALSE, Ipc_A_internal);
            break;
    }

    return (gateHandle);
}

/*
 *  ======== GateMP_sharedMemReq ========
 */
SizeT GateMP_sharedMemReq(const GateMP_Params *params)
{
    SizeT memReq, minAlign;
    UInt16 regionId;
    GateMP_RemoteSystemProxy_Params systemParams;
    GateMP_RemoteCustom1Proxy_Params custom1Params;
    GateMP_RemoteCustom2Proxy_Params custom2Params;

    if (params->sharedAddr) {
        regionId = SharedRegion_getId(params->sharedAddr);
    }
    else {
        regionId = params->regionId;
    }
    
    Assert_isTrue(regionId != SharedRegion_INVALIDREGIONID, Ipc_A_internal);

    minAlign = Memory_getMaxDefaultTypeAlign();
    if (SharedRegion_getCacheLineSize(regionId) > minAlign) {
        minAlign = SharedRegion_getCacheLineSize(regionId);
    }

    memReq = _Ipc_roundup(sizeof(GateMP_Attrs), minAlign);

    /* add the amount of shared memory required by proxy */
    if (params->remoteProtect == GateMP_RemoteProtect_SYSTEM) {
        GateMP_RemoteSystemProxy_Params_init(&systemParams);
        systemParams.regionId = regionId;
        memReq += GateMP_RemoteSystemProxy_sharedMemReq(
                (IGateMPSupport_Params *)&systemParams);
    }
    else if (params->remoteProtect == GateMP_RemoteProtect_CUSTOM1) {
        GateMP_RemoteCustom1Proxy_Params_init(&custom1Params);
        custom1Params.regionId = regionId;
        memReq += GateMP_RemoteCustom1Proxy_sharedMemReq(
                (IGateMPSupport_Params *)&custom1Params);
    }
    else {
        GateMP_RemoteCustom2Proxy_Params_init(&custom2Params);
        custom2Params.regionId = regionId;
        memReq += GateMP_RemoteCustom2Proxy_sharedMemReq(
                (IGateMPSupport_Params *)&custom2Params);
    }

    return (memReq);
}

/*
 *  ======== GateMP_getRegion0ReservedSize ========
 */
SizeT GateMP_getRegion0ReservedSize(Void)
{
    SizeT reserved, minAlign;

    minAlign = Memory_getMaxDefaultTypeAlign();

    if (SharedRegion_getCacheLineSize(0) > minAlign) {
        minAlign = SharedRegion_getCacheLineSize(0);
    }

    reserved = _Ipc_roundup(sizeof(GateMP_Reserved), minAlign);

    reserved += _Ipc_roundup(GateMP_module->numRemoteSystem, minAlign);

    if (GateMP_module->proxyMap[GateMP_ProxyOrder_CUSTOM1] ==
        GateMP_ProxyOrder_CUSTOM1) {
            reserved += _Ipc_roundup(GateMP_module->numRemoteCustom1, minAlign);
    }

    if (GateMP_module->proxyMap[GateMP_ProxyOrder_CUSTOM2] ==
        GateMP_ProxyOrder_CUSTOM2) {
            reserved += _Ipc_roundup(GateMP_module->numRemoteCustom2, minAlign);
    }

    return (reserved);
}

/*
 *  ======== GateMP_setRegion0ReservedSize ========
 */
Void GateMP_setRegion0Reserved(Ptr sharedAddr)
{
    GateMP_Reserved *reserve;
    UInt32 minAlign, offset;

    minAlign = Memory_getMaxDefaultTypeAlign();
    if (SharedRegion_getCacheLineSize(0) > minAlign) {
        minAlign = SharedRegion_getCacheLineSize(0);
    }

    /* setup GateMP_Reserved fields */
    reserve = (GateMP_Reserved *)sharedAddr;
    reserve->version = GateMP_VERSION;

    if (SharedRegion_isCacheEnabled(0)) {
        Cache_wbInv(sharedAddr, sizeof(GateMP_Reserved), Cache_Type_ALL, TRUE);
    }

    /*
     *  initialize the in-use array in shared memory for the system gates.
     */
    offset = _Ipc_roundup(sizeof(GateMP_Reserved), minAlign);
    GateMP_module->remoteSystemInUse =
        (Ptr)((UInt32)sharedAddr + offset);

    memset(GateMP_module->remoteSystemInUse, 0, GateMP_module->numRemoteSystem);

    if (SharedRegion_isCacheEnabled(0)) {
        Cache_wbInv(GateMP_module->remoteSystemInUse,
            GateMP_module->numRemoteSystem, Cache_Type_ALL, TRUE);
    }

    /*
     *  initialize the in-use array in shared memory for the custom1 gates.
     *  Need to check if this proxy is the same as system
     */
    offset = _Ipc_roundup(GateMP_module->numRemoteSystem, minAlign);
    if (GateMP_module->proxyMap[GateMP_ProxyOrder_CUSTOM1] ==
        GateMP_ProxyOrder_CUSTOM1) {
        if (GateMP_module->numRemoteCustom1 != 0) {
            GateMP_module->remoteCustom1InUse =
                GateMP_module->remoteSystemInUse + offset;
        }

        memset(GateMP_module->remoteCustom1InUse, 0,
            GateMP_module->numRemoteCustom1);

        if (SharedRegion_isCacheEnabled(0)) {
            Cache_wbInv(GateMP_module->remoteCustom1InUse,
                 GateMP_module->numRemoteCustom1, Cache_Type_ALL, TRUE);
        }
    }
    else {
        GateMP_module->remoteCustom1InUse = GateMP_module->remoteSystemInUse;
        GateMP_module->remoteCustom1Gates = GateMP_module->remoteSystemGates;
    }

    /*
     *  initialize the in-use array in shared memory for the custom2 gates.
     *  Need to check if this proxy is the same as system or custom1
     */
    offset = _Ipc_roundup(GateMP_module->numRemoteCustom1, minAlign);
    if (GateMP_module->proxyMap[GateMP_ProxyOrder_CUSTOM2] ==
        GateMP_ProxyOrder_CUSTOM2) {
        if (GateMP_module->numRemoteCustom2 != 0) {
                GateMP_module->remoteCustom2InUse =
                    GateMP_module->remoteCustom1InUse + offset;
        }
        memset(GateMP_module->remoteCustom2InUse, 0,
            GateMP_module->numRemoteCustom2);

        if (SharedRegion_isCacheEnabled(0)) {
            Cache_wbInv(GateMP_module->remoteCustom2InUse,
                 GateMP_module->numRemoteCustom2, Cache_Type_ALL, TRUE);
        }
    }
    else if (GateMP_module->proxyMap[GateMP_ProxyOrder_CUSTOM2] ==
             GateMP_ProxyOrder_CUSTOM1) {
        GateMP_module->remoteCustom2InUse =
            GateMP_module->remoteCustom1InUse;
        GateMP_module->remoteCustom2Gates =
            GateMP_module->remoteCustom1Gates;
    }
    else {
        GateMP_module->remoteCustom2InUse = GateMP_module->remoteSystemInUse;
        GateMP_module->remoteCustom2Gates = GateMP_module->remoteSystemGates;
    }
}

/*
 *  ======== GateMP_openRegion0Reserved ========
 */
Void GateMP_openRegion0Reserved(Ptr sharedAddr)
{
    GateMP_Reserved *reserve;
    SizeT minAlign, offset;

    minAlign = Memory_getMaxDefaultTypeAlign();
    if (SharedRegion_getCacheLineSize(0) > minAlign) {
        minAlign = SharedRegion_getCacheLineSize(0);
    }


    /* setup GateMP_Reserved fields */
    reserve = (GateMP_Reserved *)sharedAddr;

    if (reserve->version != GateMP_VERSION) {
        /* Version mismatch: return an error */
        Error_raise(NULL, Ipc_E_versionMismatch, reserve->version,
                    GateMP_VERSION);
        return;
    }

    offset = _Ipc_roundup(sizeof(GateMP_Reserved), minAlign);
    GateMP_module->remoteSystemInUse =
        (Ptr)((UInt32)sharedAddr + offset);

    /*
     *  initialize the in-use array in shared memory for the custom1 gates.
     *  Need to check if this proxy is the same as system
     */
    offset = _Ipc_roundup(GateMP_module->numRemoteSystem, minAlign);
    if (GateMP_module->proxyMap[GateMP_ProxyOrder_CUSTOM1] ==
        GateMP_ProxyOrder_CUSTOM1) {
        if (GateMP_module->numRemoteCustom1 != 0) {
            GateMP_module->remoteCustom1InUse =
                GateMP_module->remoteSystemInUse + offset;
        }
    }
    else {
        GateMP_module->remoteCustom1InUse = GateMP_module->remoteSystemInUse;
        GateMP_module->remoteCustom1Gates = GateMP_module->remoteSystemGates;
    }

    offset = _Ipc_roundup(GateMP_module->numRemoteCustom1, minAlign);
    if (GateMP_module->proxyMap[GateMP_ProxyOrder_CUSTOM2] ==
        GateMP_ProxyOrder_CUSTOM2) {
        if (GateMP_module->numRemoteCustom2 != 0) {
            GateMP_module->remoteCustom2InUse =
                GateMP_module->remoteCustom1InUse + offset;
        }
    }
    else if (GateMP_module->proxyMap[GateMP_ProxyOrder_CUSTOM2] ==
             GateMP_ProxyOrder_CUSTOM1) {
        GateMP_module->remoteCustom2InUse =
            GateMP_module->remoteCustom1InUse;
        GateMP_module->remoteCustom2Gates =
            GateMP_module->remoteCustom1Gates;
    }
    else {
        GateMP_module->remoteCustom2InUse = GateMP_module->remoteSystemInUse;
        GateMP_module->remoteCustom2Gates = GateMP_module->remoteSystemGates;
    }
}

/*
 *  ======== GateMP_setDefaultRemote() ========
 */
Void GateMP_setDefaultRemote(GateMP_Handle handle)
{
    GateMP_module->defaultGate = handle;
}

/*
 *  ======== GateMP_attach ========
 */
Int GateMP_attach(UInt16 remoteProcId, Ptr sharedAddr)
{
    Ptr gateMPsharedAddr;
    SharedRegion_Entry entry;
    GateMP_Handle defaultGate;
    Int status = GateMP_S_SUCCESS;

    /* If default gate is not NULL return since its already set */
    if (GateMP_getDefaultRemote() != NULL) {
        return(GateMP_S_ALREADYSETUP);
    }

    /* get region 0 information */
    SharedRegion_getEntry(0, &entry);

    gateMPsharedAddr = (Ptr)((UInt32)sharedAddr +
                       GateMP_getRegion0ReservedSize());

    if ((entry.ownerProcId != MultiProc_self()) &&
        (entry.ownerProcId != SharedRegion_DEFAULTOWNERID)) {
        /* if not the owner of the SharedRegion */
        GateMP_openRegion0Reserved(sharedAddr);

        /* open the gate by address */
        status = GateMP_openByAddr(gateMPsharedAddr, &defaultGate);

        if (status > -1) {
            /* set the default GateMP for opener */
            GateMP_setDefaultRemote(defaultGate);
        }
    }

    return (status);
}

/*
 *  ======== GateMP_start ========
 */
Int GateMP_start(Ptr sharedAddr)
{
    SharedRegion_Entry       entry;
    ti_sdo_ipc_GateMP_Params gateMPParams;
    GateMP_Handle            defaultGate;
    Int status = GateMP_S_SUCCESS;

    /* get region 0 information */
    SharedRegion_getEntry(0, &entry);

    /* if entry owner proc is not specified return */
    if (entry.ownerProcId == SharedRegion_DEFAULTOWNERID) {
        return (status);
    }

    /* Init params for default gate */
    ti_sdo_ipc_GateMP_Params_init(&gateMPParams);
    gateMPParams.sharedAddr = (Ptr)((UInt32)sharedAddr +
                                    GateMP_getRegion0ReservedSize());
    gateMPParams.localProtect  = ti_sdo_ipc_GateMP_LocalProtect_TASKLET;

    if (MultiProc_numProcessors > 1) {
        gateMPParams.remoteProtect = ti_sdo_ipc_GateMP_RemoteProtect_SYSTEM;
    }
    else {
        gateMPParams.remoteProtect = ti_sdo_ipc_GateMP_RemoteProtect_NONE;
    }

    if (entry.ownerProcId == MultiProc_self()) {
        /* if owner of the SharedRegion */
        GateMP_setRegion0Reserved(sharedAddr);

        /* create default GateMP */
        defaultGate = ti_sdo_ipc_GateMP_create(&gateMPParams, NULL);

        if (defaultGate != NULL) {
            /* set the default GateMP for creator */
            GateMP_setDefaultRemote(defaultGate);
        }
        else {
            status = GateMP_E_FAIL;
        }
    }

    return (status);
}

/*
 *************************************************************************
 *                       Internal functions
 *************************************************************************
 */
UInt GateMP_getFreeResource(UInt8 *inUse, Int num, Error_Block *eb)
{
    IArg key;
    Bool flag = FALSE;
    UInt resourceId;
    GateMP_Handle defaultGate;

    /* Need to look at shared memory. Enter default gate */
    defaultGate = GateMP_getDefaultRemote();

    if (defaultGate){
        key = GateMP_enter(defaultGate);
    }

    /* Invalidate cache before looking at the in-use flags */
    if (SharedRegion_isCacheEnabled(0)) {
        Cache_inv(inUse, num * sizeof(UInt8), Cache_Type_ALL, TRUE);
    }

    /*
     *  Find a free resource id. Note: zero is reserved on the
     *  system proxy for the default gate.
     */
    for (resourceId = 0; resourceId < num; resourceId++) {
        /*
         *  If not in-use, set the inUse to TRUE to prevent other
         *  creates from getting this one.
         */
        if (inUse[resourceId] == FALSE) {
           flag = TRUE;

           /* Denote in shared memory that the resource is used */
           inUse[resourceId] = TRUE;
           break;
       }
    }

    /* Write-back if a in-use flag was changed */
    if (flag == TRUE && SharedRegion_isCacheEnabled(0)) {
        Cache_wbInv(inUse, num * sizeof(UInt8), Cache_Type_ALL, TRUE);
    }

    /* Done with the critical section */
    if (defaultGate) {
        GateMP_leave(defaultGate, key);
    }

    if (flag == FALSE) {
        resourceId = (UInt)-1;
        Error_raise(eb, GateMP_E_gateUnavailable, 0, 0);
    }

    return (resourceId);
}
/*
 *                      SDOCM00068518 (handle NS_add failure)
 *! 14-Mar-2010 skp     SDOCM00068026 (GateMP_open incorrectly uses &&) 
 *! 13-Mar-2010 skp     SDOCM00067604 (GateMP_get[Local/Remote]Protect needed)
 *! 09-Mar-2010 skp     Fixed SDOCM00066819 (assert that region has valid heap)
 *! 18-Feb-2010 skp     Fixes for SDOCM00066913 (ROUND_UP issues)
 *! 17-Feb-2010 skp     Fixed SDOCM00066850: Need to set params->regionId
 *! 31-Jan-2010 jv      Added attach(), updated start().
 *! 29-Jan-2009 skp     Fixes for SDOCM00066370
 *! 20-Nov-2009 jv      Fixed bug with not checking if a GateMP is created
 *!                     before attempting to openByAddr().
 *! 11-Aug-2009 toddm   created
 */
/*
 *  @(#) ti.sdo.ipc; 1, 0, 0, 0,398; 4-9-2010 18:37:47; /db/vtree/library/trees/ipc/ipc-d23x/src/
 */

