/* 
 * Copyright (c) 2010
 * Texas Instruments
 *
 *  All rights reserved.  Property of Texas Instruments
 *  Restricted rights to use, duplicate or disclose this code are
 *  granted through contract.
 * 
 * */
/*
 *  ======== Ipc.c ========
 */

#include <xdc/std.h>
#include <xdc/runtime/Error.h>
#include <xdc/runtime/Assert.h>
#include <xdc/runtime/Memory.h>
#include <xdc/runtime/Startup.h>

#include <ti/sdo/ipc/NameServerRemoteNotify.h>

#include <ti/sysbios/BIOS.h>
#include <ti/sysbios/hal/Cache.h>
#include <ti/sysbios/knl/Task.h>

#include "package/internal/Ipc.xdc.h"

#include <ti/sdo/ipc/_Ipc.h>
#include <ti/sdo/ipc/_Notify.h>
#include <ti/sdo/ipc/_GateMP.h>
#include <ti/sdo/ipc/_SharedRegion.h>
#include <ti/sdo/utils/_MultiProc.h>
#include <ti/sdo/utils/_NameServer.h>
#include <ti/sdo/ipc/_MessageQ.h>

#ifdef __ti__
    #pragma FUNC_EXT_CALLED(Ipc_attach);
    #pragma FUNC_EXT_CALLED(Ipc_detach);
    #pragma FUNC_EXT_CALLED(Ipc_readConfig);
    #pragma FUNC_EXT_CALLED(Ipc_start);
    #pragma FUNC_EXT_CALLED(Ipc_writeConfig);
#endif

/*
 *  For no MMU case, it should be set to TRUE always.
 *  For MMU, it would be set by IPC link between local and HOST processors.
 */
extern __FAR__ Bits32 Ipc_sr0MemorySetup;


/*
 *  ======== Ipc_attach ========
 */
Int Ipc_attach(UInt16 remoteProcId)
{
    Ptr notifySharedAddr;
    Ptr nsrnSharedAddr;
    Ptr msgqSharedAddr;
    SizeT notifyMemReq;
    volatile Ipc_Reserved *slave;
    Ipc_ProcEntry *ipc;
    Error_Block eb;
    SizeT reservedSize = Ipc_reservedSizePerProc();
    Bool cacheEnabled = SharedRegion_isCacheEnabled(0);
    Int status;
    
    /* Assert that the remoteProc is valid and isn't our own */
    Assert_isTrue(remoteProcId < MultiProc_INVALIDID,
        MultiProc_A_invalidMultiProcId);
    Assert_isTrue(remoteProcId != MultiProc_self(), Ipc_A_invArgument);
    
    Error_init(&eb);

    /* determine if self is master or slave */
    if (MultiProc_self() < remoteProcId) {
        Ipc_module->procEntry[remoteProcId].slave = TRUE;
    }
    else {
        Ipc_module->procEntry[remoteProcId].slave = FALSE;
    }

    /* determine the slave's slot */
    slave = Ipc_getSlaveAddr(remoteProcId, Ipc_module->ipcSharedAddr);

    if (cacheEnabled) {
        Cache_inv((Ptr)slave, reservedSize, Cache_Type_ALL, TRUE);
    }

    /* get the attach paramters associated with remoteProcId */
    ipc = &(Ipc_module->procEntry[remoteProcId]);

    /* Synchronize the processors. */
    status = Ipc_procSyncStart(remoteProcId, Ipc_module->ipcSharedAddr);
    if (status < 0) {
        return (status);
    }

    /* must be called before SharedRegion_attach */ 
    status = GateMP_attach(remoteProcId, Ipc_module->gateMPSharedAddr);
    if (status < 0) {
        return (status);
    }

    /* retrieves the SharedRegion Heap handles */
    status = SharedRegion_attach(remoteProcId);
    if (status < 0) {
        return (status);
    }
    
    /* attach Notify if not yet attached and specified to set internal setup */
    if (!(Notify_intLineRegistered(remoteProcId, 0)) &&
        (ipc->entry.setupNotify)) {   
        /* call Notify_attach */
        if (Ipc_module->procEntry[remoteProcId].slave) {
            notifyMemReq = Notify_sharedMemReq(remoteProcId,
                                Ipc_module->ipcSharedAddr);

            /*
             *  calloc required here due to race condition.
             *  Its possible that the slave, who creates the instance,
             *  tries a sendEvent before the master has created its instance
             *  because the state of memory was enabled from a previous run.
             */
            notifySharedAddr = Memory_calloc(SharedRegion_getHeap(0),
                                   notifyMemReq,
                                   SharedRegion_getCacheLineSize(0),
                                   &eb);

            /* make sure alloc did not fail */
            if (notifySharedAddr == NULL) {
                return (Ipc_E_MEMORY);
            }

            /* if cache enabled, wbInv the calloc above */
            if (cacheEnabled) {
                Cache_wbInv(notifySharedAddr, notifyMemReq,
                            Cache_Type_ALL, TRUE);
            }

            /* set the notify SRPtr */
            slave->notifySRPtr = SharedRegion_getSRPtr(notifySharedAddr, 0);
        }
        else {
            /* get the notify SRPtr */
            notifySharedAddr = SharedRegion_getPtr(slave->notifySRPtr);
        }

        /* call attach to remote processor */
        status = Notify_attach(remoteProcId, notifySharedAddr);

        if (status < 0) {
            if (Ipc_module->procEntry[remoteProcId].slave) {
                /* free the memory back to SharedRegion 0 heap */
                Memory_free(SharedRegion_getHeap(0),
                    notifySharedAddr,
                    Notify_sharedMemReq(remoteProcId, notifySharedAddr));
            }

            return (status);
        }
    }
   
    /* Must come after GateMP_start because depends on default GateMP */
    if (!(NameServer_isRegistered(remoteProcId)) &&
        (ipc->entry.setupNotify)) {
        if (Ipc_module->procEntry[remoteProcId].slave) {
            nsrnSharedAddr = Memory_alloc(SharedRegion_getHeap(0),
                                 NameServerRemoteNotify_sharedMemReq(NULL),
                                 SharedRegion_getCacheLineSize(0),
                                 &eb);

            /* make sure alloc did not fail */
            if (nsrnSharedAddr == NULL) {
                return (Ipc_E_MEMORY);
            }

            /* set the NSRN SRPtr */
            slave->nsrnSRPtr = SharedRegion_getSRPtr(nsrnSharedAddr, 0);
        }
        else {
            /* get the NSRN SRPtr */
            nsrnSharedAddr = SharedRegion_getPtr(slave->nsrnSRPtr);
        }

        /* call attach to remote processor */
        status = NameServerRemoteNotify_attach(remoteProcId, nsrnSharedAddr);

        if (status < 0) {
            if (Ipc_module->procEntry[remoteProcId].slave) {
                /* free the memory back to SharedRegion 0 heap */
                Memory_free(SharedRegion_getHeap(0),
                            nsrnSharedAddr,
                            NameServerRemoteNotify_sharedMemReq(NULL));
            }

            return (status);
        }
    }
    
    /* Must come after GateMP_start because depends on default GateMP */
    if (!(MessageQ_SetupTransportProxy_isRegistered(remoteProcId)) &&
        (ipc->entry.setupMessageQ)) {
        if (Ipc_module->procEntry[remoteProcId].slave) {
            msgqSharedAddr = Memory_alloc(SharedRegion_getHeap(0),
                            MessageQ_sharedMemReq(Ipc_module->ipcSharedAddr),
                            SharedRegion_getCacheLineSize(0),
                            &eb);

            /* make sure alloc did not fail */
            if (msgqSharedAddr == NULL) {
                return (Ipc_E_MEMORY);
            }

            /* set the transport SRPtr */
            slave->transportSRPtr = SharedRegion_getSRPtr(msgqSharedAddr, 0);
        }
        else {
            /* get the transport SRPtr */
            msgqSharedAddr = SharedRegion_getPtr(slave->transportSRPtr);
        }

        /* call attach to remote processor */
        status = MessageQ_attach(remoteProcId, msgqSharedAddr);

        if (status < 0) {
            if (Ipc_module->procEntry[remoteProcId].slave) {
                /* free the memory back to SharedRegion 0 heap */
                Memory_free(SharedRegion_getHeap(0),
                            msgqSharedAddr,
                            MessageQ_sharedMemReq(msgqSharedAddr));
            }

            return (status);
        }
    }

    /* writeback invalidate slave's shared memory if cache enabled */
    if (cacheEnabled) {
        if (Ipc_module->procEntry[remoteProcId].slave) {
            Cache_wbInv((Ptr)slave, reservedSize, Cache_Type_ALL, TRUE);
        }
    }
    
    /* Call user attach fxn */
    if (Ipc_userFxn.attach) {
        status = Ipc_userFxn.attach(&ipc->userObj, remoteProcId);
        if (status < 0) {
            return (status);
        }
    }

    /* Finish the processor synchronization */
    status = Ipc_procSyncFinish(remoteProcId, Ipc_module->ipcSharedAddr);
    if (status < 0) {
        return (status);
    }
    
    return (status);
}

/*
 *  ======== Ipc_detach ========
 */
Int Ipc_detach(UInt16 remoteProcId)
{
    Ptr notifySharedAddr;
    Ptr nsrnSharedAddr;
    Ptr msgqSharedAddr;
    volatile Ipc_Reserved *slave, *master;
    Ipc_ProcEntry *ipc;
    SizeT reservedSize = Ipc_reservedSizePerProc();
    Bool cacheEnabled = SharedRegion_isCacheEnabled(0);
    Int status;

    /* get the paramters associated with remoteProcId */
    ipc = &(Ipc_module->procEntry[remoteProcId]);

    /* determine the slave's slot */
    slave = Ipc_getSlaveAddr(remoteProcId, Ipc_module->ipcSharedAddr);

    /* determine the master's slot */
    master = Ipc_getMasterAddr(remoteProcId, Ipc_module->ipcSharedAddr);

    if (cacheEnabled) {
        Cache_inv((Ptr)slave, reservedSize, Cache_Type_ALL, TRUE);
        Cache_inv((Ptr)master, reservedSize, Cache_Type_ALL, TRUE);
    }

    if (Ipc_module->procEntry[remoteProcId].slave) {
        /* check to make sure master is not trying to attach */
        while (master->startedKey == Ipc_PROCSYNCSTART) {
            Cache_inv((Ptr)master, reservedSize, Cache_Type_ALL, TRUE);
        }
    }
    else {
        /* check to make sure slave is not trying to attach */
        while (slave->startedKey == Ipc_PROCSYNCSTART) {
            Cache_inv((Ptr)slave, reservedSize, Cache_Type_ALL, TRUE);
        }
    }

    /* Call user detach fxn */
    if (Ipc_userFxn.detach) {
        status = Ipc_userFxn.detach(&ipc->userObj, remoteProcId);
        if (status < 0) {
            return (status);
        }
    }

    if (ipc->entry.setupMessageQ) {
        /* call MessageQ_detach for remote processor */
        status = MessageQ_detach(remoteProcId);
        if (status < 0) {
            return (status);
        }

        /* free the memory if slave processor */
        if (Ipc_module->procEntry[remoteProcId].slave) {
            /* get the pointer to MessageQ transport instance */
            msgqSharedAddr = SharedRegion_getPtr(slave->transportSRPtr);

            /* free the memory back to SharedRegion 0 heap */
            Memory_free(SharedRegion_getHeap(0),
                        msgqSharedAddr,
                        MessageQ_sharedMemReq(msgqSharedAddr));

            /* set the pointer for MessageQ transport instance back to NULL */
            slave->transportSRPtr = NULL;
        }
    }

    if (ipc->entry.setupNotify) {
        /* call NameServerRemoteNotify_detach for remote processor */
        status = NameServerRemoteNotify_detach(remoteProcId);
        if (status < 0) {
            return (status);
        }

        /* free the memory if slave processor */
        if (Ipc_module->procEntry[remoteProcId].slave) {
            /* get the pointer to NSRN instance */
            nsrnSharedAddr = SharedRegion_getPtr(slave->nsrnSRPtr);

            /* free the memory back to SharedRegion 0 heap */
            Memory_free(SharedRegion_getHeap(0),
                        nsrnSharedAddr,
                        NameServerRemoteNotify_sharedMemReq(NULL));

            /* set the pointer for NSRN instance back to NULL */
            slave->nsrnSRPtr = NULL;
        }
    }

    if (ipc->entry.setupNotify) {
        /* call Notify_detach for remote processor */
        status = Notify_detach(remoteProcId);
        if (status < 0) {
            return (status);
        }

        /* free the memory if slave processor */
        if (Ipc_module->procEntry[remoteProcId].slave) {
            /* get the pointer to Notify instance */
            notifySharedAddr = SharedRegion_getPtr(slave->notifySRPtr);

            /* free the memory back to SharedRegion 0 heap */
            Memory_free(SharedRegion_getHeap(0),
                        notifySharedAddr,
                        Notify_sharedMemReq(remoteProcId, notifySharedAddr));

            /* set the pointer for Notify instance back to NULL */
            slave->notifySRPtr = NULL;
        }
    }

    if (Ipc_module->procEntry[remoteProcId].slave) {
        slave->startedKey = 0;
        slave->configListHead = SharedRegion_INVALIDSRPTR;
        if (cacheEnabled) {
            Cache_wbInv((Ptr)slave, reservedSize, Cache_Type_ALL, TRUE);
        }
    }
    else {
        master->startedKey = 0;
        master->configListHead = SharedRegion_INVALIDSRPTR;
        if (cacheEnabled) {
            Cache_wbInv((Ptr)master, reservedSize, Cache_Type_ALL, TRUE);
        }
    }

    return (status);
}

/*
 *  ======== Ipc_getMasterAddr ========
 */
Ptr Ipc_getMasterAddr(UInt16 remoteProcId, Ptr sharedAddr)
{
    Int slot;
    UInt16 masterId;
    volatile Ipc_Reserved *master;
    SizeT reservedSize = Ipc_reservedSizePerProc();

    /* determine the master's procId and slot */
    if (MultiProc_self() < remoteProcId) {
        masterId = remoteProcId;
        slot = MultiProc_self();
    }
    else {
        masterId = MultiProc_self();
        slot = remoteProcId;
    }

    /* determine the reserve address for master between self and remote */
    master = (Ipc_Reserved *)((UInt32)sharedAddr +
             ((masterId * reservedSize) +
             (slot * sizeof(Ipc_Reserved))));

    return ((Ptr)master);
}

/*
 *  ======== Ipc_getRegion0ReservedSize ========
 */
SizeT Ipc_getRegion0ReservedSize(Void)
{
    SizeT reservedSize = Ipc_reservedSizePerProc();

    /* Calculate the total amount to reserve */
    reservedSize = reservedSize * MultiProc_numProcessors;
    
    return (reservedSize);
}

/*
 *  ======== Ipc_getSlaveAddr ========
 */
Ptr Ipc_getSlaveAddr(UInt16 remoteProcId, Ptr sharedAddr)
{
    Int slot;
    UInt16 slaveId;
    volatile Ipc_Reserved *slave;
    SizeT reservedSize = Ipc_reservedSizePerProc();

    /* determine the slave's procId and slot */
    if (MultiProc_self() < remoteProcId) {
        slaveId = MultiProc_self();
        slot = remoteProcId - 1;
    }
    else {
        slaveId = remoteProcId;
        slot = MultiProc_self() - 1;
    }

    /* determine the reserve address for slave between self and remote */
    slave = (Ipc_Reserved *)((UInt32)sharedAddr +
            ((slaveId * reservedSize) +
            (slot * sizeof(Ipc_Reserved))));

    return ((Ptr)slave);
}

/*
 *  ======== Ipc_procSyncStart ========
 *  The owner of SharedRegion 0 writes to its reserve memory address
 *  in region 0 to let the other processors know it has started.
 *  It then spins until the other processors start.
 *  The other processors write their reserve memory address in
 *  region 0 to let the owner processor know they've started.
 *  The other processors then spin until the owner processor writes
 *  to let them know that its finished the process of synchronization
 *  before continuing.
 */
Int Ipc_procSyncStart(UInt16 remoteProcId, Ptr sharedAddr)
{
    volatile Ipc_Reserved *self, *remote;
    Ipc_ProcEntry *ipc;
    SizeT reservedSize = Ipc_reservedSizePerProc();
    Bool cacheEnabled = SharedRegion_isCacheEnabled(0);

    /* don't do any synchronization if procSync is NONE */
    if (Ipc_procSync == Ipc_ProcSync_NONE) {
        return (Ipc_S_SUCCESS);
    }

    /* determine self and remote pointers */
    if (Ipc_module->procEntry[remoteProcId].slave) {
        self = Ipc_getSlaveAddr(remoteProcId, sharedAddr);
        remote = Ipc_getMasterAddr(remoteProcId, sharedAddr);
    }
    else {
        self = Ipc_getMasterAddr(remoteProcId, sharedAddr);
        remote = Ipc_getSlaveAddr(remoteProcId, sharedAddr);
    }

    /* construct the config list */
    ipc = &(Ipc_module->procEntry[remoteProcId]);

    ipc->localConfigList = (Ptr)&self->configListHead; 
    ipc->remoteConfigList = (Ptr)&remote->configListHead;

    ((Ipc_ConfigHead *)ipc->localConfigList)->first = SharedRegion_INVALIDSRPTR;

    if (cacheEnabled) {
        Cache_wbInv(ipc->localConfigList, reservedSize, Cache_Type_ALL, TRUE);
    }

    if (Ipc_module->procEntry[remoteProcId].slave) {
        /* set my processor's reserved key to start */
        self->startedKey = Ipc_PROCSYNCSTART;
        
        /* write back my processor's reserve key */
        if (cacheEnabled) {
            Cache_wbInv((Ptr)self, reservedSize, Cache_Type_ALL, TRUE);
        }
        
        /* wait for remote processor to start */
        if (cacheEnabled) {
            Cache_inv((Ptr)remote, reservedSize, Cache_Type_ALL, TRUE);
        }

        if (remote->startedKey != Ipc_PROCSYNCSTART) {
            return (Ipc_E_FAIL);
        }
    }
    else {
        /*  wait for remote processor to start */
        Cache_inv((Ptr)remote, reservedSize, Cache_Type_ALL, TRUE);

        if ((self->startedKey != Ipc_PROCSYNCSTART) &&
            (remote->startedKey != Ipc_PROCSYNCSTART)) {
            return (Ipc_E_FAIL);
        }
        
        /* set my processor's reserved key to start */
        self->startedKey = Ipc_PROCSYNCSTART;
        
        /* write my processor's reserve key back */
        if (cacheEnabled) {
            Cache_wbInv((Ptr)self, reservedSize, Cache_Type_ALL, TRUE);
        }

        /* wait for remote processor to finish */
        Cache_inv((Ptr)remote, reservedSize, Cache_Type_ALL, TRUE);

        if (remote->startedKey != Ipc_PROCSYNCFINISH) {
            return (Ipc_E_FAIL);
        }
    }

    return (Ipc_S_SUCCESS);
}

/*
 *  ======== Ipc_procSyncFinish ========
 *  Each processor writes its reserve memory address in SharedRegion 0
 *  to let the other processors know its finished the process of
 *  synchronization. 
 */
Int Ipc_procSyncFinish(UInt16 remoteProcId, Ptr sharedAddr)
{
    volatile Ipc_Reserved *self;
    SizeT reservedSize = Ipc_reservedSizePerProc();
    Bool cacheEnabled = SharedRegion_isCacheEnabled(0);

    /* don't do any synchronization if procSync is NONE */
    if (Ipc_procSync == Ipc_ProcSync_NONE) {
        return (Ipc_S_SUCCESS);
    }
    
    /* determine self pointer */
    if (Ipc_module->procEntry[remoteProcId].slave) {
        self = Ipc_getSlaveAddr(remoteProcId, sharedAddr);
    }
    else {
        self = Ipc_getMasterAddr(remoteProcId, sharedAddr);
    }
    
    /* set my processor's reserved key to finish */
    self->startedKey = Ipc_PROCSYNCFINISH;
    
    /* write back my processor's reserve key */
    if (cacheEnabled) {
        Cache_wbInv((Ptr)self, reservedSize, Cache_Type_ALL, TRUE);
    }

    return (Ipc_S_SUCCESS);
}

/*!
 *  ======== Ipc_readConfig ========
 */
Int Ipc_readConfig(UInt16 remoteProcId, UInt32 tag, Ptr cfg, SizeT size)
{
    Int status = Ipc_E_FAIL;
    volatile Ipc_ConfigEntry *entry;
    Bool cacheEnabled = SharedRegion_isCacheEnabled(0);
        
    if (cacheEnabled) {
        Cache_inv(Ipc_module->procEntry[remoteProcId].remoteConfigList, 
                  SharedRegion_getCacheLineSize(0), 
                  Cache_Type_ALL, 
                  TRUE);
    }

    entry = (Ipc_ConfigEntry *)((Ipc_ConfigHead *)
            Ipc_module->procEntry[remoteProcId].remoteConfigList)->first;

    while ((SharedRegion_SRPtr)entry != SharedRegion_INVALIDSRPTR) {
        entry = (Ipc_ConfigEntry *)
                SharedRegion_getPtr((SharedRegion_SRPtr)entry);

        /* Traverse the list to find the tag */
        if (cacheEnabled) {
            Cache_inv((Ptr)entry, 
                      size + sizeof(Ipc_ConfigEntry), 
                      Cache_Type_ALL, 
                      TRUE);
        }

        if ((entry->remoteProcId == remoteProcId) &&
            (entry->localProcId == MultiProc_self()) &&
            (entry->tag == tag)) {

            if (size == entry->size) {
                memcpy(cfg, 
                       (Ptr)((UInt32)entry + sizeof(Ipc_ConfigEntry)), 
                       entry->size);
                return (Ipc_S_SUCCESS);
            }
            else {
                return (Ipc_E_FAIL);
            }
        }

        entry = (Ipc_ConfigEntry *)entry->next;
    }

    return (status);
}

/*
 *  ======== Ipc_reservedSizePerProc ========
 */
SizeT Ipc_reservedSizePerProc(Void)
{
    SizeT reservedSize = sizeof(Ipc_Reserved) * MultiProc_numProcessors;
    SizeT cacheLineSize = SharedRegion_getCacheLineSize(0);

    /* Calculate amount to reserve per processor */
    if (cacheLineSize > reservedSize) {
        /* Use cacheLineSize if larger than reservedSize */
        reservedSize = cacheLineSize;
    }
    else {
        /* Round reservedSize to cacheLineSize */
        reservedSize = _Ipc_roundup(reservedSize, cacheLineSize);
    }

    return (reservedSize);
}

/*
 *  ======== Ipc_start ========
 */
Int Ipc_start()
{
    Int i;
    SharedRegion_Entry entry;
    Ptr ipcSharedAddr;
    Ptr gateMPSharedAddr;
    GateMP_Params gateMPParams;
    Int status;
    Bool lineAvailable;

    if (Ipc_generateSlaveDataForHost) {
        /* get Ipc_sr0MemorySetup out of the cache */
        Cache_inv(&Ipc_sr0MemorySetup,
              SharedRegion_cacheLineSize,
              Cache_Type_ALL,
              TRUE);

        /* check Ipc_sr0MemorySetup variable */
        if (Ipc_sr0MemorySetup == 0x0) {
            return (Ipc_E_FAIL);
        }
    }

    /* get region 0 information */
    SharedRegion_getEntry(0, &entry);

    /* if entry is not valid then return */
    if (entry.isValid == FALSE) {
        return (Ipc_E_FAIL);
    }

    /* 
     *  Need to reserve memory in region 0 for processor synchronization.
     *  This must done before SharedRegion_start().
     */
    ipcSharedAddr = SharedRegion_reserveMemory(0, Ipc_getRegion0ReservedSize());

    /* must reserve memory for GateMP before SharedRegion_start() */
    gateMPSharedAddr = SharedRegion_reserveMemory(0,
                       GateMP_getRegion0ReservedSize());
    
    /* Init params for default gate (must match those in GateMP_start()) */
    GateMP_Params_init(&gateMPParams);
    gateMPParams.localProtect  = GateMP_LocalProtect_TASKLET;

    if (MultiProc_numProcessors > 1) {
        gateMPParams.remoteProtect = GateMP_RemoteProtect_SYSTEM;
    }
    else {
        gateMPParams.remoteProtect = GateMP_RemoteProtect_NONE;
    }

    /* reserve memory for default gate before SharedRegion_start() */
    SharedRegion_reserveMemory(0, GateMP_sharedMemReq(&gateMPParams));

    /* clear the reserved memory */
    SharedRegion_clearReservedMemory();

    /* Set shared addresses */
    Ipc_module->ipcSharedAddr = ipcSharedAddr;
    Ipc_module->gateMPSharedAddr = gateMPSharedAddr;

    /* create default GateMP, must be called before SharedRegion start */ 
    status = GateMP_start(Ipc_module->gateMPSharedAddr);
    if (status < 0) {
        return (status);
    }

    /* create HeapMemMP in each SharedRegion */ 
    status = SharedRegion_start();
    if (status < 0) {
        return (status);
    }
    
    /* Call attach for all procs if procSync is ALL */
    if (Ipc_procSync == Ipc_ProcSync_ALL) {
        /* Must attach to owner first to get default GateMP and HeapMemMP */
        if (MultiProc_self() != entry.ownerProcId) {
            do {
                status = Ipc_attach(entry.ownerProcId);
            } while (status < 0);
        }

        /* Loop to attach to all other processors */
        for (i = 0; i < MultiProc_numProcessors; i++) {
            if ((i == MultiProc_self()) || (i == entry.ownerProcId)) {
                continue;
            }

            lineAvailable = Notify_SetupProxy_intLineAvailable(i);
            if (!lineAvailable) {
                continue;
            }

            /* call Ipc_attach for every remote processor */
            do {
                status = Ipc_attach(i);
            } while (status < 0);
        }
    }

    return (Ipc_S_SUCCESS);
}

/*!
 *  ======== Ipc_writeConfig ========
 */
Int Ipc_writeConfig(UInt16 remoteProcId, UInt32 tag, Ptr cfg, SizeT size)
{
    Int status = Ipc_S_SUCCESS;
    Ipc_ConfigEntry * entry;
    Error_Block eb;
    Bool cacheEnabled = SharedRegion_isCacheEnabled(0);
    SizeT cacheLineSize = SharedRegion_getCacheLineSize(0);
        
    Error_init(&eb);
        
    /* Allocate memory from the shared heap (System Heap) */
    entry = Memory_alloc(SharedRegion_getHeap(0), 
                         size + sizeof(Ipc_ConfigEntry),
                         cacheLineSize,
                         &eb);

    if (entry == NULL) {
        return (Ipc_E_FAIL);
    }
        
    entry->remoteProcId = remoteProcId;
    entry->localProcId = MultiProc_self();
    entry->tag = tag;
    entry->size = size;
    memcpy((Ptr)((UInt32)entry + sizeof(Ipc_ConfigEntry)), cfg, size);
        
    /* Create a linked-list of config */
    if (((Ipc_ConfigHead *)
        Ipc_module->procEntry[remoteProcId].localConfigList)->first 
        == SharedRegion_INVALIDSRPTR) {

        entry->next = SharedRegion_INVALIDSRPTR;
        ((Ipc_ConfigHead *)
            Ipc_module->procEntry[remoteProcId].localConfigList)->first = 
            (Bits32)SharedRegion_getSRPtr(entry, 0);
    }
    else {
        entry->next = ((Ipc_ConfigHead *)
            Ipc_module->procEntry[remoteProcId].localConfigList)->first;

        ((Ipc_ConfigHead *)
            Ipc_module->procEntry[remoteProcId].localConfigList)->first = 
            (Bits32)SharedRegion_getSRPtr(entry, 0);
    }

    if (cacheEnabled) {
        Cache_wbInv(Ipc_module->procEntry[remoteProcId].localConfigList, 
                    SharedRegion_getCacheLineSize(0), 
                    Cache_Type_ALL, 
                    FALSE);

        Cache_wbInv(entry, 
                    size + sizeof(Ipc_ConfigEntry), 
                    Cache_Type_ALL, 
                    TRUE);
    }

    return (status);
}

/*!
 *  ======== Ipc_dummy ========
 */
Void Ipc_dummy() 
{
}

/*
 */
/*
 *  @(#) ti.sdo.ipc; 1, 0, 0, 0,398; 4-9-2010 18:37:47; /db/vtree/library/trees/ipc/ipc-d23x/src/
 */

