/* 
 * Copyright (c) 2010
 * Texas Instruments
 *
 *  All rights reserved.  Property of Texas Instruments
 *  Restricted rights to use, duplicate or disclose this code are
 *  granted through contract.
 * 
 * */
/*
 *  ======== ListMP.c ========
 *  Implementation of functions specified in ListMP.xdc.
 */

#include <xdc/std.h>
#include <xdc/runtime/Error.h>
#include <xdc/runtime/Assert.h>
#include <xdc/runtime/Memory.h>
#include <xdc/runtime/Startup.h>
#include <xdc/runtime/System.h>
#include <xdc/runtime/IHeap.h>

#ifdef xdc_target__isaCompatible_64P
#include <ti/sysbios/family/c64p/Cache.h>
#else
#include <ti/sysbios/hal/Cache.h>
#endif

#include "package/internal/ListMP.xdc.h"

/* this must be included after the internal header file for now */
#include <ti/sdo/ipc/_Ipc.h>
#include <ti/sdo/utils/_MultiProc.h>
#include <ti/sdo/utils/_NameServer.h>
#include <ti/sdo/ipc/_GateMP.h>
#include <ti/sdo/ipc/_ListMP.h>
#include <ti/sdo/ipc/_SharedRegion.h>

#ifdef __ti__
    #pragma FUNC_EXT_CALLED(ListMP_Params_init);
    #pragma FUNC_EXT_CALLED(ListMP_create);
    #pragma FUNC_EXT_CALLED(ListMP_close);
    #pragma FUNC_EXT_CALLED(ListMP_delete);
    #pragma FUNC_EXT_CALLED(ListMP_open);
    #pragma FUNC_EXT_CALLED(ListMP_openByAddr);
    #pragma FUNC_EXT_CALLED(ListMP_sharedMemReq);
    #pragma FUNC_EXT_CALLED(ListMP_empty);
    #pragma FUNC_EXT_CALLED(ListMP_getGate);
    #pragma FUNC_EXT_CALLED(ListMP_getHead);
    #pragma FUNC_EXT_CALLED(ListMP_getTail);
    #pragma FUNC_EXT_CALLED(ListMP_insert);
    #pragma FUNC_EXT_CALLED(ListMP_next);
    #pragma FUNC_EXT_CALLED(ListMP_prev);
    #pragma FUNC_EXT_CALLED(ListMP_putHead);
    #pragma FUNC_EXT_CALLED(ListMP_putTail);
    #pragma FUNC_EXT_CALLED(ListMP_remove);
#endif

#define CACHEL1DLINESIZE        64

/*
 *************************************************************************
 *                       Instance functions
 *************************************************************************
 */
/*
 *  ======== ListMP_Instance_init ========
 */
Int ListMP_Instance_init(ListMP_Object *obj,
        const ti_sdo_ipc_ListMP_Params *params,
        Error_Block *eb)
{
    SharedRegion_SRPtr sharedShmBase;
    Ptr localAddr;
    Int status;
    ListMP_Params sparams;
    IHeap_Handle regionHeap;

    if (params->openFlag == TRUE) {
        /* Open by sharedAddr */
        obj->objType = Ipc_ObjType_OPENDYNAMIC;
        obj->attrs = (ListMP_Attrs *)params->sharedAddr;
        obj->regionId = SharedRegion_getId(&(obj->attrs->head));
        obj->cacheEnabled = SharedRegion_isCacheEnabled(obj->regionId);
        obj->cacheLineSize = SharedRegion_getCacheLineSize(obj->regionId);
    
        /* get the local address of the SRPtr */
        localAddr = SharedRegion_getPtr(obj->attrs->gateMPAddr);
        
        status = GateMP_openByAddr(localAddr, (GateMP_Handle *)&(obj->gate));
        if (status != GateMP_S_SUCCESS) {
            return (1);
        }
        
        return (0);
    }

    /* init the gate */
    if (params->gate != NULL) {
        obj->gate = params->gate;
    }
    else {
        obj->gate = GateMP_getDefaultRemote();
    }
    
    if (params->sharedAddr == NULL) {
        /* Creating using a shared region ID */                
        obj->objType = Ipc_ObjType_CREATEDYNAMIC_REGION;
        obj->regionId = params->regionId;
        obj->cacheEnabled = SharedRegion_isCacheEnabled(obj->regionId);
        obj->cacheLineSize = SharedRegion_getCacheLineSize(obj->regionId);
                
        /* Need to allocate from the heap */
        ListMP_Params_init(&sparams);
        sparams.regionId = params->regionId;
        obj->allocSize = ListMP_sharedMemReq(&sparams);
            
        regionHeap = SharedRegion_getHeap(obj->regionId);
        Assert_isTrue(regionHeap != NULL, SharedRegion_A_noHeap);

        /* The region heap will take care of the alignment */
        obj->attrs = Memory_alloc(regionHeap, obj->allocSize, 0, eb);

        if (obj->attrs == NULL) {
            return (2);
        }
    }
    else {
        /* Creating using sharedAddr */
        obj->regionId = SharedRegion_getId(params->sharedAddr);
            
        /* Assert that the buffer is in a valid shared region */
        Assert_isTrue(obj->regionId != SharedRegion_INVALIDREGIONID, 
                      Ipc_A_addrNotInSharedRegion);
                        
        /* Assert that sharedAddr is cache aligned */
        Assert_isTrue(((UInt32)params->sharedAddr % 
                      SharedRegion_getCacheLineSize(obj->regionId) == 0),
                      Ipc_A_addrNotCacheAligned);
                
        /* set object's cacheEnabled, objType, and attrs  */
        obj->cacheEnabled = SharedRegion_isCacheEnabled(obj->regionId);
        obj->cacheLineSize = SharedRegion_getCacheLineSize(obj->regionId);
        obj->objType = Ipc_ObjType_CREATEDYNAMIC;
        obj->attrs = (ListMP_Attrs *)params->sharedAddr;
    }

    /* init the head (to be empty) */
    ListMP_elemClear(&(obj->attrs->head));
       
    /* store the GateMP sharedAddr in the Attrs */
    obj->attrs->gateMPAddr = GateMP_getSharedAddr(obj->gate);

    /* last thing, set the status */
    obj->attrs->status = ListMP_CREATED;
        
    if (obj->cacheEnabled) {
        Cache_wbInv(obj->attrs, sizeof(ListMP_Attrs), Cache_Type_ALL, TRUE);
    }
        
    /* add to NameServer if name not NULL */
    if (params->name != NULL) {
        sharedShmBase = SharedRegion_getSRPtr(obj->attrs, obj->regionId);
        obj->nsKey = NameServer_addUInt32(ListMP_module->nameServer,
                                          params->name,
                                          (UInt32)sharedShmBase);
        if (obj->nsKey == NULL) {
            return (3);
        }
    }

    return (0);
}

/*
 *  ======== ListMP_Instance_finalize ========
 */
Void ti_sdo_ipc_ListMP_Instance_finalize(ListMP_Object *obj, Int status)
{
    if (obj->objType | (Ipc_ObjType_CREATEDYNAMIC |
                        Ipc_ObjType_CREATEDYNAMIC_REGION)) {
        /* List is being deleted */
        /* Remove entry from NameServer */
        if (obj->nsKey != NULL) {
            NameServer_removeEntry(ListMP_module->nameServer, obj->nsKey);
        }

        /* Set status to 'not created' */
        obj->attrs->status = 0;
        if (obj->cacheEnabled == TRUE) {
            Cache_wbInv(obj->attrs, sizeof(ListMP_Attrs), Cache_Type_ALL, TRUE);
        }

        /*
         *  Free the shared memory back to the region heap. If NULL, then the 
         *  Memory_alloc failed.
         */
        if (obj->objType == Ipc_ObjType_CREATEDYNAMIC_REGION &&
            obj->attrs != NULL) {
            Memory_free(SharedRegion_getHeap(obj->regionId), obj->attrs, 
                        obj->allocSize);
        }
    }
    else {
        /* List is being closed */
        /* Close the gate. If NULL, then GateMP_openByAddr failed. */
        if (obj->gate != NULL) {
            GateMP_close((GateMP_Handle *)&(obj->gate));
        }
    }
}

/*
 *  ======== ListMP_create ========
 */
ListMP_Handle ListMP_create(const ListMP_Params *params)
{
    ti_sdo_ipc_ListMP_Params mpParams;
    ListMP_Object *obj;
    Error_Block eb;

    Error_init(&eb);

    if (params != NULL) {
        /* init the module params struct */ 
        ti_sdo_ipc_ListMP_Params_init(&mpParams);
        mpParams.gate       = params->gate;
        mpParams.openFlag   = FALSE;
        mpParams.sharedAddr = params->sharedAddr;
        mpParams.name       = params->name;
        mpParams.regionId   = params->regionId;

        /* call the module create */
        obj = ti_sdo_ipc_ListMP_create(&mpParams, &eb);
    }
    else {
        obj = ti_sdo_ipc_ListMP_create(NULL, &eb);
    }

    return ((ListMP_Handle)obj);
}

/*
 *  ======== ListMP_delete ========
 */
Int ListMP_delete(ListMP_Handle *handlePtr)
{
    ti_sdo_ipc_ListMP_delete(handlePtr);
    
    return (ListMP_S_SUCCESS);
}

/*
 *  ======== ListMP_empty ========
 */
Bool ListMP_empty(ListMP_Object *obj)
{
    Bool flag = FALSE;
    IArg key;
    SharedRegion_SRPtr sharedHead;

    /* prevent another thread or processor from modifying the ListMP */
    key = GateMP_enter(obj->gate);

    if (obj->cacheEnabled) {
        /* invalidate the head to make sure we are not getting stale data */
        Cache_inv(&(obj->attrs->head), sizeof(ListMP_Elem), Cache_Type_ALL,
                  TRUE);
    }
    
    /* get the SRPtr for the head */
    sharedHead = SharedRegion_getSRPtr(&(obj->attrs->head), obj->regionId);

    /* if 'next' is ourself, then the ListMP must be empty */
    if (obj->attrs->head.next == sharedHead) {
        flag = TRUE;
    }

    /* leave the gate */
    GateMP_leave(obj->gate, key);

    return (flag);
}

/*
 *  ======== ListMP_getGate ========
 */
GateMP_Handle ListMP_getGate(ListMP_Object *obj)
{
    return (obj->gate);
}

/*
 *  ======== ListMP_getHead ========
 */
Ptr ListMP_getHead(ListMP_Object *obj)
{
    ListMP_Elem *elem;
    ListMP_Elem *localHeadNext;
    ListMP_Elem *localNext;
    UInt key;

    /* prevent another thread or processor from modifying the ListMP */
    key = GateMP_enter(obj->gate);

    if (obj->cacheEnabled) {
        /* invalidate the head to make sure we are not getting stale data */
        Cache_inv(&(obj->attrs->head), sizeof(ListMP_Elem), Cache_Type_ALL,
                  TRUE);
    }
    
    localHeadNext = SharedRegion_getPtr(obj->attrs->head.next);

    /* Assert that pointer is not NULL */
    Assert_isTrue(localHeadNext != NULL, Ipc_A_nullPointer);

    /* See if the ListMP was empty */
    if (localHeadNext == (ListMP_Elem *)(&(obj->attrs->head))) {
        /* Empty, return NULL */
        elem = NULL;
    }
    else {
        if (obj->cacheEnabled) {
            /* invalidate elem */
            Cache_inv(localHeadNext, sizeof(ListMP_Elem), Cache_Type_ALL, TRUE);
        }

        /* Elem to return */
        elem = localHeadNext;

        localNext = SharedRegion_getPtr(elem->next);
        
        /* Assert that pointer is not NULL */
        Assert_isTrue(localNext != NULL, Ipc_A_nullPointer);
    
        if (obj->cacheEnabled) {
            Cache_inv(localNext, sizeof(ListMP_Elem), Cache_Type_ALL, TRUE);
        }
        
        /* Fix the head of the list next pointer */
        obj->attrs->head.next = elem->next;

        /* Fix the prev pointer of the new first elem on the list */
        localNext->prev = localHeadNext->prev;

        if (obj->cacheEnabled) {
            Cache_wbInv(&(obj->attrs->head), sizeof(ListMP_Elem),
                        Cache_Type_ALL, FALSE);
            Cache_wbInv(localNext, sizeof(ListMP_Elem), Cache_Type_ALL, TRUE);
        }
    }

    GateMP_leave(obj->gate, key);

    return (elem);
}

/*
 *  ======== ListMP_getTail ========
 */
Ptr ListMP_getTail(ListMP_Object *obj)
{
    ListMP_Elem *elem;
    ListMP_Elem *localHeadPrev;
    ListMP_Elem *localPrev;
    UInt key;

    /* prevent another thread or processor from modifying the ListMP */
    key = GateMP_enter(obj->gate);

    if (obj->cacheEnabled) {
        /* invalidate the head to make sure we are not getting stale data */
        Cache_inv(&(obj->attrs->head), sizeof(ListMP_Elem), Cache_Type_ALL,
                  TRUE);
    }
    
    localHeadPrev = SharedRegion_getPtr(obj->attrs->head.prev);

    /* Assert that pointer is not NULL */
    Assert_isTrue(localHeadPrev != NULL, Ipc_A_nullPointer);

    /* See if the ListMP was empty */
    if (localHeadPrev == (ListMP_Elem *)(&obj->attrs->head)) {
        /* Empty, return NULL */
        elem = NULL;
    }
    else {
        if (obj->cacheEnabled) {
            /* invalidate elem */
            Cache_inv(localHeadPrev, sizeof(ListMP_Elem), Cache_Type_ALL, TRUE);
        }

        /* Elem to return */
        elem = localHeadPrev;

        localPrev = SharedRegion_getPtr(elem->prev);
        
        /* Assert that pointer is not NULL */
        Assert_isTrue(localPrev != NULL, Ipc_A_nullPointer);
        
        if (obj->cacheEnabled) {
            Cache_inv(localPrev, sizeof(ListMP_Elem), Cache_Type_ALL, TRUE);
        }

        /* Fix the head of the list prev pointer */
        obj->attrs->head.prev   = elem->prev;

        /* Fix the next pointer of the new last elem on the list */
        localPrev->next = localHeadPrev->next;

        if (obj->cacheEnabled) {
            Cache_wbInv(&(obj->attrs->head), sizeof(ListMP_Elem),
                        Cache_Type_ALL, FALSE);
            Cache_wbInv(localPrev, sizeof(ListMP_Elem), Cache_Type_ALL, TRUE);
        }
    }

    GateMP_leave(obj->gate, key);

    return (elem);
}


/*
 *  ======== ListMP_Params_init ========
 */
Void ListMP_Params_init(ListMP_Params *params)
{
    /* init the params to the defaults */
    params->gate       = NULL;
    params->sharedAddr = 0;
    params->name       = NULL;
    params->regionId   = 0;
}

/*
 *  ======== ListMP_putHead ========
 */
Int ListMP_putHead(ListMP_Object *obj, ListMP_Elem *elem)
{
    UInt key;
    Int  id;
    ListMP_Elem *localNextElem;
    SharedRegion_SRPtr sharedElem;

    /* safe to do outside the gate */
    id = SharedRegion_getId(elem);
    sharedElem = SharedRegion_getSRPtr(elem, id);

    /* prevent another thread or processor from modifying the ListMP */
    key = GateMP_enter(obj->gate);

    if (obj->cacheEnabled) {
        /* invalidate the head to make sure we are not getting stale data */
        Cache_inv(&(obj->attrs->head), sizeof(ListMP_Elem), Cache_Type_ALL,
                  TRUE);
    }
    
    /* add the new elem into the list */
    elem->next = obj->attrs->head.next;
    localNextElem = SharedRegion_getPtr(elem->next);
    
    /* Assert that pointer is not NULL */
    Assert_isTrue(localNextElem != NULL, Ipc_A_nullPointer);

    if (obj->cacheEnabled) {
        Cache_inv(localNextElem, sizeof(ListMP_Elem), Cache_Type_ALL, TRUE);
    }

    elem->prev = localNextElem->prev;
    localNextElem->prev = sharedElem;
    obj->attrs->head.next = sharedElem;

    if (obj->cacheEnabled) {
        /* need to do cache operations */
        Cache_wbInv(localNextElem, sizeof(ListMP_Elem), Cache_Type_ALL,
                    FALSE);
        Cache_wbInv(&(obj->attrs->head), sizeof(ListMP_Elem), Cache_Type_ALL,
                    FALSE);
    
        /* writeback invalidate only the elem structure */
        Cache_wbInv(elem, sizeof(ListMP_Elem), Cache_Type_ALL, TRUE);
    }

    GateMP_leave(obj->gate, key);
    
    return (ListMP_S_SUCCESS);
}

/*
 *  ======== ListMP_putTail ========
 */
Int ListMP_putTail(ListMP_Object *obj, ListMP_Elem *elem)
{
    UInt key;
    Int  id;
    ListMP_Elem *localPrevElem;
    SharedRegion_SRPtr sharedElem;

    /* safe to do outside the gate */
    id = SharedRegion_getId(elem);
    sharedElem = SharedRegion_getSRPtr(elem, id);

    /* prevent another thread or processor from modifying the ListMP */
    key = GateMP_enter(obj->gate);

    if (obj->cacheEnabled) {
        /* invalidate the head to make sure we are not getting stale data */
        Cache_inv(&(obj->attrs->head), sizeof(ListMP_Elem), Cache_Type_ALL,
                  TRUE);
    }
    
    /* add the new elem into the list */
    elem->prev = obj->attrs->head.prev;
    localPrevElem = SharedRegion_getPtr(elem->prev);

    /* Assert that pointer is not NULL */
    Assert_isTrue(localPrevElem != NULL, Ipc_A_nullPointer);

    if (obj->cacheEnabled) {
        Cache_inv(localPrevElem, sizeof(ListMP_Elem), Cache_Type_ALL, TRUE);
    }

    elem->next = localPrevElem->next;
    localPrevElem->next = sharedElem;
    obj->attrs->head.prev = sharedElem;

    if (obj->cacheEnabled) {
        /* need to do cache operations */
        Cache_wbInv(localPrevElem, sizeof(ListMP_Elem), Cache_Type_ALL, FALSE);
        Cache_wbInv(&(obj->attrs->head), sizeof(ListMP_Elem), Cache_Type_ALL,
                    FALSE);
    
        /* writeback invalidate only the elem structure */
        Cache_wbInv(elem, sizeof(ListMP_Elem), Cache_Type_ALL, TRUE);   
    }

    GateMP_leave(obj->gate, key);
    
    return (ListMP_S_SUCCESS);
}

/*
 *  ======== ListMP_insert ========
 */
Int ListMP_insert(ListMP_Object *obj, ListMP_Elem *newElem,
                  ListMP_Elem *curElem)
{
    UInt key;
    Int  id;
    ListMP_Elem *localPrevElem;
    SharedRegion_SRPtr sharedNewElem;
    SharedRegion_SRPtr sharedCurElem;

    /* get SRPtr for newElem */
    id = SharedRegion_getId(newElem);
    sharedNewElem = SharedRegion_getSRPtr(newElem, id);

    /* get SRPtr for curElem */
    id = SharedRegion_getId(curElem);
    sharedCurElem = SharedRegion_getSRPtr(curElem, id);

    /* prevent another thread or processor from modifying the ListMP */
    key = GateMP_enter(obj->gate);

    if (obj->cacheEnabled) {
        Cache_inv(curElem, sizeof(ListMP_Elem), Cache_Type_ALL, TRUE);
    }
    
    /* get Ptr for curElem->prev */
    localPrevElem = SharedRegion_getPtr(curElem->prev);

    if (obj->cacheEnabled) {
        Cache_inv(localPrevElem, sizeof(ListMP_Elem), Cache_Type_ALL, TRUE);
    }
    
    newElem->next       = sharedCurElem;
    newElem->prev       = curElem->prev;
    localPrevElem->next = sharedNewElem;
    curElem->prev       = sharedNewElem;

    if (obj->cacheEnabled) {
        Cache_wbInv(localPrevElem, sizeof(ListMP_Elem), Cache_Type_ALL, FALSE);
        Cache_wbInv(curElem, sizeof(ListMP_Elem), Cache_Type_ALL, FALSE);
    
        /* writeback invalidate only the elem structure  */
            Cache_wbInv(newElem, sizeof(ListMP_Elem), Cache_Type_ALL, TRUE);
    }

    GateMP_leave(obj->gate, key);
    
    return (ListMP_S_SUCCESS);
}

/*
 *  ======== ListMP_next ========
 */
Ptr ListMP_next(ListMP_Object *obj, ListMP_Elem *elem)
{
    ListMP_Elem *retElem;  /* returned elem */

    /* elem == NULL -> start at the head */
    if (elem == NULL) {
        elem = (ListMP_Elem *)&(obj->attrs->head);
    }

    if (obj->cacheEnabled) {
        Cache_inv(elem, sizeof(ListMP_Elem), Cache_Type_ALL, TRUE);
    }

    retElem = SharedRegion_getPtr(elem->next);

    if (retElem == (ListMP_Elem *)(&obj->attrs->head)) {
        retElem = NULL;
    }

    return (retElem);
}

/*
 *  ======== ListMP_prev ========
 */
Ptr ListMP_prev(ListMP_Object *obj, ListMP_Elem *elem)
{
    ListMP_Elem *retElem;  /* returned elem */

    /* elem == NULL -> start at the head */
    if (elem == NULL) {
        elem = (ListMP_Elem *)&(obj->attrs->head);
    }
    
    if (obj->cacheEnabled) {
        Cache_inv(elem, sizeof(ListMP_Elem), Cache_Type_ALL, TRUE);
    }

    retElem = SharedRegion_getPtr(elem->prev);

    if (retElem == (ListMP_Elem *)(&(obj->attrs->head))) {
        retElem = NULL;
    }

    return (retElem);
}

/*
 *  ======== ListMP_remove ========
 */
Int ListMP_remove(ListMP_Object *obj, ListMP_Elem *elem)
{
    UInt key;
    ListMP_Elem *localPrevElem;
    ListMP_Elem *localNextElem;

    /* prevent another thread or processor from modifying the ListMP */
    key = GateMP_enter(obj->gate);

    if (obj->cacheEnabled) {
        Cache_inv(elem, sizeof(ListMP_Elem), Cache_Type_ALL, TRUE);
    }
    
    localPrevElem = SharedRegion_getPtr(elem->prev);
    localNextElem = SharedRegion_getPtr(elem->next);

    if (obj->cacheEnabled) {
        Cache_inv(localPrevElem, sizeof(ListMP_Elem), Cache_Type_ALL, FALSE);
        Cache_inv(localNextElem, sizeof(ListMP_Elem), Cache_Type_ALL, TRUE);
    }
    
    localPrevElem->next = elem->next;
    localNextElem->prev = elem->prev;

    if (obj->cacheEnabled) {
        Cache_wbInv(localPrevElem, sizeof(ListMP_Elem), Cache_Type_ALL, FALSE);
        Cache_wbInv(localNextElem, sizeof(ListMP_Elem), Cache_Type_ALL, TRUE);
    }
    
    GateMP_leave(obj->gate, key);
    
    return (ListMP_S_SUCCESS);
}

/*
 *************************************************************************
 *                       Module functions
 *************************************************************************
 */
/*
 *  ======== ListMP_elemClear ========
 */
Void ListMP_elemClear(ti_sdo_ipc_ListMP_Elem *elem)
{
    SharedRegion_SRPtr sharedElem;
    Int id;

    id = SharedRegion_getId(elem);
    sharedElem = SharedRegion_getSRPtr(elem, id);

    elem->next = elem->prev = sharedElem;

    if (SharedRegion_isCacheEnabled(id)) {
        Cache_wbInv(elem, sizeof(ListMP_Elem), Cache_Type_ALL, TRUE);
    }
}

/*
 *  ======== ListMP_close ========
 */
Int ListMP_close(ListMP_Handle *handle)
{
    ListMP_delete(handle);
    
    return (ListMP_S_SUCCESS);
}

/*
 *  ======== ListMP_open ========
 */
Int ListMP_open(String name, ListMP_Handle *handlePtr)
{
    SharedRegion_SRPtr sharedShmBase;
    Ptr sharedAddr;
    Int status;
    
    /* Assert that a pointer has been supplied */
    Assert_isTrue(handlePtr != NULL, Ipc_A_nullArgument);
    
    /* Assert that we have enough information to open the gate */
    Assert_isTrue(name != NULL, Ipc_A_invParam);
                  
    /* Search NameServer */
    status = NameServer_getUInt32(ListMP_module->nameServer,
                                  name,
                                  &sharedShmBase,
                                  NULL);

    if (status < 0) {
        /* Name was not found */
        *handlePtr = NULL;
        return (ListMP_E_NOTFOUND);
    }

    sharedAddr = SharedRegion_getPtr(sharedShmBase);
    
    status = ListMP_openByAddr(sharedAddr, handlePtr);

    return (status);
}

/*
 *  ======== ListMP_openByAddr ========
 */
Int ListMP_openByAddr(Ptr sharedAddr, ListMP_Handle *handlePtr)
{
    ti_sdo_ipc_ListMP_Params params;
    ti_sdo_ipc_ListMP_Attrs *attrs;
    Error_Block eb;
    Int status;
    UInt16 id;
    
    Error_init(&eb);
    
    ti_sdo_ipc_ListMP_Params_init(&params);
    
    /* Tell Instance_init() that we're opening */
    params.openFlag = TRUE;
    
    attrs = (ti_sdo_ipc_ListMP_Attrs *)sharedAddr;
    params.sharedAddr = sharedAddr;
    id = SharedRegion_getId(sharedAddr);
    
    if (SharedRegion_isCacheEnabled(id)) {
        Cache_inv(attrs, sizeof(ti_sdo_ipc_ListMP_Attrs), Cache_Type_ALL, TRUE);
    }
    
    if (attrs->status != ListMP_CREATED) {
        *handlePtr = NULL;
        status = ListMP_E_NOTFOUND;
    }
    else {
        /* Create the object */
        *handlePtr = (ListMP_Handle)ti_sdo_ipc_ListMP_create(&params, &eb);
        if (*handlePtr == NULL) {
            status = ListMP_E_FAIL;
        }
        else {
            status = ListMP_S_SUCCESS;
        }
    }

    return (status);
}

/*
 *  ======== ListMP_sharedMemReq ========
 */
SizeT ListMP_sharedMemReq(const ListMP_Params *params)
{
    SizeT  memReq, minAlign;
    UInt16 regionId;

    if (params->sharedAddr == NULL) {
        regionId = params->regionId;
    }
    else {
        regionId = SharedRegion_getId(params->sharedAddr);
    }
    
    /* Assert that the region is valid */
    Assert_isTrue(regionId != SharedRegion_INVALIDREGIONID,
                  Ipc_A_addrNotInSharedRegion);

    minAlign = Memory_getMaxDefaultTypeAlign();
    if (SharedRegion_getCacheLineSize(regionId) > minAlign) {
        minAlign = SharedRegion_getCacheLineSize(regionId);
    }

    memReq = _Ipc_roundup(sizeof(ListMP_Attrs), minAlign);
    
    return (memReq);
}

/*
 */
/*
 *  @(#) ti.sdo.ipc; 1, 0, 0, 0,398; 4-9-2010 18:37:48; /db/vtree/library/trees/ipc/ipc-d23x/src/
 */

