/* 
 * Copyright (c) 2010
 * Texas Instruments
 *
 *  All rights reserved.  Property of Texas Instruments
 *  Restricted rights to use, duplicate or disclose this code are
 *  granted through contract.
 * 
 * */
/*
 *  ======== MessageQ.c ========
 *  Implementation of functions specified in MessageQ.xdc.
 */

#include <xdc/std.h>

#include <string.h>

#include <xdc/runtime/Error.h>
#include <xdc/runtime/Assert.h>
#include <xdc/runtime/Log.h>
#include <xdc/runtime/Memory.h>
#include <xdc/runtime/IHeap.h>
#include <xdc/runtime/Gate.h>
#include <xdc/runtime/IGateProvider.h>
#include <xdc/runtime/Startup.h>

#include <xdc/runtime/knl/ISync.h>
#include <xdc/runtime/knl/GateThread.h>

#include <ti/sysbios/syncs/SyncSem.h>

#include <ti/sdo/ipc/interfaces/IMessageQTransport.h>
#include <ti/sdo/utils/List.h>

#include "package/internal/MessageQ.xdc.h"

/* must be included after the internal header file for now */
#include <ti/sdo/ipc/_MessageQ.h>
#include <ti/sdo/utils/_MultiProc.h>
#include <ti/sdo/utils/_NameServer.h>

#ifdef __ti__
    #pragma FUNC_EXT_CALLED(MessageQ_Params_init);
    #pragma FUNC_EXT_CALLED(MessageQ_alloc);
    #pragma FUNC_EXT_CALLED(MessageQ_close);
    #pragma FUNC_EXT_CALLED(MessageQ_count);
    #pragma FUNC_EXT_CALLED(MessageQ_create);
    #pragma FUNC_EXT_CALLED(MessageQ_delete);
    #pragma FUNC_EXT_CALLED(MessageQ_open);
    #pragma FUNC_EXT_CALLED(MessageQ_free);
    #pragma FUNC_EXT_CALLED(MessageQ_get);
    #pragma FUNC_EXT_CALLED(MessageQ_getProcId);
    #pragma FUNC_EXT_CALLED(MessageQ_getQueueId);
    #pragma FUNC_EXT_CALLED(MessageQ_put);
    #pragma FUNC_EXT_CALLED(MessageQ_registerHeap);
    #pragma FUNC_EXT_CALLED(MessageQ_setReplyQueue);
    #pragma FUNC_EXT_CALLED(MessageQ_setMsgTrace);
    #pragma FUNC_EXT_CALLED(MessageQ_staticMsgInit);
    #pragma FUNC_EXT_CALLED(MessageQ_unblock);
    #pragma FUNC_EXT_CALLED(MessageQ_unregisterHeap);
#endif


/*
 *************************************************************************
 *                       Internal functions
 *************************************************************************
 */
/*
 *  ======== MessageQ_grow ========
 */
UInt16 MessageQ_grow(MessageQ_Object *obj, Error_Block *eb)
{
    UInt16 oldSize;
    UInt16 queueIndex = MessageQ_module->numQueues;    
    MessageQ_Handle *queues;
    MessageQ_Handle *oldQueues;

    oldSize = (MessageQ_module->numQueues) * sizeof(MessageQ_Handle);
    
    /* Allocate larger table */
    queues = Memory_alloc(MessageQ_Object_heap(), 
                          oldSize + sizeof(MessageQ_Handle), 0, eb);

    if (queues == NULL) {        
        return (MessageQ_INVALIDMESSAGEQ);
    }
    
    /* Copy contents into new table */
    memcpy(queues, MessageQ_module->queues, oldSize);
    
    /* Fill in the new entry */
    queues[queueIndex] = obj;
    
    /* Hook-up new table */
    oldQueues = MessageQ_module->queues;
    MessageQ_module->queues = queues;
    MessageQ_module->numQueues++;

    /* Delete old table if not statically defined */
    if (MessageQ_module->canFreeQueues == TRUE) {
        Memory_free(MessageQ_Object_heap(), oldQueues, oldSize);
    }
    else {
        MessageQ_module->canFreeQueues = TRUE;
    }

    return (queueIndex);
}
   
/*
 *************************************************************************
 *                       Instance functions
 *************************************************************************
 */
 
/*
 *  ======== MessageQ_Params_init ========
 */
Void MessageQ_Params_init(MessageQ_Params *params)
{
    params->synchronizer = NULL;
}

/*
 *  ======== MessageQ_Instance_init ========
 */
Int MessageQ_Instance_init(MessageQ_Object *obj, String name,
        const ti_sdo_ipc_MessageQ_Params *params, Error_Block *eb)
{
    Int              i;
    UInt16           start;
    UInt16           count;
    UInt             key;
    Bool             found = FALSE;    
    List_Handle      listHandle;
    SyncSem_Handle   syncSemHandle;
    MessageQ_QueueIndex queueIndex;    

    /* lock */
    key = IGateProvider_enter(MessageQ_module->gate);    

    start = MessageQ_Object_count(); 
    count = MessageQ_module->numQueues;
         
    /* Search the dynamic array for any holes */
    for (i = start; (i < count) && (found == FALSE); i++) {
        if (MessageQ_module->queues[i] == NULL) {
            MessageQ_module->queues[i] = (MessageQ_Handle)obj;            
            queueIndex = i;
            found = TRUE;            
        }
    } 
   
    /*
     *  If no free slot was found:
     *     - if no growth allowed, raise and error
     *     - if growth is allowed, grow the array
     */
    if (found == FALSE) {
        if (MessageQ_maxRuntimeEntries != NameServer_ALLOWGROWTH) {
            /* unlock scheduler */
            IGateProvider_leave(MessageQ_module->gate, key);    
            
            Error_raise(eb, MessageQ_E_maxReached, MessageQ_maxRuntimeEntries,
                        0);
            return (1);
        }
        else {
            queueIndex = MessageQ_grow(obj, eb);
            if (queueIndex == MessageQ_INVALIDMESSAGEQ) {
                /* unlock scheduler */
                IGateProvider_leave(MessageQ_module->gate, key);    
                return (2);
            }
        }
    }

    /* create default sync if not specified */
    if (params->synchronizer == NULL) {
        /* Create a SyncSem as the synchronizer */
        syncSemHandle = SyncSem_create(NULL, eb);

        if (syncSemHandle == NULL) {
            /* unlock scheduler */
            IGateProvider_leave(MessageQ_module->gate, key);    
            return (3);
        }

        /* store handle for use in finalize ...  */
        obj->syncSemHandle = syncSemHandle;

        obj->synchronizer = SyncSem_Handle_upCast(syncSemHandle);
    }
    else {
        obj->syncSemHandle = NULL;

        obj->synchronizer = params->synchronizer;
    }
    
    /* unlock scheduler */
    IGateProvider_leave(MessageQ_module->gate, key);    
         
    /* Fill in the message queue object */
    listHandle = MessageQ_Instance_State_normalList(obj);
    List_construct(List_struct(listHandle), NULL);                 

    listHandle = MessageQ_Instance_State_highList(obj);
    List_construct(List_struct(listHandle), NULL);                 

    obj->queue = ((MessageQ_QueueId)(MultiProc_self()) << 16) | queueIndex;
    
    obj->unblocked = FALSE;
    
    /* Add into NameServer */
    if (name != NULL) {
        obj->nsKey = NameServer_addUInt32(MessageQ_module->nameServer, name,
                                          obj->queue);
    }    
    
    return (0);
}

/*
 *  ======== MessageQ_Instance_finalize ========
 */
Void MessageQ_Instance_finalize(MessageQ_Object* obj, Int status)
{
    UInt key;
    MessageQ_QueueIndex index = (MessageQ_QueueIndex)(obj->queue);
    MessageQ_Msg tempMsg;
    List_Handle listHandle;

    if (obj->syncSemHandle != NULL) {
        SyncSem_delete(&obj->syncSemHandle);
    }

    listHandle = MessageQ_Instance_State_normalList(obj);
    
    /* Remove all the messages for the message queue's queue */
    while (TRUE) {
        tempMsg = (MessageQ_MsgHeader *)List_get(listHandle);
        if (tempMsg == NULL) {
            break;
        }
        MessageQ_free(tempMsg);
    }
    
    /* Destruct the list */
    List_destruct(List_struct(listHandle));
    
    listHandle = MessageQ_Instance_State_highList(obj);
    
    /* Remove all the messages for the message queue's queue */
    while (TRUE) {
        tempMsg = (MessageQ_MsgHeader *)List_get(listHandle);
        if (tempMsg == NULL) {
            break;
        }
        MessageQ_free(tempMsg);
    }
    
    /* Destruct the list */
    List_destruct(List_struct(listHandle));
    
    /* lock */
    key = IGateProvider_enter(MessageQ_module->gate);    
        
    /* Null out entry in the array. */
    MessageQ_module->queues[index] = NULL;
        
    /* unlock scheduler */
    IGateProvider_leave(MessageQ_module->gate, key);    
    
    if (obj->nsKey != NULL) {
        NameServer_removeEntry(MessageQ_module->nameServer, obj->nsKey);
    }
}

/*
 *  ======== MessageQ_close ======== 
 */
Int MessageQ_close(MessageQ_QueueId *queueId)
{
    *queueId = MessageQ_INVALIDMESSAGEQ;

    return (MessageQ_S_SUCCESS);
}

/*
 *  ======== MessageQ_count ========
 */
Int MessageQ_count(MessageQ_Object *obj)
{
    Int              count = 0;
    UInt             key;
    List_Elem       *tempMsg = NULL;
    List_Handle      listHandle;

    /* lock */
    key = Gate_enterSystem();
    
    /* Get the list */
    listHandle = MessageQ_Instance_State_highList(obj);
        
    /* Loop through and count the messages */
    while ((tempMsg = List_next(listHandle, tempMsg)) != NULL) {
        count++;
    }
    
    listHandle = MessageQ_Instance_State_normalList(obj);
        
    /* Loop through and count the messages */
    while ((tempMsg = List_next(listHandle, tempMsg)) != NULL) {
        count++;
    }

    /* unlock scheduler */
    Gate_leaveSystem(key);    
    
    return (count);
}

/*
 *  ======== MessageQ_get ========
 */
Int MessageQ_get(MessageQ_Object *obj, MessageQ_Msg *msg, UInt timeout)
{
    Int status;
    List_Handle highList;
    List_Handle normalList;
        
    /* Get the list */
    normalList = MessageQ_Instance_State_normalList(obj);
    highList   = MessageQ_Instance_State_highList(obj);
    
    /* Keep looping while there are no elements on either list */
    *msg = (MessageQ_Msg)List_get(highList);
    while (*msg == NULL) {
        *msg = (MessageQ_Msg)List_get(normalList);
        if (*msg == NULL) {
            /*  Block until notified. */
            status = ISync_wait(obj->synchronizer, timeout, NULL);
            if (status == ISync_WaitStatus_TIMEOUT) {
                return (MessageQ_E_TIMEOUT);
            }
            else if (status < 0) {
                return (MessageQ_E_FAIL);
            }
            
            if (obj->unblocked) {
                /* *(msg) may be NULL */
                return (MessageQ_E_UNBLOCKED);
            }
            
            *msg = (MessageQ_Msg)List_get(highList);
        }
    }
    
    if ((MessageQ_traceFlag == TRUE) ||
        (((*msg)->flags & MessageQ_TRACEMASK) != 0)) {                    
        Log_write4(MessageQ_LM_get, (UArg)(*msg), (UArg)((*msg)->seqNum), 
                   (UArg)((*msg)->srcProc), (UArg)(obj));
    }

    return (MessageQ_S_SUCCESS);
}

/*
 *  ======== MessageQ_getProcId ========
 *  Determine where the queue resides
 */
UInt16 MessageQ_getProcId(MessageQ_Object *obj)
{
    UInt16  flag;

    flag = (UInt16)(obj->queue >> 16);

    return (flag);
}

/*
 *  ======== MessageQ_getQueueId ======== 
 */
MessageQ_QueueId MessageQ_getQueueId(MessageQ_Object *obj)
{
    MessageQ_QueueId  queueId;

    queueId = (obj->queue);

    return (queueId);
}

/*
 *  ======== MessageQ_setReplyQueue ========
 *  Embed the source message queue into a message.
 */
Void MessageQ_setReplyQueue(MessageQ_Object *obj, MessageQ_Msg msg)
{
    msg->replyId   = (UInt16)(obj->queue);
    msg->replyProc = (UInt16)(obj->queue >> 16);
}

/*
 *************************************************************************
 *                      Module functions
 *************************************************************************
 */

/*
 *  ======== MessageQ_Module_startup ========
 */
Int MessageQ_Module_startup(Int phase)
{
    Int i;
    
    /* Ensure NameServer Module_startup() has completed */
    if (NameServer_Module_startupDone() == FALSE) {
        return (Startup_NOTDONE);
    }
    
    if (GateThread_Module_startupDone() == FALSE) {
        return (Startup_NOTDONE);
    }
    
    if (MessageQ_module->gate == NULL) {
        MessageQ_module->gate = 
            GateThread_Handle_upCast(GateThread_create(NULL, NULL));
    }
        
    /* Loop through all the static objects and set the id */
    for (i = 0; i < MessageQ_Object_count(); i++) {
        MessageQ_module->queues[i] =
            (MessageQ_Object *)MessageQ_Object_get(NULL, i);
    }

    /* Null out the dynamic ones */
    for (i = MessageQ_Object_count(); i < MessageQ_module->numQueues; i++) {
        MessageQ_module->queues[i] = NULL;
    }
    
    return (Startup_DONE);    
}

/*
 *  ======== MessageQ_msgInit ========
 *  This is a helper function to initialize a message. 
 */
static Void MessageQ_msgInit(MessageQ_Msg msg)
{
    IArg key;
    
    msg->replyId = (UInt16)MessageQ_INVALIDMESSAGEQ;
    msg->msgId   = MessageQ_INVALIDMSGID;
    msg->dstId   = (UInt16)MessageQ_INVALIDMESSAGEQ;    
    msg->flags   = MessageQ_HEADERVERSION | 
                   MessageQ_NORMALPRI |
                   (MessageQ_TRACEMASK & 
                   (MessageQ_traceFlag << MessageQ_TRACESHIFT));
    msg->srcProc = MultiProc_self();
    
    key = Gate_enterSystem();
    msg->seqNum  = MessageQ_module->seqNum++;
    Gate_leaveSystem(key);
}

/*
 *  ======== MessageQ_staticMsgInit ========
 */
Void MessageQ_staticMsgInit(MessageQ_Msg msg, Uint32 size)
{
    Assert_isTrue((msg != NULL), MessageQ_A_invalidMsg);
    
    MessageQ_msgInit(msg);

    msg->heapId  = MessageQ_STATICMSG;    
    msg->msgSize = size;
    
    if (MessageQ_traceFlag == TRUE) {
        Log_write3(MessageQ_LM_staticMsgInit, (UArg)(msg), (UArg)(msg->seqNum), 
                   (UArg)(msg->srcProc));
    }
}

/*
 *  ======== MessageQ_alloc ========
 *  Allocate a message and initial the needed fields (note some
 *  of the fields in the header at set via other APIs or in the
 *  MessageQ_put function.
 */
MessageQ_Msg MessageQ_alloc(UInt16 heapId, Uint32 size)
{
    MessageQ_Msg msg;
    Error_Block eb;

    Assert_isTrue((heapId < MessageQ_module->numHeaps),
                  MessageQ_A_heapIdInvalid);

    Assert_isTrue((MessageQ_module->heaps[heapId] != NULL),
                  MessageQ_A_heapIdInvalid);

    /* Allocate the message. No alignment requested */
    Error_init(&eb);
    msg = Memory_alloc(MessageQ_module->heaps[heapId], size, 0, &eb);

    if (msg == NULL) {
        return (NULL);
    }

    /* Fill in the fields of the message */
    MessageQ_msgInit(msg);
    msg->msgSize = size;
    msg->heapId  = heapId;
        
    if (MessageQ_traceFlag == TRUE) {
        Log_write3(MessageQ_LM_alloc, (UArg)(msg), (UArg)(msg->seqNum), 
                (UArg)(msg->srcProc));
    }

    return (msg);
}

/*
 *  ======== MessageQ_free ========
 */
Int MessageQ_free(MessageQ_Msg msg)
{
    IHeap_Handle heap;

    /* make sure msg is not NULL */
    Assert_isTrue((msg != NULL), MessageQ_A_invalidMsg);
    
    /* Cannot free a message that was initialized via MessageQ_staticMsgInit */
    Assert_isTrue((msg->heapId != MessageQ_STATICMSG), 
                  MessageQ_A_cannotFreeStaticMsg);
    
    Assert_isTrue((msg->heapId < MessageQ_module->numHeaps),
                  MessageQ_A_heapIdInvalid);

    Assert_isTrue((MessageQ_module->heaps[msg->heapId] != NULL),
                  MessageQ_A_heapIdInvalid);

    if ((MessageQ_traceFlag == TRUE) ||
        (msg->flags & MessageQ_TRACEMASK) != 0) {                    
        Log_write3(MessageQ_LM_free, (UArg)(msg), (UArg)(msg->seqNum), 
                   (UArg)(msg->srcProc));
    }
    
    heap = MessageQ_module->heaps[msg->heapId];
    
    if (heap != NULL) {
        Memory_free(heap, msg, msg->msgSize);
    }

    return (MessageQ_S_SUCCESS);
}

/*
 *  ======== MessageQ_open ========
 */
Int MessageQ_open(String name, MessageQ_QueueId *queueId)
{
    Int         status;
    Error_Block eb;

    Assert_isTrue(name != NULL, MessageQ_A_invalidParam);
    Assert_isTrue(queueId != NULL, MessageQ_A_invalidParam);

    Error_init(&eb);
    
    /* Search NameServer */
    status = NameServer_getUInt32(MessageQ_module->nameServer, name,
                                  queueId, NULL);

    if (status >= 0) {
        return (MessageQ_S_SUCCESS);    /* name found */
    }
    else {
        return (MessageQ_E_NOTFOUND);   /* name not found */
    }
}

/*
 *  ======== MessageQ_put ========
 */
Int MessageQ_put(MessageQ_QueueId queueId, MessageQ_Msg msg)
{
    IMessageQTransport_Handle transport;
    MessageQ_QueueIndex dstProcId = (MessageQ_QueueIndex)(queueId >> 16);
    List_Handle       listHandle;
    Int               status;
    UInt              priority;
    UInt16            flags;
    UInt16            seqNum;
    UInt16            srcProc;
    MessageQ_Object   *obj;

    Assert_isTrue((msg != NULL), MessageQ_A_invalidMsg);    
    
    msg->dstId   = (UInt16)(queueId);
    msg->dstProc = (UInt16)(queueId >> 16);
    
    if (dstProcId != MultiProc_self()) {
        /* assert that dstProcId is valid */
        Assert_isTrue(dstProcId < MultiProc_numProcessors,
                      MessageQ_A_procIdInvalid);
        
        /* Put the high and urgent messages to the high priority transport */
        priority = (UInt)((msg->flags) & MessageQ_TRANSPORTPRIORITYMASK);

        /* Call the transport associated with this message queue */
        transport = MessageQ_module->transports[dstProcId][priority];
        if (transport == NULL) {
            /* Try the other transport */
            priority = !priority;
            transport = MessageQ_module->transports[dstProcId][priority];
        }
        
        /* assert transport is not null */
        Assert_isTrue(transport != NULL, MessageQ_A_unregisteredTransport);

        /* use local vars so msg does not get cached after put */
        flags = msg->flags;

        if ((MessageQ_traceFlag) ||
            (flags & MessageQ_TRACEMASK) != 0) {
            /* use local vars so msg does not get cached after put */
            seqNum  = msg->seqNum;
            srcProc = msg->srcProc;
        }

        /* put msg to remote processor using transport */
        if (IMessageQTransport_put(transport, msg)) {
            status = MessageQ_S_SUCCESS;

            /* if trace enabled */
            if ((MessageQ_traceFlag) ||
                (flags & MessageQ_TRACEMASK) != 0) {                    
                Log_write4(MessageQ_LM_putRemote, (UArg)(msg), 
                          (UArg)(seqNum), (UArg)(srcProc),
                          (UArg)(dstProcId));
            }
        }
        else {
            status = MessageQ_E_FAIL;
        }       
    }
    else {
        /* Assert queueId is valid */
        Assert_isTrue((UInt16)queueId < MessageQ_module->numQueues,
                      MessageQ_A_invalidQueueId);

        /* It is a local MessageQ */
        obj = MessageQ_module->queues[(UInt16)(queueId)];

        /* Assert object is not NULL */
        Assert_isTrue(obj != NULL, MessageQ_A_invalidObj);

        if ((msg->flags & MessageQ_PRIORITYMASK) == MessageQ_URGENTPRI) {
            listHandle = MessageQ_Instance_State_highList(obj);
            List_putHead(listHandle, (List_Elem *)msg);
        }
        else {
            if ((msg->flags & MessageQ_PRIORITYMASK) == MessageQ_NORMALPRI) {
                listHandle = MessageQ_Instance_State_normalList(obj);
            }
            else {
                listHandle = MessageQ_Instance_State_highList(obj);
            }
            /* put on the queue */
            List_put(listHandle, (List_Elem *)msg);
        }

        ISync_signal(obj->synchronizer);

        status = MessageQ_S_SUCCESS;

        if ((MessageQ_traceFlag == TRUE) ||
            (msg->flags & MessageQ_TRACEMASK) != 0) {
            Log_write4(MessageQ_LM_putLocal, (UArg)(msg), 
                       (UArg)(msg->seqNum), (UArg)(msg->srcProc), (UArg)(obj));
        }
    }

    return (status);
}

/*
 *  ======== MessageQ_registerHeap ========
 *  Register a heap
 */
Int MessageQ_registerHeap(Ptr heap, UInt16 heapId)
{
    Int status;
    UInt key;
    IHeap_Handle iheap = (IHeap_Handle)heap;

    /* Make sure the heapId is valid */
    Assert_isTrue((heapId < MessageQ_module->numHeaps),
                  MessageQ_A_heapIdInvalid);

    /* lock scheduler */
    key = Gate_enterSystem();

    /* Make sure the id is not already in use */
    if (MessageQ_module->heaps[heapId] == NULL) {
        MessageQ_module->heaps[heapId] = iheap;    
        status = MessageQ_S_SUCCESS;
    }
    else {
        status = MessageQ_E_ALREADYEXISTS;
    }
        
    /* unlock scheduler */
    Gate_leaveSystem(key);
        
    return (status);
}

/*
 *  ======== MessageQ_unregisterHeap ========
 *  Unregister a heap
 */
Int MessageQ_unregisterHeap(UInt16 heapId)
{
    UInt key;

    Assert_isTrue((heapId < MessageQ_module->numHeaps),
                  MessageQ_A_heapIdInvalid);
    
    /* lock scheduler */
    key = Gate_enterSystem();

    MessageQ_module->heaps[heapId] = NULL;
        
    /* unlock scheduler */
    Gate_leaveSystem(key);

    return (MessageQ_S_SUCCESS);
}

/*
 *  ======== MessageQ_registerTransport ========
 *  Register a transport
 */
Bool MessageQ_registerTransport(IMessageQTransport_Handle handle, UInt16 procId,
                                UInt priority)
{
    Bool flag = FALSE;
    IArg key;

    /* Make sure the procId is valid */
    Assert_isTrue(procId < MultiProc_numProcessors, MessageQ_A_procIdInvalid);

    /* lock scheduler */
    key = Gate_enterSystem();

    /* Make sure the id is not already in use */
    if (MessageQ_module->transports[procId][priority] == NULL) {
        MessageQ_module->transports[procId][priority] = handle;
        flag = TRUE;
    }
        
    /* unlock scheduler */
    Gate_leaveSystem(key);

    return (flag);    
}

/*
 *  ======== MessageQ_unregisterTransport ========
 *  Unregister a heap
 */
Void MessageQ_unregisterTransport(UInt16 procId, UInt priority)
{
    IArg key;

    /* Make sure the procId is valid */
    Assert_isTrue(procId < MultiProc_numProcessors, MessageQ_A_procIdInvalid);

    /* lock scheduler */
    key = Gate_enterSystem();

    MessageQ_module->transports[procId][priority] = NULL;
    
    /* unlock scheduler */
    Gate_leaveSystem(key);
}

/*
 *  ======== MessageQ_setMsgTrace ======== 
 */
Void MessageQ_setMsgTrace(MessageQ_Msg msg, Bool traceFlag)
{
    msg->flags = (msg->flags & ~MessageQ_TRACEMASK) |  
                 (traceFlag << MessageQ_TRACESHIFT);
    Log_write4(MessageQ_LM_setTrace, (UArg)(msg), (UArg)(msg->seqNum),
               (UArg)(msg->srcProc), (UArg)(traceFlag));
}

/*
 *  ========  MessageQ_sharedMemReq ========
 */
SizeT MessageQ_sharedMemReq(Ptr sharedAddr)
{
    SizeT memReq = 0;
    
    if (MultiProc_numProcessors > 1) {
        /* Determine device-specific shared memory requirements */
        memReq =  MessageQ_SetupTransportProxy_sharedMemReq(sharedAddr);
    }

    return (memReq);
}

/*
 *  ======== MessageQ_attach ========
 */
Int MessageQ_attach(UInt16 remoteProcId, Ptr sharedAddr)
{
    Int status = MessageQ_E_FAIL;

    if (MultiProc_numProcessors > 1) {
        /* Use the MessageQ_SetupTransportProxy to attach transports */
        status = MessageQ_SetupTransportProxy_attach(remoteProcId, sharedAddr);
    }

    return (status);
}

/*
 *  ======== MessageQ_detach ========
 */
Int MessageQ_detach(UInt16 remoteProcId)
{
    Int status = MessageQ_E_FAIL;

    if (MultiProc_numProcessors > 1) {
        /* Use the MessageQ_SetupTransportProxy to detach transports */
        status = MessageQ_SetupTransportProxy_detach(remoteProcId);
    }

    return (status);
}

/*
 *  ======== MessageQ_create ========
 */
MessageQ_Handle MessageQ_create(String name, const MessageQ_Params *params)
{
    ti_sdo_ipc_MessageQ_Handle handle;
    ti_sdo_ipc_MessageQ_Params prms;
    Error_Block eb;

    Error_init(&eb);

    if (params != NULL) {
        ti_sdo_ipc_MessageQ_Params_init(&prms);
        
        prms.synchronizer = params->synchronizer;

        handle = ti_sdo_ipc_MessageQ_create(name, &prms, &eb);
    }
    else {
        handle = ti_sdo_ipc_MessageQ_create(name, NULL, &eb);
    }

    return ((MessageQ_Handle)handle);
}

/*
 *  ======== MessageQ_delete ========
 */
Int MessageQ_delete(MessageQ_Handle *handleP)
{
    ti_sdo_ipc_MessageQ_Handle *instp;

    instp = (ti_sdo_ipc_MessageQ_Handle *)handleP;

    ti_sdo_ipc_MessageQ_delete(instp);

    return (MessageQ_S_SUCCESS);
}

/*
 *  ======== MessageQ_unblock ========
 */
Void MessageQ_unblock(MessageQ_Object *obj)
{
    /* Assert that the queue is using a blocking synchronizer */
    Assert_isTrue(ISync_query((obj->synchronizer), ISync_Q_BLOCKING) == TRUE,
        MessageQ_A_invalidUnblock);
    
    /* Set instance to 'unblocked' state */
    obj->unblocked = TRUE;
    
    /* Signal the synchronizer */
    ISync_signal(obj->synchronizer);
}

/*
 */
/*
 *  @(#) ti.sdo.ipc; 1, 0, 0, 0,398; 4-9-2010 18:37:48; /db/vtree/library/trees/ipc/ipc-d23x/src/
 */

