/* 
 * Copyright (c) 2010
 * Texas Instruments
 *
 *  All rights reserved.  Property of Texas Instruments
 *  Restricted rights to use, duplicate or disclose this code are
 *  granted through contract.
 *
 * */
/*
 *  ======== NameServerRemoteNotify.c ========
 */

/* Remove Windows strncpy warning */
#define _CRT_SECURE_NO_DEPRECATE 1

#include <xdc/std.h>
#include <string.h>
#include <stdlib.h>

#include <xdc/runtime/Assert.h>
#include <xdc/runtime/Error.h>
#include <xdc/runtime/knl/ISync.h>

#include <ti/sysbios/BIOS.h>
#include <ti/sysbios/hal/Cache.h>
#include <ti/sysbios/knl/Swi.h>
#include <ti/sysbios/ipc/Semaphore.h>

#include "package/internal/NameServerRemoteNotify.xdc.h"

#include <ti/sdo/ipc/_Ipc.h>
#include <ti/sdo/ipc/_Notify.h>
#include <ti/sdo/ipc/_GateMP.h>
#include <ti/sdo/utils/_MultiProc.h>
#include <ti/sdo/utils/_NameServer.h>
#include <ti/sdo/ipc/_SharedRegion.h>

/*
 *************************************************************************
 *                       Instance functions
 *************************************************************************
 */
Int NameServerRemoteNotify_Instance_init(NameServerRemoteNotify_Object *obj,
                        UInt16 remoteProcId,
                        const NameServerRemoteNotify_Params *params,
                        Error_Block *eb)
{
    Int               offset = 0;
    Int               status;
    Semaphore_Params  semParams;
    Semaphore_Handle  semRemoteWait;
    Semaphore_Handle  semMultiBlock;
    Swi_Handle        swiHandle;
    Swi_Params        swiParams;
    
    /* Assert that a NameServerRemoteNotify_Params has been supplied */
    Assert_isTrue(params != NULL, Ipc_A_nullArgument);

    /* Assert that remoteProcId is valid */
    Assert_isTrue(remoteProcId != MultiProc_self() &&
                  remoteProcId != MultiProc_INVALIDID,
                  Ipc_A_invParam);
   
    /* determine the offset based upon own id and remote's id */
    if (MultiProc_self() > remoteProcId) {
            offset = 1;
    }
    
    obj->regionId = SharedRegion_getId(params->sharedAddr);
    
    /* Assert that sharedAddr is cache aligned */
    Assert_isTrue(((UInt32)params->sharedAddr % 
                  SharedRegion_getCacheLineSize(obj->regionId) == 0),
                  Ipc_A_addrNotCacheAligned);

    semRemoteWait = NameServerRemoteNotify_Instance_State_semRemoteWait(obj);
    semMultiBlock = NameServerRemoteNotify_Instance_State_semMultiBlock(obj);
    swiHandle = NameServerRemoteNotify_Instance_State_swiObj(obj);

    obj->msg[0] = (NameServerRemoteNotify_Message *)(params->sharedAddr);
    obj->msg[1] = (NameServerRemoteNotify_Message *)((UInt32)obj->msg[0] +
                          sizeof(NameServerRemoteNotify_Message));
    obj->gate = params->gate;
    obj->remoteProcId = remoteProcId;
    obj->notifyEventId = NameServerRemoteNotify_notifyEventId;

    /* construct the Semaphore */
    Semaphore_Params_init(&semParams);
    Semaphore_construct(Semaphore_struct(semRemoteWait), 0, &semParams);

    /* constrcut the semaphore */
    Semaphore_construct(Semaphore_struct(semMultiBlock), 1, &semParams);
    
    /* swi created with lowest priority and fxn = swiFxn */
    Swi_Params_init(&swiParams);
    swiParams.arg0 = (UArg)obj;
    swiParams.priority = 0;
    Swi_construct(Swi_struct(swiHandle),
                 (ti_sysbios_knl_Swi_FuncPtr)NameServerRemoteNotify_swiFxn,
                 &swiParams, eb);

    /* initialize own side of message struct only */
    obj->msg[offset]->request = 0;
    obj->msg[offset]->response = 0;
    obj->msg[offset]->requestStatus = 0;
    obj->msg[offset]->value = 0;
    obj->msg[offset]->valueLen = 0;
    
    memset(obj->msg[offset]->instanceName, 0,
               sizeof(obj->msg[offset]->instanceName));
    memset(obj->msg[offset]->name, 0,
               sizeof(obj->msg[offset]->name));

    /* determine cacheability of the object from the regionId */
    obj->cacheEnable = SharedRegion_isCacheEnabled(obj->regionId);
    if (obj->cacheEnable) {
        /* write back shared memory that was modified */
        Cache_wbInv(obj->msg[offset], sizeof(NameServerRemoteNotify_Message),
                    Cache_Type_ALL, TRUE);
    }

    /* register the call back function and event Id with notify */
    status = Notify_registerEventSingle(
                    remoteProcId,
                0,
                obj->notifyEventId,
                (Notify_FnNotifyCbck)NameServerRemoteNotify_cbFxn,
                (UArg)swiHandle);

    /* if not successful return */
    if (status != Notify_S_SUCCESS) {
        return (1);
    }

    /* register the remote driver with NameServer */
    NameServer_registerRemoteDriver(NameServerRemoteNotify_Handle_upCast(obj),
            remoteProcId);

    return (0);
}

/*
 *  ======== NameServerRemoteNotify_Instance_finalize ========
 */
Void NameServerRemoteNotify_Instance_finalize(NameServerRemoteNotify_Object *obj,
    Int status)
{
    Semaphore_Handle  semRemoteWait;
    Semaphore_Handle  semMultiBlock;
    Swi_Handle        swiHandle;

    if (status == 0) {
        /* unregister remote driver from NameServer module */
        NameServer_unregisterRemoteDriver(obj->remoteProcId);

        /* unregister event from Notify module */
        Notify_unregisterEventSingle(
                       obj->remoteProcId,
                       0,
                       obj->notifyEventId);
    }

    semRemoteWait = NameServerRemoteNotify_Instance_State_semRemoteWait(obj);
    if (semRemoteWait != NULL) {
        Semaphore_destruct(Semaphore_struct(semRemoteWait));
    }

    semMultiBlock = NameServerRemoteNotify_Instance_State_semMultiBlock(obj);
    if (semMultiBlock != NULL) {
        Semaphore_destruct(Semaphore_struct(semMultiBlock));
    }

    swiHandle = NameServerRemoteNotify_Instance_State_swiObj(obj);
    if (swiHandle != NULL) {
        Swi_destruct(Swi_struct(swiHandle));
    }
}

/*
 *************************************************************************
 *                       Module functions
 *************************************************************************
 */

/*
 *  ======== NameServerRemoteNotify_cbFxn ========
 */
Void NameServerRemoteNotify_cbFxn(UInt16 procId,
                                  UInt16 lineId,
                                  UInt32 eventId, 
                                  UArg arg, 
                                  UInt32 payload)
{
    Swi_Handle swiHandle;
        
    /* Swi_Handle was passed as arg in register */
    swiHandle = (Swi_Handle)arg;

    /* post the Swi */
    Swi_post(swiHandle);
}

/*
 *  ======== NameServerRemoteNotify_get ========
 */
Int NameServerRemoteNotify_get(NameServerRemoteNotify_Object *obj,
                               String instanceName, 
                               String name, 
                               Ptr value, 
                               UInt32 *valueLen, 
                               ISync_Handle syncHandle, 
                               Error_Block *eb)    
{
    Int len;
    Int retval = NameServer_E_NOTFOUND;
    Int offset = 0;
    Int status;
    Int notifyStatus;
    IArg key;
    Semaphore_Handle semRemoteWait;
    Semaphore_Handle semMultiBlock;
    
    Assert_isTrue(*valueLen <= 300,
                      NameServerRemoteNotify_A_invalidValueLen);

    semRemoteWait = NameServerRemoteNotify_Instance_State_semRemoteWait(obj);
    semMultiBlock = NameServerRemoteNotify_Instance_State_semMultiBlock(obj);
    
    /* prevent multiple threads from entering */
    Semaphore_pend(semMultiBlock, BIOS_WAIT_FOREVER);
    
    if (MultiProc_self() > obj->remoteProcId) {
            offset = 1;
    }    

    if (obj->cacheEnable) {
        /* Make sure there's no outstanding message */
        Cache_inv(obj->msg[offset], sizeof(NameServerRemoteNotify_Message),
                  Cache_Type_ALL, TRUE);
    }

    /* enter gate so nobody can be modifying message in shared memory */
    key = GateMP_enter(obj->gate);

    /* this is a request message */
    obj->msg[offset]->request = 1;
    obj->msg[offset]->response = 0;
    obj->msg[offset]->requestStatus = 0;
    obj->msg[offset]->valueLen = *valueLen;

    /* copy the name of instance into shared memory */
    len = strlen(instanceName);
    strncpy((Char *)obj->msg[offset]->instanceName, instanceName, len + 1);

    /* copy the name of nameserver entry into shared memory */
    len = strlen(name);
    strncpy((Char *)obj->msg[offset]->name, name, len + 1);

    if (obj->cacheEnable) {
        Cache_wbInv(obj->msg[offset], sizeof(NameServerRemoteNotify_Message),
                    Cache_Type_ALL, TRUE);
    }

    /* send the notification to remote processor */
    notifyStatus = Notify_sendEvent(
                           obj->remoteProcId,
                       0,
                       obj->notifyEventId,
                       0,
                       FALSE);

    /* now we can leave the gate after we send the notification */
    GateMP_leave(obj->gate, key);
    
    if (notifyStatus < 0) {
        /* undo previous options */
        obj->msg[offset]->request = 0;
        obj->msg[offset]->valueLen = 0;

        /* post the semaphore to make sure it doesn't block */
        Semaphore_post(semMultiBlock);

        return (retval);
    }
    
    /* pend here until we get a notification back from remote processor */
    status = Semaphore_pend(semRemoteWait, BIOS_WAIT_FOREVER);
    
    if (status == FALSE) {
        retval = NameServer_E_OSFAILURE;
    }
    else {
        /* getting here means we got the notification back */
        /* enter gate so nobody can be modifying message in shared memory */
        key = GateMP_enter(obj->gate);

        if (obj->cacheEnable) {
            Cache_inv(obj->msg[offset], sizeof(NameServerRemoteNotify_Message),
                      Cache_Type_ALL, TRUE);
        }

        /* if successful request then copy to value */
        if (obj->msg[offset]->requestStatus == TRUE) {
            /* clear out requestStatus */
            obj->msg[offset]->requestStatus = FALSE;

            /* copy to value */
            if (obj->msg[offset]->valueLen == sizeof(UInt32)) {                
                memcpy(value, &(obj->msg[offset]->value), sizeof(UInt32));
            }
            else {
                memcpy(value, &(obj->msg[offset]->valueBuf),
                       obj->msg[offset]->valueLen); 
            }

            /* set length to amount of data that was copied */
            *valueLen = obj->msg[offset]->valueLen;

            /* set the status */
            retval = NameServer_S_SUCCESS;
        }
    
        /* clear out the request and response flags */
        obj->msg[offset]->request  = 0;
        obj->msg[offset]->response = 0;
    
        if (obj->cacheEnable) {
            Cache_wbInv(obj->msg[offset],
                        sizeof(NameServerRemoteNotify_Message),
                        Cache_Type_ALL, TRUE);
        }

        /* now we can leave the gate after we send the notification */
        GateMP_leave(obj->gate, key);
    }

    Semaphore_post(semMultiBlock);
    
    return (retval);
}

/*
 *  ======== NameServerRemoteNotify_sharedMemReq ========
 */
SizeT NameServerRemoteNotify_sharedMemReq(
    const NameServerRemoteNotify_Params *params)
{
    /*
     *  Two Message structs are required.
     *  One for sending request and one for sending response.
     */
    if (MultiProc_numProcessors > 1) {
        return (2 * sizeof(NameServerRemoteNotify_Message));
    }
    
    return (0);
}

/*
 *  ======== NameServerRemoteNotify_swiFxn ========
 */
Void NameServerRemoteNotify_swiFxn(UArg arg)
{
    Int count = NameServer_E_FAIL;
    Int offset = 0;
    UInt32 valueLen;
    IArg key;
    NameServer_Handle handle;
    NameServerRemoteNotify_Handle nsrHandle;
    Int status;

    nsrHandle = (NameServerRemoteNotify_Handle)arg;

    if (MultiProc_self() > nsrHandle->remoteProcId) {
            offset = 1;
    }
 
    if (nsrHandle->cacheEnable) {
        /* invalidate both request and response messages */
        Cache_inv(nsrHandle->msg[0],
                  sizeof(NameServerRemoteNotify_Message) << 1,
                  Cache_Type_ALL, TRUE);
    }

    /* in case of request */
    if (nsrHandle->msg[1 - offset]->request == TRUE) {
        /* get the NameServer handle */
        handle = NameServer_getHandle(
                     (String)nsrHandle->msg[1 - offset]->instanceName);
        valueLen = nsrHandle->msg[1 - offset]->valueLen;              

        if (handle != NULL) {
            /* Search for the NameServer entry */
            if (valueLen == sizeof(UInt32)) {
                count = NameServer_getLocalUInt32(handle, 
                    (String)nsrHandle->msg[1 - offset]->name, 
                    &nsrHandle->msg[1 - offset]->value);
            }
            else {
                count = NameServer_getLocal(handle, 
                    (String)nsrHandle->msg[1 - offset]->name, 
                    &nsrHandle->msg[1 - offset]->valueBuf, &valueLen);
            }
        }

        /* enter gate so nobody can be modifying message in shared memory */
            key = GateMP_enter(nsrHandle->gate);
        
        /*
         *  If an entry was found, set requestStatus to TRUE
         *  and valueLen to the size of data that was copied.
         */
        if (count == NameServer_S_SUCCESS) {
            nsrHandle->msg[1 - offset]->requestStatus = TRUE;
            nsrHandle->msg[1 - offset]->valueLen = valueLen;
        }
        
        /* Send a response back */
        nsrHandle->msg[1 - offset]->response = TRUE;
        nsrHandle->msg[1 - offset]->request = FALSE;

        if (nsrHandle->cacheEnable) {
            Cache_wbInv(nsrHandle->msg[1 - offset],
                        sizeof(NameServerRemoteNotify_Message),
                        Cache_Type_ALL, TRUE);
        }

        /* now we can leave the gate */
        GateMP_leave(nsrHandle->gate, key);
 
        /*
         *  The NotifyDriver handle must exists at this point,
         *  otherwise the Notify_sendEvent should have failed
         */
        status = Notify_sendEvent(
                nsrHandle->remoteProcId,
                 0,
                 nsrHandle->notifyEventId,
                 0,
                 FALSE);
        
        /* The NS query could fail, but the reply should never fail */
        Assert_isTrue(status >= 0, Ipc_A_internal);
    }
    
    /* in case of response */
    if (nsrHandle->msg[offset]->response == TRUE) {
        /* post the Semaphore */
        Semaphore_post(
            NameServerRemoteNotify_Instance_State_semRemoteWait(nsrHandle));
    }
}

/*
 *  ======== NameServerRemoteNotify_attach ========
 */
Int NameServerRemoteNotify_attach(UInt16 remoteProcId, Ptr sharedAddr)
{
    NameServerRemoteNotify_Params nsrParams;
    NameServerRemoteNotify_Handle handle;
    Int status = NameServerRemoteNotify_E_FAIL;

    /* Assert that the default GateMP is not NULL */
    Assert_isTrue(GateMP_getDefaultRemote() != NULL, Ipc_A_internal);
    
    /* Use default GateMP */
    NameServerRemoteNotify_Params_init(&nsrParams);
    nsrParams.gate = GateMP_getDefaultRemote();
        
    /* determine the address */
    nsrParams.sharedAddr = (Ptr)((UInt32)sharedAddr);
        
    /* create only if notify driver has been created to remote proc */
    if (Notify_intLineRegistered(remoteProcId, 0)) {
        handle = NameServerRemoteNotify_create(remoteProcId,
                                               &nsrParams,
                                               NULL);
        if (handle != NULL) {
            status = NameServerRemoteNotify_S_SUCCESS;
        }
    }

    return (status);
}

/*
 *  ======== NameServerRemoteNotify_detach ========
 */
Int NameServerRemoteNotify_detach(UInt16 remoteProcId)
{
    NameServerRemoteNotify_Handle handle;
    Int status = NameServerRemoteNotify_S_SUCCESS;

    handle = NameServerRemoteNotify_getHandle(remoteProcId);

    if (handle == NULL) {
        status = NameServerRemoteNotify_E_FAIL;
    }
    else {
        NameServerRemoteNotify_delete(&handle);
    }

    return (status);
}

/*
 *  ======== NameServerRemoteNotify_getHandle ========
 */
NameServerRemoteNotify_Handle NameServerRemoteNotify_getHandle(
    UInt16 remoteProcId)
{
    NameServerRemoteNotify_Handle obj;

    for (obj = NameServerRemoteNotify_Object_first(); obj != NULL; 
        obj = NameServerRemoteNotify_Object_next(obj)) {
        if (obj->remoteProcId == remoteProcId) {
            break;
        }
    }

    return (obj);
}

/*
 *                      Fix SDOCM00068400 (assert that NSRN reply succeeded)
 *! 26-Feb-2010 jv      Fix SDOCM00067180 - prevent multiple threads from
 *!                     calling get() at the same time.
 *! 18-Feb-2010 skp     Fixes for SDOCM00066913 (ROUND_UP issues)
 *! 22-Jan-2010 jv      Fixed SDOCM66115.
 *! 24-Aug-2009 jv      Fixed SDOCM61119.
 *! 20-Apr-2009 jv      Support values > UInt32.
 *! 24-Feb-2009 jv      Created.
 */     
/*
 *  @(#) ti.sdo.ipc; 1, 0, 0, 0,398; 4-9-2010 18:37:48; /db/vtree/library/trees/ipc/ipc-d23x/src/
 */

