/* 
 * Copyright (c) 2010
 * Texas Instruments
 *
 *  All rights reserved.  Property of Texas Instruments
 *  Restricted rights to use, duplicate or disclose this code are
 *  granted through contract.
 * 
 * */
/*
 *  ======== NameServerRemoteNotify.xdc ========
 *
 */
 
import xdc.runtime.Error;
import xdc.runtime.Assert;
import ti.sysbios.knl.Swi;
import ti.sysbios.ipc.Semaphore;
import ti.sdo.utils.INameServerRemote;

/*! @_nodoc
 *  ======== NameServerRemoteNotify ========
 *  Used by NameServer to communicate to remote processors.
 *
 *  This module is used by {@link #ti.sdo.utils.NameServer} to communicate
 *  to remote processors using {@link #ti.sdo.ipc.Notify} and shared memory.
 *  There needs to be one instance between each two cores in the system.
 *  Interrupts must be enabled before using this module.  For critical
 *  memory management, a GateMP {@link #gate} can be specified.  Currently
 *  supports transferring up to 300-bytes between two cores.
 */
@InstanceInitError
@InstanceFinalize

module NameServerRemoteNotify inherits INameServerRemote
{
    /* structure in shared memory for retrieving value */
    struct Message {
        Bits32  request;                /* if this is a request set to 1    */
        Bits32  response;               /* if this is a response set to 1   */
        Bits32  requestStatus;          /* if request sucessful set to 1    */
        Bits32  value;                  /* holds value if len <= 4          */
        Bits32  valueLen;               /* len of value                     */
        Bits32  instanceName[8];        /* name of NameServer instance      */
        Bits32  name[8];                /* name of NameServer entry         */
        Bits32  valueBuf[75];           /* supports up to 300-byte value    */
    };

    /*!
     *  Assert raised when length of value larger then 300 bytes.
     */
    config xdc.runtime.Assert.Id A_invalidValueLen =
        {msg: "A_invalidValueLen: Invalid valueLen (too large)"};

    /*!
     *  ======== notifyEventId ========
     *  The Notify event ID.
     */
    config UInt notifyEventId = 7;
    
    /*!
     *  ======== sharedMemReq ========
     *  Returns the shared memory size requirement for a single instance
     *
     *  Can be used to make sure the {@link #sharedAddr} buffer is large
     *  enough before calling create.
     *
     *  @param(params)       NameServerRemoteNotify create parameters
     *
     *  @b(returns)          Number of MAUs needed to create the instance.
     */
    SizeT sharedMemReq(const Params *params);
    
    /*! @_nodoc
     *  ======== attach ========
     */
    Int attach(UInt16 remoteProcId, Ptr sharedAddr);

    /*! @_nodoc
     *  ======== detach ========
     */
    Int detach(UInt16 remoteProcId);


instance:

    /*!
     *  ======== sharedAddr ========
     *  Physical address of the shared memory
     *
     *  The shared memory that will be used for maintaining shared state
     *  information.  This value must be the same for both processors when
     *  creating the instance between a pair of processors.
     */
    config Ptr sharedAddr = null;

    /*!
     *  ======== gate ========
     *  GateMP used for critical region management of the shared memory
     *
     *  Using the default value of NULL will result in the default GateMP
     *  being used for context protection.
     */
    config GateMP.Handle gate = null;

internal:    

    /*!
     *  ======== cbFxn ========
     *  The call back function registered with Notify.
     *
     *  This function is registered with Notify as a call back function
     *  when the specified event is triggered.  This function simply posts
     *  a Swi which will process the event.
     *
     *  @param(procId)          Source proc id
     *  @param(lineId)          Interrupt line id
     *  @param(eventId)         the Notify event id.
     *  @param(arg)             the argument for the function.
     *  @param(payload)         a 32-bit payload value.
     */
    Void cbFxn(UInt16 procId, UInt16 lineId, UInt32 eventId, UArg arg, 
               UInt32 payload);
    
    /*!
     *  ======== getHandle ========
     *  Returns the handle of the instance
     *
     *  @param(remoteProcId)    The remote processor's id.
     */
    Handle getHandle(UInt16 remoteProcId);
    
    /*!
     *  ======== swiFxn ========
     *  The swi function that will be executed during the call back.
     *
     *  @param(arg)     argument to swi function
     */
    Void swiFxn(UArg arg);
    
    /* instance state */
    struct Instance_State {
        Message            *msg[2];        /* Ptrs to messages in shared mem */
        UInt32             notifyEventId;  /* Notify event Id                */
        UInt16             regionId;       /* SharedRegion ID                */
        GateMP.Handle      gate;           /* remote and local gate protect  */
        Semaphore.Object   semRemoteWait;  /* sem to wait on remote proc     */
        Semaphore.Object   semMultiBlock;  /* sem to block multiple threads  */
        Swi.Object         swiObj;         /* instance swi object            */
        UInt16             remoteProcId;   /* remote MultiProc id            */
        Bool               cacheEnable;    /* cacheability                   */
    };
}
/*
 *  @(#) ti.sdo.ipc; 1, 0, 0, 0,398; 4-9-2010 18:37:48; /db/vtree/library/trees/ipc/ipc-d23x/src/
 */

