/* 
 * Copyright (c) 2010
 * Texas Instruments
 *
 *  All rights reserved.  Property of Texas Instruments
 *  Restricted rights to use, duplicate or disclose this code are
 *  granted through contract.
 *
 * */
/*
 *  ======== Notify.c ========
 */

#include <xdc/std.h>
#include <xdc/runtime/Assert.h>
#include <xdc/runtime/Gate.h>
#include <xdc/runtime/Memory.h>
#include <xdc/runtime/Startup.h>
#include <xdc/runtime/Error.h>

#include "package/internal/Notify.xdc.h"

#include <ti/sdo/ipc/interfaces/INotifyDriver.h>
#include <ti/sdo/ipc/Ipc.h>
#include <ti/sdo/utils/List.h>

#include <ti/sdo/ipc/_Notify.h>
#include <ti/sdo/utils/_MultiProc.h>

#define ISRESERVED(eventId) ((eventId & 0xFFFF) >= Notify_reservedEvents || \
                                ((eventId >> 16) == Notify_SYSTEMKEY))

/* Bit mask operations */
#define SET_BIT(num,pos)            ((num) |= (1u << (pos)))
#define CLEAR_BIT(num,pos)          ((num) &= ~(1u << (pos)))
#define TEST_BIT(num,pos)           ((num) & (1u << (pos)))

#ifdef __ti__
    #pragma FUNC_EXT_CALLED(Notify_attach);
    #pragma FUNC_EXT_CALLED(Notify_disable);
    #pragma FUNC_EXT_CALLED(Notify_disableEvent);
    #pragma FUNC_EXT_CALLED(Notify_enableEvent);
    #pragma FUNC_EXT_CALLED(Notify_eventAvailable);
    #pragma FUNC_EXT_CALLED(Notify_intLineRegistered);
    #pragma FUNC_EXT_CALLED(Notify_registerEvent);
    #pragma FUNC_EXT_CALLED(Notify_registerEventSingle);
    #pragma FUNC_EXT_CALLED(Notify_restore);
    #pragma FUNC_EXT_CALLED(Notify_sendEvent);
    #pragma FUNC_EXT_CALLED(Notify_sharedMemReq);
    #pragma FUNC_EXT_CALLED(Notify_unregisterEvent);
    #pragma FUNC_EXT_CALLED(Notify_unregisterEventSingle);
#endif

/*
 *  ======== Notify_Module_startup ========
 */
Int Notify_Module_startup(Int phase)
{
    /* Wait for Startup to be done because a user fxn may set MultiProc id */
    if (!Startup_Module_startupDone()) {
        return (Startup_NOTDONE);
    }
    /* Create a local instance */
    Notify_create(NULL, MultiProc_self(), 0, NULL, NULL);

    return (Startup_DONE);
}


/*
 *  ======== Notify_Instance_init ========
 */
Int Notify_Instance_init(Notify_Object *obj, INotifyDriver_Handle driverHandle,
                         UInt16 remoteProcId, UInt16 lineId,
                         const Notify_Params *params,
                         Error_Block *eb)
{
    UInt i;
    List_Handle eventList;

    Assert_isTrue(remoteProcId < MultiProc_numProcessors &&
                  lineId < Notify_numLines, Ipc_A_invArgument);
    Assert_isTrue(Notify_module->notifyHandles[remoteProcId][lineId] == NULL,
                  Notify_A_alreadyRegistered);

    obj->remoteProcId = remoteProcId;
    obj->lineId = lineId;
    obj->nesting = 0;

    /* Allocate and initialize (to 0 with calloc()) the callbacks array. */
    obj->callbacks = Memory_calloc(Notify_Object_heap(),
        (sizeof(Notify_EventCallback) * Notify_numEvents), 0, eb);
    if (obj->callbacks == NULL) {
        return (2);
    }

    obj->eventList = Memory_alloc(Notify_Object_heap(),
        sizeof(List_Struct) * Notify_numEvents, 0, eb);
    if (obj->eventList == NULL) {
        return (1);
    }

    for (i = 0; i < Notify_numEvents; i++) {
        eventList = List_Object_get(obj->eventList, i);

        /* Pass in NULL for eb since construct should never fail */
        List_construct(List_struct(eventList), NULL);
    }

    /* Used solely for remote driver (NULL if remoteProcId == self) */
    obj->driverHandle = driverHandle;

    if (driverHandle != NULL) {
        /* Send this handle to the INotifyDriver */
        INotifyDriver_setNotifyHandle(driverHandle, obj);
    }

    Notify_module->notifyHandles[remoteProcId][lineId] = obj;

    return (0);
}

/*
 *  ======== Notify_Instance_finalize ========
 */
Void Notify_Instance_finalize(Notify_Object *obj, int status)
{
    UInt i;

    switch (status) {
        case 0:
            /* Unregister the notify instance from the Notify module */
            Notify_module->notifyHandles[obj->remoteProcId][obj->lineId] = NULL;

            /* Destruct the event lists */
            for (i = 0; i < Notify_numEvents; i++) {
                List_destruct(List_struct(List_Object_get(obj->eventList, i)));
            }

            /* Free memory used for the List.Objects */
            Memory_free(Notify_Object_heap(), obj->eventList,
                sizeof(List_Struct) * Notify_numEvents);

        case 1:
            /* Free memory used for callbacks array */
            Memory_free(Notify_Object_heap(), obj->callbacks,
                sizeof(Notify_EventCallback) * Notify_numEvents);
    }
}

/*
 *  ======== Notify_attach ========
 *  Called within Ipc module or directly by the user
 */
Int Notify_attach(UInt16 remoteProcId, Ptr sharedAddr)
{
    Int status;

    /* Use the NotifySetup proxy to setup drivers */
    status = Notify_SetupProxy_attach(remoteProcId, sharedAddr);

    return (status);
}

/*
 *  ======== Notify_detach ========
 *  Called within Ipc module
 */
Int Notify_detach(UInt16 remoteProcId)
{
    Notify_Object *obj;
    INotifyDriver_Handle driverHandle;
    UInt i;

    for (i = 0; i < Notify_numLines; i++) {
        /* First, unregister the driver since it was registered last */
        obj = (Notify_Object *)Notify_module->notifyHandles[remoteProcId][i];
        
        /* Ensure that notify is actually registered */
        Assert_isTrue(obj != NULL, Notify_A_notRegistered);
        
        driverHandle = obj->driverHandle;
        Notify_delete(&obj);

        /* Second, delete the driver itself */
        INotifyDriver_delete(&driverHandle);
    }

    /* Notify_detech never fails on RTOS */
    return (Notify_S_SUCCESS);
}

/*
 *************************************************************************
 *                      Module functions
 *************************************************************************
 */

/*
 *  A Note about interrupt latency:
 *
 *  The next four functions below modify the registration of callback functions
 *  (registerEvent, unregisterEvent, registerEventMany, unregisterEventMany)
 *  Because the state of event registration is modified in these functions and
 *  the Notify_exec function (run within an ISR context) reads this state, we
 *  have to consider the possibility and outcome of these functions being
 *  interrupted by the Notify_exec function.  In order to ensure the integrity
 *  of the callback registration state during any call to Notify_exec, we
 *  eliminate the possiblity of preemption altogether by adhering to the
 *  following rules:
 *
 *  - If registering a single (or many) callback functions, modify state in the
 *    following order:
 *
 *    1.  Add to the event list (if registerMany)
 *    2.  Add the callback function (if registerMany for the first time or
 *        if doing a single register)
 *    3.  Finally enable the event in the notify driver, thus opening the gate
 *        for incoming interrupts.
 *
 *    Performing step 3) last during a register is crucial as it ensures that
 *    the remote processor never sends an interrupt that disrupts the operations
 *    performed in/between 1) and 2).
 *
 *  - If unregistering a single (or many) callback functions, perform the
 *    above steps in reverse order.  Doing so ensures that possibility of
 *    getting incoming interrupts is eliminated before local callback
 *    registration state is modified.
 *
 */

/*
 *  ======== Notify_registerEventSingle ========
 */
Int Notify_registerEventSingle(UInt16                 procId,
                               UInt16                 lineId,
                               UInt32                 eventId,
                               Notify_FnNotifyCbck    fnNotifyCbck,
                               UArg                   cbckArg)
{
    UInt32        strippedEventId = (eventId & 0xFFFF);
    IArg          modKey;
    Int           status;
    Notify_Object *obj;

    Assert_isTrue(procId < MultiProc_numProcessors && lineId < Notify_numLines,
                  Ipc_A_invArgument);
    Assert_isTrue(strippedEventId < Notify_numEvents, Ipc_A_invArgument);
    Assert_isTrue(ISRESERVED(eventId), Notify_A_reservedEvent);

    modKey = Gate_enterModule();

    obj = (Notify_Object *)Notify_module->notifyHandles[procId][lineId];

    Assert_isTrue(obj != NULL, Notify_A_notRegistered);

    if (obj->callbacks[strippedEventId].fnNotifyCbck) {
        /* A callback is already registered. Fail. */
        status = Notify_E_ALREADYEXISTS;
    }
    else {
        /*
         *  No callback is registered. Register it. There is no need to protect
         *  these modifications because the event isn't registered yet.
         */
        obj->callbacks[strippedEventId].fnNotifyCbck = (Fxn)fnNotifyCbck;
        obj->callbacks[strippedEventId].cbckArg = cbckArg;

        if (procId != MultiProc_self()) {
            /* Tell the remote notify driver that the event is now registered */
            INotifyDriver_registerEvent(obj->driverHandle, strippedEventId);
        }

        status = Notify_S_SUCCESS;
    }

    Gate_leaveModule(modKey);

    return (status);
}

/*
 *  ======== Notify_registerEvent ========
 */
Int Notify_registerEvent(UInt16                 procId,
                         UInt16                 lineId,
                         UInt32                 eventId,
                         Notify_FnNotifyCbck    fnNotifyCbck,
                         UArg                   cbckArg)
{
    UInt32               strippedEventId = (eventId & 0xFFFF);
    Int                  status;
    Notify_Object        *obj;
    IArg                 modKey;
    List_Handle          eventList;
    Notify_EventListener *listener;
    Bool                 listWasEmpty;
    Error_Block          eb;

    Assert_isTrue(procId < MultiProc_numProcessors && lineId < Notify_numLines,
                  Ipc_A_invArgument);
    Assert_isTrue(strippedEventId < Notify_numEvents, Ipc_A_invArgument);
    Assert_isTrue(ISRESERVED(eventId), Notify_A_reservedEvent);

    Error_init(&eb);
    
    modKey = Gate_enterModule();

    obj = (Notify_Object *)Notify_module->notifyHandles[procId][lineId];

    Assert_isTrue(obj != NULL, Notify_A_notRegistered);

    /* Allocate a new EventListener */
    listener = Memory_alloc(Notify_Object_heap(), sizeof(Notify_EventListener),
        0, &eb);
    if (listener == NULL) {
        /* Listener memory allocation failed.  Leave module gate & return */
        Gate_leaveModule(modKey);
        
        return (Notify_E_MEMORY);
    }

    listener->callback.fnNotifyCbck = (Fxn)fnNotifyCbck;
    listener->callback.cbckArg = cbckArg;

    eventList = List_Object_get(obj->eventList, strippedEventId);

    /*
     *  Store whether the list was empty so we know whether to register the
     *  callback
     */
    listWasEmpty = List_empty(eventList);

    /*
     *  Need to atomically add to the list using the system gate because
     *  Notify_exec might preempt List_remove.  List_put is atomic.
     */
    List_put(eventList, (List_Elem *)listener);

    if (listWasEmpty) {
        /*
         *  Registering this event for the first time.  Need to register the
         *  callback function.
         */
        status = Notify_registerEventSingle(procId, lineId, eventId,
            Notify_execMany, (UArg)obj);
            
        /* Notify_registerEventSingle should always succeed */
        Assert_isTrue(status == Notify_S_SUCCESS, Ipc_A_internal);
    }

    status = Notify_S_SUCCESS;

    Gate_leaveModule(modKey);

    return (status);
}

/*
 *  ======== Notify_unregisterEventSingle ========
 */
Int Notify_unregisterEventSingle(UInt16 procId, UInt16 lineId, UInt32 eventId)
{
    UInt32               strippedEventId = (eventId & 0xFFFF);
    Int                  status;
    Notify_Object        *obj;
    IArg                 modKey;

    Assert_isTrue(procId < MultiProc_numProcessors && lineId < Notify_numLines,
                  Ipc_A_invArgument);
    Assert_isTrue(strippedEventId < Notify_numEvents, Ipc_A_invArgument);
    Assert_isTrue(ISRESERVED(eventId), Notify_A_reservedEvent);

    modKey = Gate_enterModule();

    obj = (Notify_Object *)Notify_module->notifyHandles[procId][lineId];

    Assert_isTrue(obj != NULL, Notify_A_notRegistered);

    if (obj->callbacks[strippedEventId].fnNotifyCbck) {
        if (procId != MultiProc_self()) {
            /*
             *  First, Tell the remote notify driver that the event is now
             *  unregistered
             */
            INotifyDriver_unregisterEvent(obj->driverHandle,
                                          strippedEventId);
        }

        /*
         *  No need to protect these modifications with the system gate because
         *  we shouldn't get preempted by Notify_exec after INotifyDriver_
         *  unregisterEvent.
         */
        obj->callbacks[strippedEventId].fnNotifyCbck = NULL;
        obj->callbacks[strippedEventId].cbckArg = NULL;

        status = Notify_S_SUCCESS;
    }
    else {
        /* No callback is registered. Fail. */
        status = Notify_E_FAIL;
    }

    Gate_leaveModule(modKey);

    return (status);
}

/*
 *  ======== Notify_unregisterEvent ========
 */
Int Notify_unregisterEvent(UInt16                 procId,
                           UInt16                 lineId,
                           UInt32                 eventId,
                           Notify_FnNotifyCbck    fnNotifyCbck,
                           UArg                   cbckArg)
{
    UInt32               strippedEventId = (eventId & 0xFFFF);
    Int                  status;
    IArg                 sysKey, modKey;
    Notify_Object        *obj;
    List_Handle          eventList;
    Notify_EventListener *listener;
    UInt                 count = 0;

    Assert_isTrue(procId < MultiProc_numProcessors && lineId < Notify_numLines,
                  Ipc_A_invArgument);
    Assert_isTrue(strippedEventId < Notify_numEvents, Ipc_A_invArgument);
    Assert_isTrue(ISRESERVED(eventId), Notify_A_reservedEvent);

    modKey = Gate_enterModule();

    obj = (Notify_Object *)Notify_module->notifyHandles[procId][lineId];

    Assert_isTrue(obj != NULL, Notify_A_notRegistered);

    eventList = List_Object_get(obj->eventList, strippedEventId);

    if (List_empty(eventList)) {
        return (Notify_E_NOTFOUND);
    }

    /* Get the first listener on the list */
    listener = (Notify_EventListener *)List_next(eventList, NULL);

    while (listener != NULL) {
        count++;
        if (listener->callback.fnNotifyCbck == (Fxn)fnNotifyCbck &&
                listener->callback.cbckArg == cbckArg ) {
            break;      /* found a match! */
        }
        listener = (Notify_EventListener *)
            List_next(eventList, (List_Elem *)listener);
    }

    if (listener == NULL) {
        /* Event listener not found */
        status = Notify_E_NOTFOUND;
    }
    else {
        if (count == 1 && List_next(eventList, (List_Elem *)listener) == NULL) {
            /*
             *  If only one element counted so far and the List_next returns
             *  NULL, the list will be empty after unregistering.  Therefore,
             *  unregister the callback function.
             */
            status = Notify_unregisterEventSingle(procId, lineId, eventId);
            /* unregisterEvent should always suceed */
            Assert_isTrue(status == Notify_S_SUCCESS, Ipc_A_internal);

            /*  No need to protect the list removal: the event's unregistered */
            List_remove(eventList, (List_Elem *)listener);
        }
        else {
            /*
             *  Need to atomically remove from the list using the system gate
             *  because Notify_exec might preempt List_remove (the event is
             *  still registered)
             */
            sysKey = Gate_enterSystem();
            List_remove(eventList, (List_Elem *)listener);
            Gate_leaveSystem(sysKey);
        }

        /* Free the memory alloc'ed for the event listener */
        Memory_free(Notify_Object_heap(), listener,
            sizeof(Notify_EventListener));

        status = Notify_S_SUCCESS;
    }

    Gate_leaveModule(modKey);

    return (status);
}

/*
 *  ======== Notify_sendEvent ========
 */
Int Notify_sendEvent(UInt16   procId,
                     UInt16   lineId,
                     UInt32   eventId,
                     UInt32   payload,
                     Bool     waitClear)
{
    UInt32        strippedEventId = (eventId & 0xFFFF);
    Int           status;
    Notify_Object *obj;
    IArg          sysKey;

    Assert_isTrue(procId < MultiProc_numProcessors && lineId < Notify_numLines,
                  Ipc_A_invArgument);
    Assert_isTrue(strippedEventId < Notify_numEvents, Ipc_A_invArgument);
    Assert_isTrue(ISRESERVED(eventId), Notify_A_reservedEvent);

    obj = (Notify_Object *)Notify_module->notifyHandles[procId][lineId];

    Assert_isTrue(obj != NULL, Notify_A_notRegistered);

    if (procId != MultiProc_self()) {
        /* Send a remote event */
        status = INotifyDriver_sendEvent(obj->driverHandle,
                                         strippedEventId,
                                         payload, waitClear);
    }
    else {
        /*  
         *  The check agaist non-NULL fnNotifyCbCk must be atomic with 
         *  Notify_exec so Notify_exec doesn't encounter a null callback.
         */
        sysKey = Gate_enterSystem();
        
        /*
         *  If nesting == 0 (the driver is enabled) and the event is enabled,
         *  send the event
         */
        if (obj->callbacks[eventId].fnNotifyCbck == NULL) {
            /* No callbacks are registered locally for the event. */
            status = Notify_E_EVTNOTREGISTERED;
        }
        else if (obj->nesting != 0) {
            /* Driver is disabled */
            status = Notify_E_FAIL;
        }
        else if (!TEST_BIT (Notify_module->localEnableMask, strippedEventId)){
            /* Event is disabled */
            status = Notify_E_EVTDISABLED;
        }
        else {
            /* Execute the callback function registered to the event */
            Notify_exec((Notify_Handle)obj, eventId, payload);
        }
        
        Gate_leaveSystem(sysKey);
    }

    return (status);
}

/*
 *  ======== Notify_disable ========
 */
UInt Notify_disable(UInt16 procId, UInt16 lineId)
{
    UInt key;
    IArg modKey;
    Notify_Object *obj;

    Assert_isTrue(procId < MultiProc_numProcessors && lineId < Notify_numLines,
                  Ipc_A_invArgument);

    obj = (Notify_Object *)Notify_module->notifyHandles[procId][lineId];

    Assert_isTrue(obj != NULL, Notify_A_notRegistered);

    /* Nesting value and enable state have to be in sync */
    modKey = Gate_enterModule();

    obj->nesting++;
    if (obj->nesting == 1) {
        /* Disable receiving all events */
        if (procId != MultiProc_self()) {
            INotifyDriver_disable(obj->driverHandle);
        }
    }

    key = obj->nesting;

    Gate_leaveModule(modKey);

    return (key);
}

/*
 *  ======== Notify_restore ========
 */
Void Notify_restore(UInt16 procId, UInt16 lineId, UInt key)
{
    IArg modKey;
    Notify_Object *obj;

    Assert_isTrue(procId < MultiProc_numProcessors && lineId < Notify_numLines,
                  Ipc_A_invArgument);
    
    obj = (Notify_Object *)Notify_module->notifyHandles[procId][lineId];

    Assert_isTrue(obj != NULL, Notify_A_notRegistered);
    Assert_isTrue(key == obj->nesting, Notify_A_outOfOrderNesting);

    /* Nesting value and enable state have to be in sync */
    modKey = Gate_enterModule();

    obj->nesting--;
    if (obj->nesting == 0) {
        /* Enable receiving events */
        if (procId != MultiProc_self()) {
            INotifyDriver_enable(obj->driverHandle);
        }
    }

    Gate_leaveModule(modKey);
}

/*
 *  ======== Notify_disableEvent ========
 */
Void Notify_disableEvent(UInt16 procId, UInt16 lineId, UInt32 eventId)
{
    UInt32 strippedEventId = (eventId & 0xFFFF);
    Notify_Object *obj;
    IArg sysKey;

    Assert_isTrue(procId < MultiProc_numProcessors && lineId < Notify_numLines,
                  Ipc_A_invArgument);
    Assert_isTrue(strippedEventId < Notify_numEvents, Ipc_A_invArgument);
    Assert_isTrue(ISRESERVED(eventId), Notify_A_reservedEvent);

    obj = (Notify_Object *)Notify_module->notifyHandles[procId][lineId];

    Assert_isTrue(obj != NULL, Notify_A_notRegistered);

    if (procId != MultiProc_self()) {
        /* disable the remote event */
        INotifyDriver_disableEvent(obj->driverHandle, strippedEventId);
    }
    else {
        /* disable the local event */
        sysKey = Gate_enterSystem();
        CLEAR_BIT(Notify_module->localEnableMask, strippedEventId);
        Gate_leaveSystem(sysKey);
    }
}

/*
 *  ======== Notify_enableEvent ========
 */
Void Notify_enableEvent(UInt16 procId, UInt16 lineId, UInt32 eventId)
{
    UInt32 strippedEventId = (eventId & 0xFFFF);
    Notify_Object *obj;
    IArg sysKey;

    Assert_isTrue(procId < MultiProc_numProcessors && lineId < Notify_numLines,
                  Ipc_A_invArgument);
    Assert_isTrue(strippedEventId < Notify_numEvents, Ipc_A_invArgument);
    Assert_isTrue(ISRESERVED(eventId), Notify_A_reservedEvent);

    obj = (Notify_Object *)Notify_module->notifyHandles[procId][lineId];

    Assert_isTrue(obj != NULL, Notify_A_notRegistered);

    if (procId != MultiProc_self()) {
        /* enable the remote event */
        INotifyDriver_enableEvent(obj->driverHandle, strippedEventId);
    }
    else {
        /* enable the local event */
        sysKey = Gate_enterSystem();
        SET_BIT(Notify_module->localEnableMask, strippedEventId);
        Gate_leaveSystem(sysKey);
    }
}

/*
 *  ======== Notify_eventAvailable ========
 */
Bool Notify_eventAvailable(UInt16 procId, UInt16 lineId, UInt32 eventId)
{
    UInt32 strippedEventId = (eventId & 0xFFFF);
    Bool available;
    Notify_Object *obj;

    Assert_isTrue(procId < MultiProc_numProcessors && lineId < Notify_numLines,
                  Ipc_A_invArgument);
    Assert_isTrue(strippedEventId < Notify_numEvents, Ipc_A_invArgument);

    obj = (Notify_Object *)Notify_module->notifyHandles[procId][lineId];
    if (obj == NULL) {
        /* Driver not registered */
        return (FALSE);
    }
    
    if (!ISRESERVED(eventId)) {
        /* Event is reserved and the caller isn't using Notify_SYSTEMKEY */
        return (FALSE);
    }
    
    available = (obj->callbacks[strippedEventId].fnNotifyCbck == NULL);

    return (available);
}

/*
 *  ======== Notify_intLineRegistered ========
 */
Bool Notify_intLineRegistered(UInt16 procId, UInt16 lineId)
{
    Bool registered;

    Assert_isTrue(procId < MultiProc_numProcessors && lineId < Notify_numLines,
                  Ipc_A_invArgument);

    registered = (Notify_module->notifyHandles[procId][lineId] != NULL);

    return (registered);
}

/*
 *  ======== Notify_sharedMemReq ========
 */
SizeT Notify_sharedMemReq(UInt16 procId, Ptr sharedAddr)
{
    SizeT memReq;

    if (procId == MultiProc_self()) {
        return (0);
    }

    /* Determine device-specific shared memory requirements */
    memReq = Notify_SetupProxy_sharedMemReq(procId, sharedAddr);

    return (memReq);
}

/*
 *  ======== Notify_exec ========
 */
Void Notify_exec(Notify_Object *obj, UInt32 eventId, UInt32 payload)
{
    Notify_EventCallback *callback;

    callback = &(obj->callbacks[eventId]);

    /* Make sure the callback function is not null */
    Assert_isTrue(callback->fnNotifyCbck != NULL, Ipc_A_internal);

    /* Execute the callback function with its argument and the payload */
    callback->fnNotifyCbck(obj->remoteProcId, obj->lineId, eventId,
        callback->cbckArg, payload);
}


/*
 *************************************************************************
 *                       Internal functions
 *************************************************************************
 */

/*
 *  ======== Notify_execMany ========
 */
Void Notify_execMany(UInt16 procId, UInt16 lineId, UInt32 eventId,
                     UArg arg, UInt32 payload)
{
    Notify_Object *obj = (Notify_Object *)arg;
    List_Handle eventList;
    Notify_EventListener *listener;

    /* Both loopback and the the event itself are enabled */
    eventList = List_Object_get(obj->eventList, eventId);

    /* Use "NULL" to get the first EventListener on the list */
    listener = (Notify_EventListener *)List_next(eventList, NULL);
    while (listener != NULL) {
        /* Execute the callback function */
        listener->callback.fnNotifyCbck(procId, lineId,
            eventId, listener->callback.cbckArg, payload);

        listener = (Notify_EventListener *)List_next(eventList,
             (List_Elem *)listener);
    }
}

/*
 *                      Added procId/lineId Asserts to [un]registerEvent[Single]
 *! 24-Mar-2010 skp     Fix SDOCM00068331 (sendEvent shouldn't enter mod gate)
 *! 04-Mar-2010 skp     Fix SDOCM00067532 (added Notify_eventAvailable)
 *! 19-Feb-2010 skp     Added FUNC_EXT_CALLED #pragma's
 *! 21-Jan-2010 skp     Fix SDOCM00066064 (misc notify bugs)
 *! 01-Dec-2009 skp     Loopback instance created in Module_startup
 *! 19-Nov-2009 skp     fixed high interrupt latency issues
 */
/*
 *  @(#) ti.sdo.ipc; 1, 0, 0, 0,398; 4-9-2010 18:37:48; /db/vtree/library/trees/ipc/ipc-d23x/src/
 */

