/* 
 * Copyright (c) 2010
 * Texas Instruments
 *
 *  All rights reserved.  Property of Texas Instruments
 *  Restricted rights to use, duplicate or disclose this code are
 *  granted through contract.
 *
 * */
/*
 *  ======== Notify.xs ========
 *
 */

var Notify      = null;
var MultiProc   = null;
var Memory      = null;
var Ipc         = null;
var List        = null;
var Settings    = null;

/*
 *  ======== module$use ========
 */
function module$use()
{
    Notify      = this;
    Memory      = xdc.useModule('xdc.runtime.Memory');
    MultiProc   = xdc.useModule('ti.sdo.utils.MultiProc');
    List        = xdc.useModule('ti.sdo.utils.List');
    Ipc         = xdc.useModule('ti.sdo.ipc.Ipc');
    Settings    = xdc.useModule('ti.sdo.ipc.family.Settings');

    /* 
     *  Check for proper MultiProc.nameList if it exists.  I.e. if building on
     *  evmDM6446, the only possible names in nameList are "ARM" and "DSP".
     */
    for each (var name in MultiProc.nameList) {
        if (name != null && !Settings.procInDevice(name)) {
            Notify.$logFatal("The MultiProc name (" + name + ") does not " +
                "match any of the following possible names on the device (" +
                Program.cpu.deviceName + "): " + Settings.getDeviceProcNames(),
                Notify);
        }
    }
    
    /* Check for valid numEvents */
    if (Notify.numEvents > Notify.MAXEVENTS) {
        Notify.$logFatal("Notify.numEvents may not be set to a value greater" +
                          " than " + Notify.MAXEVENTS, Notify);
    }
    
    /* 
     *  Plug the module gate which will be used for protecting Notify APIs from
     *  each other. Use GateSwi 
     */
    if (Notify.common$.gate === undefined) {
        var GateSwi = xdc.useModule('ti.sysbios.gates.GateSwi');
        Notify.common$.gate = GateSwi.create();
    }

    /* Plug the NotifySetup proxy */
    if (Notify.SetupProxy == null) {
        if (MultiProc.numProcessors == 1) {
            /* 
             *  On a uniprocessor system (or Notify is disabled). We will never 
             *  need to setup remote notify driver instances, so plug in 
             *  NotifySetupNull.
             */
            Notify.SetupProxy 
                = xdc.useModule('ti.sdo.ipc.notifyDrivers.NotifySetupNull');
        }
        else {
            /* 
             *  On a multiprocessor system.  Query the Settings module for the
             *  device-specific NotifySetup delegate.
             */
            try {
                delegateName = Settings.getNotifySetupDelegate();
                Notify.SetupProxy = xdc.useModule(delegateName);
            }
            catch (e) {
                Notify.$logFatal(String(e), Notify);
            }
        }
    }

}

/*
 *  ======== module$static$init ========
 */
function module$static$init(mod, params)
{
    var handles = Notify.$object.notifyHandles;

    handles.length = MultiProc.numProcessors;

    /* Initialize every handle to null */
    for (var i = 0; i < handles.length; i++) {
        handles[i].length = params.numLines;

        for (var j = 0; j < handles[i].length; j++) {
            handles[i][j] = null;
        }
    }

    mod.localEnableMask = ~0;   /* all enabled by default */
}


/*
 *************************************************************************
 *                       ROV View functions
 *************************************************************************
 */

/*
 *  ======== viewInitBasic ========
 */
function viewInitBasic(view, obj)
{
    var MultiProc = xdc.useModule('ti.sdo.utils.MultiProc');

    /* view.remoteProcId */
    view.remoteProcId = obj.remoteProcId;
    
    /* view.remoteProcName */
    try {
        view.remoteProcName = MultiProc.getName$view(obj.remoteProcId);
    }
    catch(e) {
        Program.displayError(view, 'remoteProcName', 
                             "Problem retrieving proc name: " + e);
    }

    /* view.lineId */
    view.lineId = obj.lineId;

    /* view.disabled */
    view.disabled = obj.nesting;
}

/*
 *  ======== viewInitData ========
 *  Instance data view.
 */
function viewInitData(view, obj)
{
    var Notify  = xdc.useModule('ti.sdo.ipc.Notify');
    var List    = xdc.useModule('ti.sdo.utils.List');
    var modCfg = Program.getModuleConfig('ti.sdo.ipc.Notify');

    /* Display the instance label in the tree */
    view.label = "procId = " + obj.remoteProcId + " lineId = " + obj.lineId;

    /* Fetch the callback array */
    try {
        var callbacks = Program.fetchArray(obj.callbacks$fetchDesc,
                                           obj.callbacks,
                                           modCfg.numEvents);
    }
    catch(e) {
        Program.displayError(view, "eventId", "Problem retrieving callback " +
                             " array from instance state.");
        return;
    }

    /* Fetch the eventList */
    try {
        var eventList = Program.fetchArray(obj.eventList$fetchDesc,
                                           obj.eventList,
                                           modCfg.numEvents);
    }
    catch(e) {
        Program.displayError(view, "eventId", "Problem retrieving eventList " +
                             " from instance state.");
        return;
    }
    
    /* For each element of the buffer... */
    for (var eventId = 0; eventId < modCfg.numEvents; eventId++) {   
        var fnNotifyCbck = callbacks[eventId].fnNotifyCbck;
        if (Number(fnNotifyCbck) == 0) {
            /* Event is unplugged */
            continue;
        }
        
        var fxnName = Program.lookupFuncName(Number(fnNotifyCbck))[0];
        if (fxnName == "ti_sdo_ipc_Notify_execMany__I") {
            /* Multiple callbacks registered.  View all of them */
            try {
                listView = Program.scanObjectView('ti.sdo.utils.List',
                                                   eventList[eventId],
                                                  'Basic');
                }
            catch (e) {
                Program.displayError(view, "eventId", "Problem scanning ROV " +
                    " for embedded event list object");
                return;
            }
            for each (var listElemPtr in listView.elems) {
                try {
                    var eventListener = Program.fetchStruct(
                            Notify.EventListener$fetchDesc,
                            $addr(listElemPtr));
                }
                catch(e) {
                    Program.displayError(view, "eventId", "Problem " +
                        "retrieving event listener ");
                    return;
                }

                var elem = Program.newViewStruct('ti.sdo.ipc.Notify',
                                                 'EventListeners');
                elem.eventId = eventId;
                elem.fnNotifyCbck = Program.lookupFuncName(
                    Number(eventListener.callback.fnNotifyCbck))[0];
                elem.cbckArg = "0x" + 
                    Number(eventListener.callback.cbckArg).toString(16);

                /* Create a new row in the instance data view */
                view.elements.$add(elem);
            }   
        }
        else {
            /* Single callback function registered. Just view one. */        
            var elem = Program.newViewStruct('ti.sdo.ipc.Notify',
                                             'EventListeners');
            elem.fnNotifyCbck = fxnName;
            elem.eventId = eventId;
            elem.cbckArg = "0x" + 
                Number(callbacks[eventId].cbckArg).toString(16);
                
            /* Create a new row in the instance data view */
            view.elements.$add(elem);
        }
    }
}
/*
 *  @(#) ti.sdo.ipc; 1, 0, 0, 0,398; 4-9-2010 18:37:49; /db/vtree/library/trees/ipc/ipc-d23x/src/
 */

