/* 
 * Copyright (c) 2010
 * Texas Instruments
 *
 *  All rights reserved.  Property of Texas Instruments
 *  Restricted rights to use, duplicate or disclose this code are
 *  granted through contract.
 *
 * */
/*
 *  ======== notify_video.c ========
 *  Notify on a multiprocessor system
 *
 *  This is an example program that uses Notify between the three
 *  processors (VIDEO, DSS and DSP) of a DM740 device.
 * 
 *  The M3/VIDEO processor sends a notify event to
 *  the DSP. Once the DSP receives the notify, it sends a notify 
 *  event to the DSS processor. The DSS processor then sends it
 *  back to the VIDEO processor. The loop is completed NUMLOOPS times.
 *
 *  Each processor has a Task that is pending on a semaphore. Once
 *  the event comes in, the cbFxn posts the semaphore, which releases
 *  the pending Task.
 *
 *  See notify_video.k file for expected output.
 *
 *  Refer to the readme.txt on how to load and run the example.
 */

#include <xdc/std.h>

/*  -----------------------------------XDC.RUNTIME module Headers    */
#include <xdc/runtime/System.h>

/*  ----------------------------------- IPC module Headers           */
#include <ti/ipc/MultiProc.h>
#include <ti/ipc/Notify.h>
#include <ti/ipc/Ipc.h>

/*  ----------------------------------- BIOS6 module Headers         */
#include <ti/sysbios/ipc/Semaphore.h>
#include <ti/sysbios/BIOS.h>

/*  ----------------------------------- To get globals from .cfg Header */
#include <xdc/cfg/global.h>

/* Number of times to run the loop */
#define NUMLOOPS        10 

/* Notify event number that the app uses */
#define EVENTID         10

UInt32 seq = 0;
UInt16 dssProcId;
UInt16 dspProcId;

/*
 *  ======== cbFxn ========
 *  This function was registered with Notify. It is called when 
 *  an event is sent to this processor.
 */
Void cbFxn(UInt16 procId, UInt16 lineId,
           UInt32 eventId, UArg arg, UInt32 payload)
{
    /* Quick sanity check. This should never fail in this example */
    if ((arg != 0xBAAD) || 
        (eventId != EVENTID)) {
        System_abort("Wrong values\n");
    }

    /* store in seq to show that payload made it all the way around */
    seq = payload;
    Semaphore_post(semHandle);
}

/*
 *  ======== tsk1_func ========
 *  Send an interrupt to the DSP processor and wait on semaphore.
 */
Void tsk1_func(UArg arg0, UArg arg1)
{   
    Int   i = 1;
    Int status;    
    Int16 remoteProcId = MultiProc_getId("DSP");
    
    while (i <= NUMLOOPS) {
        /* Send an event to the DSP */
        status = Notify_sendEvent(dspProcId, 0, EVENTID, i, TRUE);

        /* Continue until remote side is up */      
        if (status < 0) {
            continue;
        }

        System_printf("tsk1_func: Sent request #%d to DSP\n", i);

        /* Wait to be released by the cbFxn posting the semaphore */
        Semaphore_pend(semHandle, BIOS_WAIT_FOREVER);
        
        System_printf("tsk1_func: Received request #%d from DSS\n", seq);        
        
        /* increment for next iteration */
        i++;
    }
    System_printf("Test completed\n");
    System_exit(0);
}

/*
 *  ======== main ========
 */
Int main(Int argc, Char* argv[])
{
    Int status;

    dspProcId = MultiProc_getId("DSP");
    dssProcId = MultiProc_getId("DSS");

    System_printf("main: MultiProc id = %d\n", MultiProc_self());

    /*  
     *  Ipc_start() calls Ipc_attach() to synchronize all remote processors
     *  because 'Ipc.procSync' is set to 'Ipc.ProcSync_ALL' in *.cfg
     */
    Ipc_start();

    /*
     *  Register cbFxn with Notify. It will be called when DSS
     *  sends event number EVENTID to instance.
     *  Passing in 0xBAAD as the arg just for validation.
     */
    status = Notify_registerEvent(dssProcId, 0, EVENTID,
                              (Notify_FnNotifyCbck)cbFxn, 0xBAAD);    

    if (status < 0) {
        System_abort("Notify_registerEvent failed\n");
    }

    BIOS_start();
    return (0);
}

/*
 */
/*
 *  @(#) ti.sdo.ipc.examples.multicore.evmDM740; 1, 0, 0, 0,95; 4-9-2010 18:39:01; /db/vtree/library/trees/ipc/ipc-d23x/src/
 */

