/* 
 * Copyright (c) 2010
 * Texas Instruments
 *
 *  All rights reserved.  Property of Texas Instruments
 *  Restricted rights to use, duplicate or disclose this code are
 *  granted through contract.
 *
 * */
/*
 *  ======== notify_single.c ========
 *  Single-image version of Notify example
 *
 *  See notify_single_[CORENUM].k file for expected output.
 *
 *  Refer to the readme.txt on how to load and run the example.
 */

#include <xdc/std.h>

/*  -----------------------------------XDC.RUNTIME module Headers    */
#include <xdc/runtime/System.h>

/*  ----------------------------------- IPC module Headers           */
#include <ti/ipc/MultiProc.h>
#include <ti/ipc/Notify.h>
#include <ti/ipc/Ipc.h>
/*  ----------------------------------- BIOS6 module Headers         */
#include <ti/sysbios/ipc/Semaphore.h>
#include <ti/sysbios/knl/Task.h>
#include <ti/sysbios/BIOS.h>

/*  ----------------------------------- To get globals from .cfg Header */
#include <xdc/cfg/global.h>

#define INTERRUPT_LINE  0

/* Notify event number that the app uses */
#define EVENTID         10

/* Number of times to run the loop */
#define NUMLOOPS        10 

UInt32 seq = 0;
UInt16 recvProcId;

UInt16 srcProc, dstProc;

/*
 *  ======== cbFxn ========
 *  This function was registered with Notify. It is called when any event is
 *  sent to this processor.
 */
Void cbFxn(UInt16 procId, UInt16 lineId,
           UInt32 eventId, UArg arg, UInt32 payload)
{
    /* The payload is a sequence number. */
    recvProcId = procId;
    seq = payload;
    Semaphore_post(semHandle);
}

/*
 *  ======== setMyId ========
 *  This function is run by xdc.runtime.Startup and is executed before main().
 *  The local MultiProc id is set based on the value read from the DNUM
 *  register.
 */
Void setMyId() 
{
    extern cregister volatile UInt DNUM;
    UInt16 procId;
    Char procName[5];

    /* Determine the procName from DNUM */
    System_sprintf(procName, "CORE%d", DNUM);

    /* Translate from a procName to a procId */
    procId = MultiProc_getId(procName);
    
    /* Set the local ID */
    MultiProc_setLocalId(procId);
}


/*
 *  ======== tsk0_func ========
 *  This function is executed only on CORE0.
 *  It sends an event to the next processor then pends on a semaphore.
 *  The semaphore is posted by the callback function.
 */
Void tsk0_func(UArg arg0, UArg arg1)
{
    Int i = 1;
    Int status;

    while (i <= NUMLOOPS) {
        /* Send an event to the next processor */
        status = Notify_sendEvent(dstProc, INTERRUPT_LINE, EVENTID, i, TRUE);

        /* Continue until remote side is up */
        if (status < 0) {
            continue;
        }

        System_printf("tsk1_func: Sent request #%d to %s\n", seq,
            MultiProc_getName(dstProc));

        /* Wait to be released by the cbFxn posting the semaphore */
        Semaphore_pend(semHandle, BIOS_WAIT_FOREVER);

        System_printf("tsk1_func: Received request #%d from %s\n", seq,
            MultiProc_getName(recvProcId));

        /* increment for next iteration */
        i++;
    }

    System_printf("Test completed\n");
    System_exit(0);
}


/*
 *  ======== tsk1_func ========
 *  This function is executed on every processor except on CORE0.
 *  Pend on a semaphore. The semaphore is posted by the callback function.
 *  Once the semaphore is posted, send an event to the next processor.
 */
Void tsk1_func(UArg arg0, UArg arg1)
{
    Int status;

    while (seq < NUMLOOPS) {
        /* wait forever on a semaphore, semaphore is posted in callback */
        Semaphore_pend(semHandle, BIOS_WAIT_FOREVER);

        System_printf("tsk1_func: Received request #%d from %s\n", seq,
            MultiProc_getName(recvProcId));

        /* Send an event to the next processor */
        status = Notify_sendEvent(dstProc, INTERRUPT_LINE, EVENTID, seq, TRUE);
        if (status < 0) {
            System_abort("sendEvent failed\n");
        }

        System_printf("tsk1_func: Sent request #%d to %s\n", seq,
            MultiProc_getName(dstProc));
    }

    System_printf("Test completed\n");
    System_exit(0);
}

/*
 *  ======== main ========
 *  Create a task.
 *  Synchronize all processors.
 *  Register an event with Notify.
 */
Int main(Int argc, Char* argv[])
{
    Int status;
    UInt numProcs = MultiProc_getNumProcessors();

    /*
     *  Determine which processors Notify will communicate with based on the
     *  local MultiProc id.  Also, create a processor-specific Task.
     */
    srcProc = ((MultiProc_self() - 1 + numProcs) % numProcs);
    dstProc = ((MultiProc_self() + 1) % numProcs);
    if (MultiProc_self() == 0) {
        Task_create(tsk0_func, NULL, NULL);
    }
    else {
        Task_create(tsk1_func, NULL, NULL);
    }

    System_printf("main: MultiProc id = %d\n", MultiProc_self());
    System_printf("main: MultiProc name = %s\n", 
        MultiProc_getName(MultiProc_self()));
    
    /*  
     *  Ipc_start() calls Ipc_attach() to synchronize all remote processors
     *  because 'Ipc.procSync' is set to 'Ipc.ProcSync_ALL' in *.cfg
     */
    Ipc_start();

    /*
     *  Register call back with Notify. It will be called when the processor
     *  with id = srcProc sends event number EVENTID to this processor.
     */
    status = Notify_registerEvent(srcProc, INTERRUPT_LINE, EVENTID,
                                  (Notify_FnNotifyCbck)cbFxn, NULL);
    if (status < 0) {
        System_abort("Notify_registerEvent failed\n");
    }

    BIOS_start();
    return (0);
}

/*
 */
/*
 *  @(#) ti.sdo.ipc.examples.multicore.simTCI6498; 1, 0, 0, 0,85; 4-9-2010 18:39:10; /db/vtree/library/trees/ipc/ipc-d23x/src/
 */

