/* 
 * Copyright (c) 2010
 * Texas Instruments
 *
 *  All rights reserved.  Property of Texas Instruments
 *  Restricted rights to use, duplicate or disclose this code are
 *  granted through contract.
 * 
 * */
/*
 *  ======== Interrupt.c ========
 *  c6472 based interrupt manager.
 */

#include <xdc/std.h>
#include <xdc/runtime/Assert.h>

#include <ti/sysbios/family/c64p/Hwi.h>

#include <ti/sdo/ipc/Ipc.h>
#include <ti/sdo/ipc/notifyDrivers/IInterrupt.h>

#include "package/internal/Interrupt.xdc.h"

/* shift value used for setting/identifying interrupt source */
#define SRCSx_SHIFT     4


/*
 *************************************************************************
 *                      Module functions
 *************************************************************************
 */


/*
 *  ======== Interrupt_intEnable ========
 *  Enable interrupt
 */     
Void Interrupt_intEnable(UInt16 remoteProcId, IInterrupt_IntInfo *intInfo)
{
    Hwi_enableInterrupt(intInfo->intVectorId);
}
        
/*
 *  ======== Interrupt_intDisable ========
 *  Disables interrupts
 */     
Void Interrupt_intDisable(UInt16 remoteProcId, IInterrupt_IntInfo *intInfo)
{
    Hwi_disableInterrupt(intInfo->intVectorId);
}

/*
 *  ======== Interrupt_intRegister ========
 *  Register ISR for remote processor interrupt
 */
Void Interrupt_intRegister(UInt16 remoteProcId, IInterrupt_IntInfo *intInfo,
                           Fxn func, UArg arg)
{
    UInt       key;
    Hwi_Params hwiAttrs;
    Interrupt_FxnTable *table;

    /* Disable global interrupts */
    key = Hwi_disable();

    /* setup the function table using the same Hwi int */
    table = &(Interrupt_module->fxnTable[remoteProcId]);
    table->func = func;
    table->arg  = arg;

    Interrupt_intClear(remoteProcId, intInfo);
    
    /* Make sure the interrupt only gets plugged once */
    Interrupt_module->numPlugged++;
    if (Interrupt_module->numPlugged == 1) {
        /* Register interrupt to remote processor */
        Hwi_Params_init(&hwiAttrs);
        hwiAttrs.maskSetting = Hwi_MaskingOption_SELF;
        hwiAttrs.arg         = arg;
        hwiAttrs.eventId     = Interrupt_INTERDSPINT;

        Hwi_create(intInfo->intVectorId,
            (Hwi_FuncPtr)Interrupt_intShmStub, &hwiAttrs, NULL);

        Hwi_enableInterrupt(intInfo->intVectorId);
    }

    /* Restore global interrupts */
    Hwi_restore(key);
}

/*
 *  ======== Interrupt_intUnregister ========
 */
Void Interrupt_intUnregister(UInt16 remoteProcId, IInterrupt_IntInfo *intInfo)
{
    Hwi_Handle hwiHandle;
    Interrupt_FxnTable *table;
   
    Interrupt_module->numPlugged--;
    if (Interrupt_module->numPlugged == 0) {
        /* No need to disable interrupt: Hwi_delete takes care of this */
        hwiHandle = Hwi_getHandle(intInfo->intVectorId);
        Hwi_delete(&hwiHandle);        
    }
    
    /* Unset the function table */
    table = &(Interrupt_module->fxnTable[remoteProcId]);
    table->func = NULL;
    table->arg  = 0;
}

/*
 *  ======== Interrupt_intSend ========
 *  Send interrupt to the remote processor
 */
Void Interrupt_intSend(UInt16 remoteProcId, IInterrupt_IntInfo *intInfo,
                       UArg arg)
{
    UInt32 val;
    extern volatile cregister Uns DNUM;
    volatile UInt32 *ipcgr = (volatile UInt32 *)Interrupt_IPCGR0;

    /*
     *  bit 0 is set to generate interrupt.
     *  bits 4-7 is set to specify the interrupt generation source.
     *  The convention is that bit 4 (SRCS0) is used for core 0,
     *  bit 5 (SRCS1) for core 1, etc... .
     */
    val = (1 << (DNUM + SRCSx_SHIFT)) | 1;

    ipcgr[remoteProcId] =  val;
}

/*
 *  ======== Interrupt_intSend ========
 *  Send interrupt to the remote processor
 */
Void Interrupt_intPost(UInt16 srcProcId, IInterrupt_IntInfo *intInfo,
                       UArg arg)
{
    UInt32 val;
    extern volatile cregister Uns DNUM;
    volatile UInt32 *ipcgr = (volatile UInt32 *)Interrupt_IPCGR0;

    /*
     *  bit 0 is set to generate interrupt.
     *  bits 4-7 is set to specify the interrupt generation source.
     *  The convention is that bit 4 (SRCS0) is used for core 0,
     *  bit 5 (SRCS1) for core 1, etc... .
     */
    val = (1 << (srcProcId + SRCSx_SHIFT)) | 1;

    ipcgr[DNUM] =  val;
}

/*
 *  ======== Interrupt_intClear ========
 *  Acknowledge interrupt by clearing the corresponding source bit.
 *  Does not clear the IFR bit by way of ICR write because that should
 *  only be done during init time.
 */
UInt Interrupt_intClear(UInt16 remoteProcId, IInterrupt_IntInfo *intInfo)
{
    extern volatile cregister Uns DNUM;
    volatile UInt32 *ipcar = (volatile UInt32 *)Interrupt_IPCAR0;

    ipcar[DNUM] =  (1 << (remoteProcId + SRCSx_SHIFT));

    return (0);
}

/*
 *************************************************************************
 *                      Internals functions
 *************************************************************************
 */

/*
 *  ======== Interrupt_intShmStub ========
 */
Void Interrupt_intShmStub(UArg arg)
{
    Int i;
    Interrupt_FxnTable *table;
    extern volatile cregister Uns DNUM;
    volatile UInt32 *ipcar = (volatile UInt32 *)Interrupt_IPCAR0;

    for (i = 0; i < Interrupt_numCores; i++) {
        if ((ipcar[DNUM]) & (1 << (i + SRCSx_SHIFT))) {
            table = &(Interrupt_module->fxnTable[i]);
            (table->func)(table->arg);
        }
    }
}

/*
 */
/*
 *  @(#) ti.sdo.ipc.family.c647x; 1, 0, 0, 0,31; 4-9-2010 18:37:58; /db/vtree/library/trees/ipc/ipc-d23x/src/
 */

