/* 
 * Copyright (c) 2010
 * Texas Instruments
 *
 *  All rights reserved.  Property of Texas Instruments
 *  Restricted rights to use, duplicate or disclose this code are
 *  granted through contract.
 * 
 * */
/*
 *  ======== InterruptArm.c ========
 *  Arm interrupt manager
 */

#include <xdc/std.h>
#include <xdc/runtime/Assert.h>
#include <ti/sysbios/family/arm/da830/Hwi.h>

#include <ti/sdo/ipc/Ipc.h>
#include <ti/sdo/ipc/notifyDrivers/IInterrupt.h>
#include "package/internal/InterruptArm.xdc.h"

/* register use to generate interrupt between cores */
#define CHIPSIGREG      0x01C14174

/* event ids associated with inter-core interrupts */
#define DSP2ARM_CHIPINT0        28
#define DSP2ARM_CHIPINT1        29
#define ARM2DSP_CHIPINT2        5
#define ARM2DSP_CHIPINT3        67

/*
 *************************************************************************
 *                      Module functions
 *************************************************************************
 */

/*!
 *  ======== InterruptArm_intEnable ========
 *  Enable GPP interrupt
 */     
Void InterruptArm_intEnable(UInt16 remoteProcId, IInterrupt_IntInfo *intInfo)
{
    Hwi_enableInterrupt(intInfo->localIntId);
}
        
/*!
 *  ======== InterruptArm_intDisable ========
 *  Disables GPP interrupt
 */     
Void InterruptArm_intDisable(UInt16 remoteProcId, IInterrupt_IntInfo *intInfo)
{
    Hwi_disableInterrupt(intInfo->localIntId);
}

/*!
 *  ======== InterruptArm_intRegister ========
 *  Register ISR for remote processor interrupt
 */
Void InterruptArm_intRegister(UInt16 remoteProcId, IInterrupt_IntInfo *intInfo,
                              Fxn func, UArg arg)
{
    Hwi_Params  hwiAttrs;
    UInt        key;
                  
    Assert_isTrue(intInfo->localIntId == DSP2ARM_CHIPINT0 || 
                  intInfo->localIntId == DSP2ARM_CHIPINT1,
                  Ipc_A_internal);
                  
    Assert_isTrue(intInfo->remoteIntId == ARM2DSP_CHIPINT2 || 
                  intInfo->remoteIntId == ARM2DSP_CHIPINT3,
                  Ipc_A_internal);

    /* Disable global interrupts */
    key = Hwi_disable();

    /* Register interrupt for communication between ARM and DSP */
    Hwi_Params_init(&hwiAttrs);
    hwiAttrs.maskSetting = Hwi_MaskingOption_SELF;
    hwiAttrs.arg         = arg;

    InterruptArm_intClear(remoteProcId, intInfo);
    Hwi_create(intInfo->localIntId,
               (Hwi_FuncPtr)func,
               &hwiAttrs,
               NULL);

    /* Restore global interrupts */
    Hwi_restore(key);

    Hwi_enableInterrupt(intInfo->localIntId);
}

/*!
 *  ======== InterruptArm_intUnregister ========
 *  Register ISR for remote processor interrupt
 */
Void InterruptArm_intUnregister(UInt16 remoteProcId, IInterrupt_IntInfo *intInfo)
{
    Hwi_Handle hwiHandle;
    
    /* Delete the Hwi and disable the corresponding interrupt */
    hwiHandle = Hwi_getHandle(intInfo->localIntId);
    Hwi_delete(&hwiHandle);
}
                              

/*!
 *  ======== InterruptArm_intSend ========
 *  Send interrupt to the remote processor
 */
Void InterruptArm_intSend(UInt16 remoteProcId, IInterrupt_IntInfo *intInfo, 
                          UArg arg)
{
    UInt16 intBitPos;
    volatile UInt32 *chipSigReg = (volatile UInt32 *)CHIPSIGREG;

    switch(intInfo->remoteIntId) {
        case ARM2DSP_CHIPINT2:
            intBitPos = 2;
            break;
        case ARM2DSP_CHIPINT3:
            intBitPos = 3;
            break;
    }

    chipSigReg[0] = (1 << intBitPos);
}

/*!
 *  ======== InterruptArm_intClear ========
 *  Clear interrupt 
 */
UInt InterruptArm_intClear(UInt16 remoteProcId, IInterrupt_IntInfo *intInfo)
{
    UInt16 statBitPos;
    volatile UInt32 *chipSigReg = (volatile UInt32 *)CHIPSIGREG;

    switch(intInfo->localIntId) {
        case DSP2ARM_CHIPINT0:
            statBitPos = 0;
            break;
        case DSP2ARM_CHIPINT1:
            statBitPos = 1;
            break;
    }

    chipSigReg[1] = (1 << statBitPos);

    return (0);
}

/*
 */
/*
 *  @(#) ti.sdo.ipc.family.da830; 1, 0, 0, 0,98; 4-9-2010 18:38:00; /db/vtree/library/trees/ipc/ipc-d23x/src/
 */

