/* 
 * Copyright (c) 2010
 * Texas Instruments
 *
 *  All rights reserved.  Property of Texas Instruments
 *  Restricted rights to use, duplicate or disclose this code are
 *  granted through contract.
 * 
 * */
/*
 *  ======== InterruptDsp.c ========
 *  Dsp interrupt manager.
 */

#include <xdc/std.h>
#include <xdc/runtime/Assert.h>
#include <ti/sysbios/family/c64p/Hwi.h>

#include <ti/sdo/ipc/Ipc.h>
#include <ti/sdo/ipc/notifyDrivers/IInterrupt.h>
#include "package/internal/InterruptDsp.xdc.h"

/* register use to generate interrupt between cores */
#define CHIPSIGREG      0x01C14174

/* event ids associated with inter-core interrupts */
#define DSP2ARM_CHIPINT0        28
#define DSP2ARM_CHIPINT1        29
#define ARM2DSP_CHIPINT2        5
#define ARM2DSP_CHIPINT3        67

/*
 *************************************************************************
 *                      Module functions
 *************************************************************************
 */

/*!
 *  ======== InterruptDsp_intEnable ========
 *  Enable GPP interrupt
 */     
Void InterruptDsp_intEnable(UInt16 remoteProcId, IInterrupt_IntInfo *intInfo)
{
    Hwi_enableInterrupt(intInfo->intVectorId);
}
        
/*!
 *  ======== InterruptDsp_intDisable ========
 *  Disables GPP interrupt
 */     
Void InterruptDsp_intDisable(UInt16 remoteProcId, IInterrupt_IntInfo *intInfo)
{
    Hwi_disableInterrupt(intInfo->intVectorId);
}

/*!
 *  ======== InterruptDsp_intRegister ========
 *  Register ISR for remote processor interrupt
 */
Void InterruptDsp_intRegister(UInt16 remoteProcId, IInterrupt_IntInfo *intInfo,
                              Fxn func, UArg arg)
{
    Hwi_Params  hwiAttrs;
    UInt        key;

    Assert_isTrue(intInfo->intVectorId <= 15, 
                  Ipc_A_internal);
                  
    Assert_isTrue(intInfo->localIntId == ARM2DSP_CHIPINT2 || 
                  intInfo->localIntId == ARM2DSP_CHIPINT3,
                  Ipc_A_internal);
                  
    Assert_isTrue(intInfo->remoteIntId == DSP2ARM_CHIPINT0 || 
                  intInfo->remoteIntId == DSP2ARM_CHIPINT1,
                  Ipc_A_internal);
    
    /* Disable global interrupts */
    key = Hwi_disable();

    InterruptDsp_intClear(remoteProcId, intInfo); //TODO

    /* Register interrupt for communication between ARM and DSP */
    Hwi_Params_init(&hwiAttrs);
    hwiAttrs.maskSetting = Hwi_MaskingOption_SELF;
    hwiAttrs.arg         = arg;
    hwiAttrs.eventId     = intInfo->localIntId;

    Hwi_create(intInfo->intVectorId,
               (Hwi_FuncPtr)func,
               &hwiAttrs,
               NULL);

    /* Restore global interrupts */
    Hwi_restore(key);

    /* enable the interrupt vector */
    Hwi_enableInterrupt(intInfo->intVectorId);
}

/*!
 *  ======== InterruptDsp_intUnregister ========
 */
Void InterruptDsp_intUnregister(UInt16 remoteProcId, 
                                IInterrupt_IntInfo *intInfo)
{
    Hwi_Handle  hwiHandle;

    /* Delete the Hwi (and disable the corresponding interrupt) */
    hwiHandle = Hwi_getHandle(intInfo->intVectorId);
    Hwi_delete(&hwiHandle);
}

/*!
 *  ======== InterruptDsp_intSend ========
 *  Send interrupt to the remote processor
 */
Void InterruptDsp_intSend(UInt16 remoteProcId, IInterrupt_IntInfo *intInfo, 
                          UArg arg)
{
    UInt32 intBitPos;
    volatile UInt32 *chipSigReg = (volatile UInt32 *)CHIPSIGREG;

    switch(intInfo->remoteIntId) {
        case DSP2ARM_CHIPINT0:
            intBitPos = 0;
            break;
        case DSP2ARM_CHIPINT1:
            intBitPos = 1;
            break;
    }

    chipSigReg[0] = (1 << intBitPos);
}

/*!
 *  ======== InterruptDsp_intClear ========
 *  Clear interrupt 
 */
UInt InterruptDsp_intClear(UInt16 remoteProcId, IInterrupt_IntInfo *intInfo)
{
    UInt32 statBitPos;
    volatile UInt32 *chipSigReg = (volatile UInt32 *)CHIPSIGREG;

    switch(intInfo->localIntId) {
        case ARM2DSP_CHIPINT2:
            statBitPos = 2;
            break;
        case ARM2DSP_CHIPINT3:
            statBitPos = 3;
            break;
    }

    chipSigReg[1] = (1 << statBitPos);

    return (0);
}

/*
 */
/*
 *  @(#) ti.sdo.ipc.family.da830; 1, 0, 0, 0,98; 4-9-2010 18:38:00; /db/vtree/library/trees/ipc/ipc-d23x/src/
 */

