/* 
 * Copyright (c) 2010
 * Texas Instruments
 *
 *  All rights reserved.  Property of Texas Instruments
 *  Restricted rights to use, duplicate or disclose this code are
 *  granted through contract.
 * 
 * */
/*
 *  ======== NotifySetup.c ========
 */
#include <xdc/std.h>
#include <xdc/runtime/Error.h> 

#include <ti/sdo/ipc/notifyDrivers/NotifyDriverShm.h>
#include <ti/sdo/ipc/_Notify.h>
#include <ti/sdo/utils/_MultiProc.h>

#include "package/internal/NotifySetup.xdc.h"

/*
 *************************************************************************
 *                      Module functions
 *************************************************************************
 */

/*!
 *  ======== NotifySetup_attach ========
 *  Initialize interrupt
 */     
Int NotifySetup_attach(UInt16 remoteProcId, Ptr sharedAddr)
{
    NotifyDriverShm_Params notifyShmParams;
    Error_Block eb;
    Int status = Notify_S_SUCCESS;
    
#ifdef xdc_target__isaCompatible_64
    NotifyDriverShm_Handle armDriverHandle0, armDriverHandle1;
    UInt armProcId = 1 - MultiProc_self();

    /* Initialize the error block */
    Error_init(&eb);

    /*
     *  Setup the notify driver to the ARM (Line 0)
     */     
    NotifyDriverShm_Params_init(&notifyShmParams);
    notifyShmParams.localIntId     = NotifySetup_dspRecvIntId0;
    notifyShmParams.remoteIntId    = NotifySetup_armRecvIntId0;
    notifyShmParams.intVectorId    = NotifySetup_dspIntVectId0;
    notifyShmParams.remoteProcId   = armProcId;
    notifyShmParams.sharedAddr     = sharedAddr;
    
    armDriverHandle0 = NotifyDriverShm_create(&notifyShmParams, &eb);
    if (armDriverHandle0 == NULL) {
        return (Notify_E_FAIL);
    }
    
    Notify_create(NotifyDriverShm_Handle_upCast(armDriverHandle0), 
                  armProcId,
                  0,
                  NULL,
                  &eb);
    if (Error_check(&eb)) {
        /* Delete the driver and then return */
        NotifyDriverShm_delete(&armDriverHandle0);
        return (Notify_E_FAIL);
    }

    if (!NotifySetup_useSecondLine) {
        return (status);
    }
    /*
     *  Setup the notify driver to the ARM (Line 1)
     */     
    NotifyDriverShm_Params_init(&notifyShmParams);
    notifyShmParams.localIntId     = NotifySetup_dspRecvIntId1;
    notifyShmParams.remoteIntId    = NotifySetup_armRecvIntId1;
    notifyShmParams.intVectorId    = NotifySetup_dspIntVectId1;
    notifyShmParams.remoteProcId   = armProcId;
    notifyShmParams.sharedAddr     = (Ptr)((UInt32)sharedAddr + 
        NotifyDriverShm_sharedMemReq(&notifyShmParams));
    armDriverHandle1 = NotifyDriverShm_create(&notifyShmParams, &eb);

    if (armDriverHandle1 == NULL) {
        return (Notify_E_FAIL);
    }

    Notify_create(NotifyDriverShm_Handle_upCast(armDriverHandle1), 
                  armProcId,
                  1,
                  NULL,
                  &eb);
    if (Error_check(&eb)) {
        /* Delete the driver and then return */
        NotifyDriverShm_delete(&armDriverHandle1);
        return (Notify_E_FAIL);
    }

#else /* ARM code */
    NotifyDriverShm_Handle dspDriverHandle0, dspDriverHandle1;
    UInt dspProcId =  1 - MultiProc_self();
    
    /* Initialize the error block */
    Error_init(&eb);
    
    /*
     *  Setup the notify driver to the DSP (Line 0)
     */     
    NotifyDriverShm_Params_init(&notifyShmParams);
    notifyShmParams.localIntId     = NotifySetup_armRecvIntId0;
    notifyShmParams.remoteIntId    = NotifySetup_dspRecvIntId0;
    notifyShmParams.remoteProcId   = dspProcId;
    notifyShmParams.sharedAddr     = sharedAddr;
    
    dspDriverHandle0 = NotifyDriverShm_create(&notifyShmParams, &eb);
    if (dspDriverHandle0 == NULL) {
        return (Notify_E_FAIL);
    }
    
    Notify_create(NotifyDriverShm_Handle_upCast(dspDriverHandle0), 
                  dspProcId,
                  0,
                  NULL,
                  &eb);
    if (Error_check(&eb)) {
        /* Delete the driver and then return */
        NotifyDriverShm_delete(&dspDriverHandle0);
        return (Notify_E_FAIL);
    }

    if (!NotifySetup_useSecondLine) {
        return (status);
    }
    /*
     *  Setup the notify driver to the DSP (Line 1)
     */     
    NotifyDriverShm_Params_init(&notifyShmParams);
    notifyShmParams.localIntId     = NotifySetup_armRecvIntId1;
    notifyShmParams.remoteIntId    = NotifySetup_dspRecvIntId1;
    notifyShmParams.remoteProcId   = dspProcId;
    notifyShmParams.sharedAddr     = (Ptr)((UInt32)sharedAddr + 
        NotifyDriverShm_sharedMemReq(&notifyShmParams));
    dspDriverHandle1 = NotifyDriverShm_create(&notifyShmParams, &eb);
    if (dspDriverHandle1 == NULL) {
        return (Notify_E_FAIL);
    }
    
    Notify_create(NotifyDriverShm_Handle_upCast(dspDriverHandle1), 
                  dspProcId,
                  1,
                  NULL,
                  &eb);
    if (Error_check(&eb)) {
        /* Delete the driver and then return */
        NotifyDriverShm_delete(&dspDriverHandle1);
        return (Notify_E_FAIL);
    }

#endif

    return (status);
}

/*!
 *  ======== NotifySetup_sharedMemReq ========
 *  Return the amount of shared memory required
 */     
SizeT NotifySetup_sharedMemReq(UInt16 remoteProcId, Ptr sharedAddr) 
{
    SizeT memReq;
    NotifyDriverShm_Params params;
    
    NotifyDriverShm_Params_init(&params);
    params.sharedAddr = sharedAddr;
    
    if (!NotifySetup_useSecondLine) {
        memReq = NotifyDriverShm_sharedMemReq(&params);
    }
    else {
        memReq = 2 * NotifyDriverShm_sharedMemReq(&params);
    }
    
    return(memReq);
}

/*!
 * ======== NotifySetup_intLineAvailable ========
 */
Bool NotifySetup_intLineAvailable(UInt16 remoteProcId)
{
    return (TRUE);
}


/*
 */

/*
 *  @(#) ti.sdo.ipc.family.da830; 1, 0, 0, 0,98; 4-9-2010 18:38:00; /db/vtree/library/trees/ipc/ipc-d23x/src/
 */

