/* 
 * Copyright (c) 2010
 * Texas Instruments
 *
 *  All rights reserved.  Property of Texas Instruments
 *  Restricted rights to use, duplicate or disclose this code are
 *  granted through contract.
 * 
 * */
/*
 *  ======== InterruptArm.c ========
 *  DM6446 based interrupt manager.
 */

#include <xdc/std.h>
#include <xdc/runtime/Assert.h>

#include <ti/sysbios/family/arm/dm6446/Hwi.h>

#include <ti/sdo/ipc/Ipc.h>
#include <ti/sdo/ipc/notifyDrivers/IInterrupt.h>

#include "package/internal/InterruptArm.xdc.h"

/* Bit mask operations */
#define SET_BIT(num,pos)            ((num) |= (1u << (pos)))
#define CLEAR_BIT(num,pos)          ((num) &= ~(1u << (pos)))

/* register use to generate interrupt between cores */
#define INTGENREG   0x01C40010

/* event ids associated with inter-core interrupts */
#define DSP_INT0 16
#define DSP_INT1 17
#define DSP_INT2 18
#define DSP_INT3 19

#define ARM_INT0 46
#define ARM_INT1 47

/*
 *************************************************************************
 *                      Module functions
 *************************************************************************
 */

/*!
 *  ======== InterruptArm_intEnable ========
 *  Enable GPP interrupt
 */     
Void InterruptArm_intEnable(UInt16 remoteProcId, IInterrupt_IntInfo *intInfo)
{
    Hwi_enableInterrupt(intInfo->localIntId);
}
        
/*!
 *  ======== InterruptArm_intDisable ========
 *  Disables GPP interrupt
 */     
Void InterruptArm_intDisable(UInt16 remoteProcId, IInterrupt_IntInfo *intInfo)
{
    Hwi_disableInterrupt(intInfo->localIntId);
}

/*!
 *  ======== InterruptArm_intRegister ========
 *  Register ISR for remote processor interrupt
 */
Void InterruptArm_intRegister(UInt16 remoteProcId, IInterrupt_IntInfo *intInfo,
                              Fxn func, UArg arg)
{
    Hwi_Params  hwiAttrs;
    UInt        key;
    
    Assert_isTrue(intInfo->localIntId == ARM_INT0 || 
                  intInfo->localIntId == ARM_INT1,
                  Ipc_A_internal);
                  
    Assert_isTrue(intInfo->remoteIntId >= DSP_INT0 && 
                  intInfo->remoteIntId <= DSP_INT3,
                  Ipc_A_internal);

    /* Disable global interrupts */
    key = Hwi_disable();

    /* Register interrupt for communication between ARM and DSP */
    Hwi_Params_init(&hwiAttrs);
    hwiAttrs.maskSetting = Hwi_MaskingOption_SELF;
    hwiAttrs.arg         = arg;

    InterruptArm_intClear(remoteProcId, intInfo);
    
    Hwi_create(intInfo->localIntId, (Hwi_FuncPtr)func, &hwiAttrs, NULL);

    /* Restore global interrupts */
    Hwi_restore(key);

    Hwi_enableInterrupt(intInfo->localIntId);
}

/*!
 *  ======== InterruptArm_intUnregister ========
 */
Void InterruptArm_intUnregister(UInt16 remoteProcId, 
                                IInterrupt_IntInfo *intInfo)
{
    Hwi_Handle hwiHandle;
    
    /* Delete the Hwi and disable the corresponding interrupt */
    hwiHandle = Hwi_getInstance(intInfo->localIntId);
    Hwi_delete(&hwiHandle);
}

/*!
 *  ======== InterruptArm_intSend ========
 *  Send interrupt to the remote processor
 */
Void InterruptArm_intSend(UInt16 remoteProcId, IInterrupt_IntInfo *intInfo, 
                          UArg arg)
{
    UInt16 intBitPos;
    
    switch(intInfo->remoteIntId) {
        case DSP_INT0:
            intBitPos = 4;
            break;
        case DSP_INT1:
            intBitPos = 5;
            break;
        case DSP_INT2:
            intBitPos = 6;
            break;
        case DSP_INT3:
            intBitPos = 7;
            break;
    }
    
    SET_BIT(*((volatile UInt32 *)INTGENREG), (intBitPos));
}

/*!
 *  ======== InterruptArm_intPost ========
 */
Void InterruptArm_intPost(UInt16 srcProcId, IInterrupt_IntInfo *intInfo, 
                          UArg arg)
{
    UInt16 intBitPos;
    
    switch(intInfo->localIntId) {
        case ARM_INT0:
            intBitPos = 12;
            break;
        case ARM_INT1:
            intBitPos = 13;
            break;
    }
    
    SET_BIT(*((volatile UInt32 *)INTGENREG), (intBitPos));
}


/*!
 *  ======== InterruptArm_intClear ========
 *  Clear interrupt 
 */
UInt InterruptArm_intClear(UInt16 remoteProcId, IInterrupt_IntInfo *intInfo)
{
    UInt16 statBitPos;

    switch(intInfo->localIntId) {
        case ARM_INT0:
            statBitPos = 28;
            break;
        case ARM_INT1:
            statBitPos = 29;
            break;
    }

    CLEAR_BIT(*((volatile UInt32 *)INTGENREG), statBitPos);

    return (0);
}

/*
 */
/*
 *  @(#) ti.sdo.ipc.family.dm6446; 1, 0, 0, 0,116; 4-9-2010 18:38:04; /db/vtree/library/trees/ipc/ipc-d23x/src/
 */

