/* 
 * Copyright (c) 2010
 * Texas Instruments
 *
 *  All rights reserved.  Property of Texas Instruments
 *  Restricted rights to use, duplicate or disclose this code are
 *  granted through contract.
 * 
 * */
/*
 *  ======== InterruptDsp.c ========
 *  DM6446 based interrupt manager.
 */

#include <xdc/std.h>
#include <xdc/runtime/Assert.h>

#include <ti/sysbios/family/c64p/Hwi.h>

#include <ti/sdo/ipc/Ipc.h>
#include <ti/sdo/ipc/notifyDrivers/IInterrupt.h>

#include "package/internal/InterruptDsp.xdc.h"

/* Bit mask operations */
#define SET_BIT(num,pos)    ((num) |= (1u << (pos)))
#define CLEAR_BIT(num,pos)  ((num) &= ~(1u << (pos)))

/* register use to generate interrupt between cores */
#define INTGENREG   0x01C40010

/* event ids associated with inter-core interrupts */
#define DSP_INT0 16
#define DSP_INT1 17
#define DSP_INT2 18
#define DSP_INT3 19

#define ARM_INT0 46
#define ARM_INT1 47

/*
 *************************************************************************
 *                      Module functions
 *************************************************************************
 */

/*!
 *  ======== InterruptDsp_intEnable ========
 *  Enable GPP interrupt
 */     
Void InterruptDsp_intEnable(UInt16 remoteProcId, IInterrupt_IntInfo *intInfo)
{
    Hwi_enableInterrupt(intInfo->intVectorId);
}
        
/*!
 *  ======== InterruptDsp_intDisable ========
 *  Disables GPP interrupt
 */     
Void InterruptDsp_intDisable(UInt16 remoteProcId, IInterrupt_IntInfo *intInfo)
{
    Hwi_disableInterrupt(intInfo->intVectorId);
}

/*!
 *  ======== InterruptDsp_intRegister ========
 *  Register ISR for remote processor interrupt
 */
Void InterruptDsp_intRegister(UInt16 remoteProcId, IInterrupt_IntInfo *intInfo,
                              Fxn func, UArg arg)
{
    Hwi_Params  hwiAttrs;
    UInt        key;

    Assert_isTrue(intInfo->intVectorId <= 15, 
                  Ipc_A_internal);
                  
    Assert_isTrue(intInfo->localIntId >= DSP_INT0 && 
                  intInfo->localIntId <= DSP_INT3,
                  Ipc_A_internal);
                  
    Assert_isTrue(intInfo->remoteIntId == ARM_INT0 || 
                  intInfo->remoteIntId == ARM_INT1,
                  Ipc_A_internal);
    
    /* Disable global interrupts */
    key = Hwi_disable();

    InterruptDsp_intClear(remoteProcId, intInfo);

    /* Register interrupt for communication between ARM and DSP */
    Hwi_Params_init(&hwiAttrs);
    hwiAttrs.maskSetting = Hwi_MaskingOption_SELF;
    hwiAttrs.arg         = arg;
    hwiAttrs.eventId     = intInfo->localIntId;

    Hwi_create(intInfo->intVectorId,
               (Hwi_FuncPtr)func,
               &hwiAttrs,
               NULL);

    /* Restore global interrupts */
    Hwi_restore(key);

    /* enable the interrupt vector */
    Hwi_enableInterrupt(intInfo->intVectorId);
}

/*!
 *  ======== InterruptDsp_intUnregister ========
 *  Register ISR for remote processor interrupt
 */
Void InterruptDsp_intUnregister(UInt16 remoteProcId, 
                                IInterrupt_IntInfo *intInfo)
{
    Hwi_Handle  hwiHandle;

    /* Delete the Hwi (and disable the corresponding interrupt) */
    hwiHandle = Hwi_getHandle(intInfo->intVectorId);
    Hwi_delete(&hwiHandle);
}

/*!
 *  ======== InterruptDsp_intSend ========
 *  Send interrupt to the remote processor
 */
Void InterruptDsp_intSend(UInt16 remoteProcId, IInterrupt_IntInfo *intInfo, 
                          UArg arg)
{
    UInt16 intBitPos;
   
    switch(intInfo->remoteIntId) {
        case ARM_INT0:
            intBitPos = 12;
            break;
        case ARM_INT1:
            intBitPos = 13;
            break;
    }

    SET_BIT(*((volatile Uint32 *)INTGENREG), (intBitPos));
}

/*!
 *  ======== InterruptDsp_intPost ========
 */
Void InterruptDsp_intPost(UInt16 srcProcId, IInterrupt_IntInfo *intInfo, 
                          UArg arg)
{
    UInt16 intBitPos;
   
    switch(intInfo->localIntId) {
        case DSP_INT0:
            intBitPos = 4;
            break;
        case DSP_INT1:
            intBitPos = 5;
            break;
        case DSP_INT2:
            intBitPos = 6;
            break;
        case DSP_INT3:
            intBitPos = 7;
            break;
    }

    SET_BIT(*((volatile Uint32 *)INTGENREG), (intBitPos));
}

/*!
 *  ======== InterruptDsp_intClear ========
 *  Clear interrupt 
 */
UInt InterruptDsp_intClear(UInt16 remoteProcId, IInterrupt_IntInfo *intInfo)
{
    UInt16 statBitPos;

    switch(intInfo->localIntId) {
        case DSP_INT0:
            statBitPos = 20;
            break;
        case DSP_INT1:
            statBitPos = 21;
            break;
        case DSP_INT2:
            statBitPos = 22;
            break;
        case DSP_INT3:
            statBitPos = 23;
            break;
    }

    CLEAR_BIT(*((volatile Uint32 *)INTGENREG), statBitPos);

    return (0);
}

/*
 */
/*
 *  @(#) ti.sdo.ipc.family.dm6446; 1, 0, 0, 0,116; 4-9-2010 18:38:04; /db/vtree/library/trees/ipc/ipc-d23x/src/
 */

