/* 
 * Copyright (c) 2010
 * Texas Instruments
 *
 *  All rights reserved.  Property of Texas Instruments
 *  Restricted rights to use, duplicate or disclose this code are
 *  granted through contract.
 * 
 * */
/*
 *  ======== InterruptDsp.c ========
 *  Mailbox based interrupt manager
 */
#include <xdc/std.h>
#include <xdc/runtime/Startup.h>
#include <xdc/runtime/System.h>
#include <xdc/runtime/Assert.h>

#include <ti/sysbios/family/c64p/Hwi.h>

#include <ti/sdo/utils/_MultiProc.h>
#include <ti/sdo/ipc/_Ipc.h>

#include <ti/sdo/ipc/notifyDrivers/IInterrupt.h>

#include "package/internal/InterruptDsp.xdc.h"

/* Register access method. */
#define REG16(A)   (*(volatile UInt16 *) (A))
#define REG32(A)   (*(volatile UInt32 *) (A))

/* 
 *  ======== Mailboxes ========
 *      Mailbox     Usage
 *      0           DSP to HOST
 *      1           DSP to VIDEO
 *      2           DSP to DSS
 *      3           HOST to DSP
 *      4           HOST to VIDEO
 *      5           HOST to DSS
 *      6           VIDEO to HOST
 *      7           VIDEO to DSP
 *      8           DSS to HOST
 *      9           DSS to DSP           
 */
#define MAILBOX_MESSAGE_0           InterruptDsp_mailboxBaseAddr + 0x40
#define MAILBOX_MESSAGE_1           InterruptDsp_mailboxBaseAddr + 0x44
#define MAILBOX_MESSAGE_2           InterruptDsp_mailboxBaseAddr + 0x48
#define MAILBOX_MESSAGE_3           InterruptDsp_mailboxBaseAddr + 0x4C
#define MAILBOX_MESSAGE_4           InterruptDsp_mailboxBaseAddr + 0x50
#define MAILBOX_MESSAGE_5           InterruptDsp_mailboxBaseAddr + 0x54
#define MAILBOX_MESSAGE_6           InterruptDsp_mailboxBaseAddr + 0x58
#define MAILBOX_MESSAGE_7           InterruptDsp_mailboxBaseAddr + 0x5C
#define MAILBOX_MESSAGE_9           InterruptDsp_mailboxBaseAddr + 0x64
#define MAILBOX_MSGSTATUS_0         InterruptDsp_mailboxBaseAddr + 0xC0
#define MAILBOX_MSGSTATUS_1         InterruptDsp_mailboxBaseAddr + 0xC4
#define MAILBOX_MSGSTATUS_2         InterruptDsp_mailboxBaseAddr + 0xC8
#define MAILBOX_MSGSTATUS_3         InterruptDsp_mailboxBaseAddr + 0xCC
#define MAILBOX_MSGSTATUS_4         InterruptDsp_mailboxBaseAddr + 0xD0
#define MAILBOX_MSGSTATUS_5         InterruptDsp_mailboxBaseAddr + 0xD4
#define MAILBOX_MSGSTATUS_6         InterruptDsp_mailboxBaseAddr + 0xD8
#define MAILBOX_MSGSTATUS_7         InterruptDsp_mailboxBaseAddr + 0xDC
#define MAILBOX_MSGSTATUS_8         InterruptDsp_mailboxBaseAddr + 0xE0
#define MAILBOX_MSGSTATUS_9         InterruptDsp_mailboxBaseAddr + 0xE4
#define MAILBOX_IRQSTATUS_CLR_DSP   InterruptDsp_mailboxBaseAddr + 0x114
#define MAILBOX_IRQENABLE_SET_DSP   InterruptDsp_mailboxBaseAddr + 0x118 
#define MAILBOX_IRQENABLE_CLR_DSP   InterruptDsp_mailboxBaseAddr + 0x11C
#define MAILBOX_EOI_REG             InterruptDsp_mailboxBaseAddr + 0x140

#define HOSTINT                77
#define M3VIDEOINT             53
#define M3DSSINT               54
#define DSPINT                 56

/*
 *************************************************************************
 *                      Module functions
 *************************************************************************
 */

/*!
 *  ======== InterruptDsp_intEnable ========
 *  Enable remote processor interrupt
 */     
Void InterruptDsp_intEnable(UInt16 remoteProcId, IInterrupt_IntInfo *intInfo)
{
    if (remoteProcId == InterruptDsp_hostProcId) {
        REG32(MAILBOX_IRQENABLE_SET_DSP) = 0x40;
    }
    else if (remoteProcId == InterruptDsp_videoProcId) {
        REG32(MAILBOX_IRQENABLE_SET_DSP) = 0x4000;
    }
    else if (remoteProcId == InterruptDsp_dssProcId) {
        REG32(MAILBOX_IRQENABLE_SET_DSP) = 0x40000;
    }
    else {
        Assert_isTrue(FALSE, Ipc_A_internal);
    }
}
        
/*!
 *  ======== InterruptDsp_intDisable ========
 *  Disables remote processor interrupt
 */     
Void InterruptDsp_intDisable(UInt16 remoteProcId, IInterrupt_IntInfo *intInfo)
{
    if (remoteProcId == InterruptDsp_hostProcId) {
        REG32(MAILBOX_IRQENABLE_CLR_DSP) = 0x40;
    }
    else if (remoteProcId == InterruptDsp_videoProcId) {
        REG32(MAILBOX_IRQENABLE_CLR_DSP) = 0x4000;
    }
    else if (remoteProcId == InterruptDsp_dssProcId) {
        REG32(MAILBOX_IRQENABLE_CLR_DSP) = 0x40000;
    }
    else {
        Assert_isTrue(FALSE, Ipc_A_internal);
    }
}

/*!
 *  ======== InterruptDsp_intRegister ======== 
 */
Void InterruptDsp_intRegister(UInt16 remoteProcId, IInterrupt_IntInfo *intInfo,
                              Fxn func, UArg arg)
{
    UInt        key;
    Int         index; 
    InterruptDsp_FxnTable *table;    
   
    Assert_isTrue(intInfo->intVectorId <= 15, Ipc_A_internal);
    
    if (remoteProcId == InterruptDsp_hostProcId) {
        index = 0;
    }
    else if (remoteProcId == InterruptDsp_videoProcId) {
        index = 1;
    }
    else if (remoteProcId == InterruptDsp_dssProcId) {
        index = 2;
    }
    else {
        Assert_isTrue(FALSE, Ipc_A_internal);
    }
    
    /* Disable global interrupts */
    key = Hwi_disable();

    table = &(InterruptDsp_module->fxnTable[index]);
    table->func = func;
    table->arg  = arg;
    
    InterruptDsp_intClear(remoteProcId, intInfo);
    
    /* Make sure the interrupt only gets plugged once */
    InterruptDsp_module->numPlugged++;
    if (InterruptDsp_module->numPlugged == 1) { 
        /* Map the interrupt number to HWI vector */
        Hwi_eventMap(intInfo->intVectorId, DSPINT);    

        Hwi_create(intInfo->intVectorId,
                   (Hwi_FuncPtr)InterruptDsp_intShmStub,
                   NULL,
                   NULL);
        Hwi_enableInterrupt(intInfo->intVectorId);
    }
    
    /* Enable the mailbox interrupt to the DSP */
    InterruptDsp_intEnable(remoteProcId, intInfo);
    
    /* Restore global interrupts */
    Hwi_restore(key);
}

/*!
 *  ======== InterruptDsp_intUnregister ======== 
 */
Void InterruptDsp_intUnregister(UInt16 remoteProcId, 
                                IInterrupt_IntInfo *intInfo)
{
    Hwi_Handle  hwiHandle;
    Int         index; 
    InterruptDsp_FxnTable *table;
    
    if (remoteProcId == InterruptDsp_hostProcId) {
        index = 0;
    }
    else if (remoteProcId == InterruptDsp_videoProcId) {
        index = 1;
    }
    else if (remoteProcId == InterruptDsp_dssProcId) {
        index = 2;
    }
    else {
        Assert_isTrue(FALSE, Ipc_A_internal);
    }
    
    /* Disable the mailbox interrupt source */
    InterruptDsp_intDisable(remoteProcId, intInfo);

    InterruptDsp_module->numPlugged--;
    if (InterruptDsp_module->numPlugged == 0) {
        /* Delete the Hwi */
        hwiHandle = Hwi_getHandle(intInfo->intVectorId);
        Hwi_delete(&hwiHandle);
    }

    table = &(InterruptDsp_module->fxnTable[index]);
    table->func = NULL;        
    table->arg  = NULL;
}

/*!
 *  ======== InterruptDsp_intSend ========
 *  Send interrupt to the remote processor
 */
Void InterruptDsp_intSend(UInt16 remoteProcId, IInterrupt_IntInfo *intInfo,
                          UArg arg)
{
    UInt key;
    
    /* 
     *  Before writing to a mailbox, check whehter it already contains a message
     *  If so, then don't write to the mailbox since we want one and only one
     *  message per interrupt.  Disable interrupts between reading 
     *  the MSGSTATUS_X register and writing to the mailbox to protect from 
     *  another thread doing an intSend at the same time
     *
     *  Note regarding possible race condition between local 'intSend' and 
     *  remote 'intClear':
     *  It is possible that we we read the MAILBOX_MSGSTATUS_X register during
     *  the remote side's intClear.  Therefore, we might choose _not_ to send
     *  write to the mailbox even though the mailbox is about to be cleared a
     *  few cycles later. In this case, the interrupt will be lost.
     *  This is OK, however. intClear should always be called by the Notify
     *  driver _before_ shared memory is read, so the event will be picked up
     *  anyway by the previous interrupt that caused intClear to be called.
     */
    if (remoteProcId == InterruptDsp_hostProcId) {
        key = Hwi_disable();
        if (REG32(MAILBOX_MSGSTATUS_0) == 0) {
            REG32(MAILBOX_MESSAGE_0) = arg;
        }
        Hwi_restore(key);
    }
    else if (remoteProcId == InterruptDsp_videoProcId) {
        key = Hwi_disable();
        if (REG32(MAILBOX_MSGSTATUS_1) == 0) {
            REG32(MAILBOX_MESSAGE_1) = arg;
        }
        Hwi_restore(key);
    }
    else { /* DSS */
        key = Hwi_disable();
        if (REG32(MAILBOX_MSGSTATUS_2) == 0) {
            REG32(MAILBOX_MESSAGE_2) = arg;
        }
        Hwi_restore(key);
    }
}

/*!
 *  ======== InterruptDsp_intPost ========
 */
Void InterruptDsp_intPost(UInt16 srcProcId, IInterrupt_IntInfo *intInfo, 
                          UArg arg)
{
    UInt key;
    
    if (srcProcId == InterruptDsp_dssProcId) {
        key = Hwi_disable();
        if (REG32(MAILBOX_MSGSTATUS_9) == 0) {
            REG32(MAILBOX_MESSAGE_9) = arg;
        }
        Hwi_restore(key);
    }
    else if (srcProcId == InterruptDsp_videoProcId) {
        key = Hwi_disable();
        if (REG32(MAILBOX_MSGSTATUS_7) == 0) {
            REG32(MAILBOX_MESSAGE_7) = arg;
        }
        Hwi_restore(key);
    }
    else {  /* hostProcId */
        key = Hwi_disable();
        if (REG32(MAILBOX_MSGSTATUS_3) == 0) {
            REG32(MAILBOX_MESSAGE_3) = arg;
        }
        Hwi_restore(key);
    }
}

/*!
 *  ======== InterruptDsp_intClear ========
 *  Clear interrupt 
 */
UInt InterruptDsp_intClear(UInt16 remoteProcId, IInterrupt_IntInfo *intInfo)
{
    UInt arg;
    
    if (remoteProcId == InterruptDsp_hostProcId) {
        arg = REG32(MAILBOX_MESSAGE_3);  
        REG32(MAILBOX_IRQSTATUS_CLR_DSP) = 0x40;
    }
    else if (remoteProcId == InterruptDsp_videoProcId) {
        arg = REG32(MAILBOX_MESSAGE_7);  
        REG32(MAILBOX_IRQSTATUS_CLR_DSP) = 0x4000;            
    }
    else { /* DSS */
        arg = REG32(MAILBOX_MESSAGE_9);  
        REG32(MAILBOX_IRQSTATUS_CLR_DSP) = 0x40000;        
    }
    
    /* Write to EOI (End Of Interrupt) register */
    REG32(MAILBOX_EOI_REG) = 0x1;

    return (arg);
}

/*
 *************************************************************************
 *                      Internals functions
 *************************************************************************
 */

/*!
 *  ======== InterruptDsp_intShmStub ======== 
 */
Void InterruptDsp_intShmStub(UArg arg)
{
    InterruptDsp_FxnTable *table;    
    
    /* Process messages from the HOST */    
    if ((REG32(MAILBOX_IRQENABLE_SET_DSP) & 0x40) 
        && REG32(MAILBOX_MSGSTATUS_3) != 0) {
        table = &(InterruptDsp_module->fxnTable[0]);    
        (table->func)(table->arg);     
    }
    
    /* Process messages from the M3 core 0  */    
    if ((REG32(MAILBOX_IRQENABLE_SET_DSP) & 0x4000)
        && REG32(MAILBOX_MSGSTATUS_7) != 0) {
        table = &(InterruptDsp_module->fxnTable[1]);    
        (table->func)(table->arg);     
    }
    
    /* Process messages from the M3 core 1  */    
    if ((REG32(MAILBOX_IRQENABLE_SET_DSP) & 0x40000)
        && REG32(MAILBOX_MSGSTATUS_9) != 0) {
        table = &(InterruptDsp_module->fxnTable[2]);    
        (table->func)(table->arg);     
    }
}

/*
 */
/*
 *  @(#) ti.sdo.ipc.family.dm740; 1, 0, 0, 0,116; 4-9-2010 18:38:06; /db/vtree/library/trees/ipc/ipc-d23x/src/
 */

