/* 
 * Copyright (c) 2010
 * Texas Instruments
 *
 *  All rights reserved.  Property of Texas Instruments
 *  Restricted rights to use, duplicate or disclose this code are
 *  granted through contract.
 *
 * */
/*
 *  ======== InterruptDucati.c ========
 *  Ducati/DM740 based interupt manager
 */

#include <xdc/std.h>
#include <xdc/runtime/Assert.h>

#include <ti/sysbios/family/arm/m3/Hwi.h>
#include <ti/sysbios/family/arm/ducati/Core.h>
#include <ti/sdo/ipc/notifyDrivers/IInterrupt.h>
#include <ti/sdo/ipc/_Ipc.h>
#include <ti/sdo/utils/_MultiProc.h>

#include "package/internal/InterruptDucati.xdc.h"

/* Register access method. */
#define REG16(A)   (*(volatile UInt16 *) (A))
#define REG32(A)   (*(volatile UInt32 *) (A))

/*
 *  Ducati control register that maintains inter-core interrupt bits.
 *
 *  Using separate core0 and core1 values to do 16-bit reads/writes
 *  because we do not want to overwrite the other cores value.
 */
#define INTERRUPT_CORE_0            (InterruptDucati_ducatiCtrlBaseAddr)
#define INTERRUPT_CORE_1            (InterruptDucati_ducatiCtrlBaseAddr + 2)

/* 
 *  ======== Mailboxes ========
 *      Mailbox     Usage
 *      0           DSP to HOST
 *      1           DSP to VIDEO
 *      2           DSP to DSS
 *      3           HOST to DSP
 *      4           HOST to VIDEO
 *      5           HOST to DSS
 *      6           VIDEO to HOST
 *      7           VIDEO to DSP
 *      8           DSS to HOST
 *      9           DSS to DSP           
 */
#define MAILBOX_MESSAGE_0           InterruptDucati_mailboxBaseAddr + 0x40
#define MAILBOX_MESSAGE_1           InterruptDucati_mailboxBaseAddr + 0x44
#define MAILBOX_MESSAGE_2           InterruptDucati_mailboxBaseAddr + 0x48
#define MAILBOX_MESSAGE_3           InterruptDucati_mailboxBaseAddr + 0x4C
#define MAILBOX_MESSAGE_4           InterruptDucati_mailboxBaseAddr + 0x50
#define MAILBOX_MESSAGE_5           InterruptDucati_mailboxBaseAddr + 0x54
#define MAILBOX_MESSAGE_6           InterruptDucati_mailboxBaseAddr + 0x58
#define MAILBOX_MESSAGE_7           InterruptDucati_mailboxBaseAddr + 0x5C
#define MAILBOX_MESSAGE_8           InterruptDucati_mailboxBaseAddr + 0x60
#define MAILBOX_MESSAGE_9           InterruptDucati_mailboxBaseAddr + 0x64
#define MAILBOX_MSGSTATUS_1         InterruptDucati_mailboxBaseAddr + 0xC4
#define MAILBOX_MSGSTATUS_2         InterruptDucati_mailboxBaseAddr + 0xC8
#define MAILBOX_MSGSTATUS_3         InterruptDucati_mailboxBaseAddr + 0xCC
#define MAILBOX_MSGSTATUS_4         InterruptDucati_mailboxBaseAddr + 0xD0
#define MAILBOX_MSGSTATUS_5         InterruptDucati_mailboxBaseAddr + 0xD4
#define MAILBOX_MSGSTATUS_6         InterruptDucati_mailboxBaseAddr + 0xD8
#define MAILBOX_MSGSTATUS_7         InterruptDucati_mailboxBaseAddr + 0xDC
#define MAILBOX_MSGSTATUS_8         InterruptDucati_mailboxBaseAddr + 0xE0
#define MAILBOX_MSGSTATUS_9         InterruptDucati_mailboxBaseAddr + 0xE4

/* Ducati CORE0 registers */
#define MAILBOX_IRQSTATUS_CLR_0     InterruptDucati_mailboxBaseAddr + 0x124
#define MAILBOX_IRQENABLE_SET_0     InterruptDucati_mailboxBaseAddr + 0x128
#define MAILBOX_IRQENABLE_CLR_0     InterruptDucati_mailboxBaseAddr + 0x12C

/* Ducati CORE1 registers */
#define MAILBOX_IRQSTATUS_CLR_1     InterruptDucati_mailboxBaseAddr + 0x134
#define MAILBOX_IRQENABLE_SET_1     InterruptDucati_mailboxBaseAddr + 0x138
#define MAILBOX_IRQENABLE_CLR_1     InterruptDucati_mailboxBaseAddr + 0x13C

#define HOSTINT                77
#define M3VIDEOINT             53
#define M3DSSINT               54
#define DSPINT                 56
#define M3INT                  19

/*
 *************************************************************************
 *                      Module functions
 *************************************************************************
 */

/*!
 *  ======== InterruptDucati_intEnable ========
 *  Enable remote processor interrupt
 */
Void InterruptDucati_intEnable(UInt16 remoteProcId, IInterrupt_IntInfo *intInfo)
{
    /*  
     *  If the remote processor communicates via mailboxes, we should enable 
     *  the Mailbox IRQ instead of enabling the Hwi because multiple mailboxes
     *  share the same Hwi
     */
    if (Core_getId() == 0) {
        if (remoteProcId == InterruptDucati_hostProcId) {
            REG32(MAILBOX_IRQENABLE_SET_0) = 0x100;
        }
        else if (remoteProcId == InterruptDucati_dspProcId) {
            REG32(MAILBOX_IRQENABLE_SET_0) = 0x4;
        }
        else {
            Hwi_enableInterrupt(M3INT);
        }
    }
    else {
        if (remoteProcId == InterruptDucati_hostProcId) {
            REG32(MAILBOX_IRQENABLE_SET_1) = 0x400;
        }
        else if (remoteProcId == InterruptDucati_dspProcId) {
            REG32(MAILBOX_IRQENABLE_SET_1) = 0x10;
        }
        else {
            Hwi_enableInterrupt(M3INT);
        }
    }
}

/*!
 *  ======== InterruptDucati_intDisable ========
 *  Disables remote processor interrupt
 */
Void InterruptDucati_intDisable(UInt16 remoteProcId, 
                                IInterrupt_IntInfo *intInfo)
{
    /*  
     *  If the remote processor communicates via mailboxes, we should disable 
     *  the Mailbox IRQ instead of disabling the Hwi because multiple mailboxes
     *  share the same Hwi
     */
    if (Core_getId() == 0) {
        if (remoteProcId == InterruptDucati_hostProcId) {
            REG32(MAILBOX_IRQENABLE_CLR_0) = 0x100;
        }
        else if (remoteProcId == InterruptDucati_dspProcId) {
            REG32(MAILBOX_IRQENABLE_CLR_0) = 0x4;
        }
        else {
            Hwi_disableInterrupt(M3INT);
        }
    }
    else {
        if (remoteProcId == InterruptDucati_hostProcId) {
            REG32(MAILBOX_IRQENABLE_CLR_1) = 0x400;
        }
        else if (remoteProcId == InterruptDucati_dspProcId) {
            REG32(MAILBOX_IRQENABLE_CLR_1) = 0x10;
        }
        else {
            Hwi_disableInterrupt(M3INT);
        }
    }
}

/*!
 *  ======== InterruptDucati_intRegister ========
 */
Void InterruptDucati_intRegister(UInt16 remoteProcId,
                                 IInterrupt_IntInfo *intInfo,
                                 Fxn func, UArg arg)
{
    Hwi_Params  hwiAttrs;
    UInt        key;
    Int         index;
    InterruptDucati_FxnTable *table;

    Assert_isTrue(remoteProcId < MultiProc_numProcessors, Ipc_A_internal);

    /* Assert that our MultiProc id is set correctly */
    Assert_isTrue((InterruptDucati_videoProcId == MultiProc_self()) ||
                  (InterruptDucati_dssProcId == MultiProc_self()),
                  Ipc_A_internal);

    /*
     *  DSS & VIDEO each have a unique interrupt ID for receiving interrupts
     *  external to the Ducati subsystem.  (M3DSSINT & M3VIDEOINT). However,
     *  they have a separate interrupt ID for receving interrupt from each other
     *  (M3INT).
     *  Store the interrupt id in the intInfo so it can be used during
     *  intUnregiseter.
     */
    if (remoteProcId == InterruptDucati_dspProcId) {
        index = 0;
        if (Core_getId() == 0) {
            intInfo->localIntId = M3VIDEOINT;
        }
        else {
            intInfo->localIntId = M3DSSINT;
        }
    }
    else if (remoteProcId == InterruptDucati_hostProcId) {
        index = 1;
        if (Core_getId() == 0) {
            intInfo->localIntId = M3VIDEOINT;
        }
        else {
            intInfo->localIntId = M3DSSINT;
        }
    }
    else {
        /* Going to the other M3 */
        index = 2;
        intInfo->localIntId = M3INT;
    }

    /* Disable global interrupts */
    key = Hwi_disable();

    table = &(InterruptDucati_module->fxnTable[index]);
    table->func = func;
    table->arg  = arg;

    InterruptDucati_intClear(remoteProcId, intInfo);

    Hwi_Params_init(&hwiAttrs);
    hwiAttrs.maskSetting = Hwi_MaskingOption_LOWER;

    /* Make sure the interrupt only gets plugged once */
    if (remoteProcId == InterruptDucati_videoProcId ||
        remoteProcId == InterruptDucati_dssProcId) {
        Hwi_create(intInfo->localIntId,
                   (Hwi_FuncPtr)InterruptDucati_intShmDucatiStub,
                   &hwiAttrs,
                   NULL);
    }
    else {
        InterruptDucati_module->numPlugged++;
        if (InterruptDucati_module->numPlugged == 1) {
            Hwi_create(intInfo->localIntId,
                       (Hwi_FuncPtr)InterruptDucati_intShmMbxStub,
                       &hwiAttrs,
                       NULL);
            
            /* Interrupt_intEnable won't enable the Hwi */
            Hwi_enableInterrupt(intInfo->localIntId);
        }
    }
    
    InterruptDucati_intEnable(remoteProcId, intInfo);
    
    /* Restore global interrupts */
    Hwi_restore(key);
}

/*!
 *  ======== InterruptDucati_intUnregister ========
 */
Void InterruptDucati_intUnregister(UInt16 remoteProcId,
                                   IInterrupt_IntInfo *intInfo)
{
    Int         index;
    InterruptDucati_FxnTable *table;
    Hwi_Handle  hwiHandle;

    if (remoteProcId == InterruptDucati_dspProcId) {
        index = 0;
    }
    else if (remoteProcId == InterruptDucati_hostProcId) {
        index = 1;
    }
    else {
        /* Going to the other M3 */
        index = 2;
    }

    /* Disable the mailbox interrupt source */
    InterruptDucati_intDisable(remoteProcId, intInfo);

    /* Disable the interrupt itself */
    if (remoteProcId == InterruptDucati_videoProcId ||
        remoteProcId == InterruptDucati_dssProcId) {
        hwiHandle = Hwi_getHandle(M3INT);
        Hwi_delete(&hwiHandle);
    }
    else {
        InterruptDucati_module->numPlugged--;
        if (InterruptDucati_module->numPlugged == 0) {
            hwiHandle = Hwi_getHandle(intInfo->localIntId);
            Hwi_delete(&hwiHandle);
        }
    }

    /* Clear the FxnTable entry for the remote processor */
    table = &(InterruptDucati_module->fxnTable[index]);
    table->func = NULL;
    table->arg  = 0;
}


/*!
 *  ======== InterruptDucati_intSend ========
 *  Send interrupt to the remote processor
 */
Void InterruptDucati_intSend(UInt16 remoteProcId, IInterrupt_IntInfo *intInfo,
                             UArg arg)
{
    UInt key;

    if (remoteProcId == InterruptDucati_videoProcId ||
        remoteProcId == InterruptDucati_dssProcId) {
        if (Core_getId() == 1) {
            /* DSS to VIDEO */
            REG16(INTERRUPT_CORE_0) |= 0x1;
        }
        else {
            /* VIDEO to DSS */
            REG16(INTERRUPT_CORE_1) |= 0x1;
        }
    }
    else if (remoteProcId == InterruptDucati_dspProcId) {
        if (Core_getId() == 1) {
            /* DSS to DSP */
            key = Hwi_disable();
            if (REG32(MAILBOX_MSGSTATUS_9) == 0) {
                REG32(MAILBOX_MESSAGE_9) = arg;
            }
            Hwi_restore(key);
        }
        else {
            /* VIDEO to DSP */
            key = Hwi_disable();
            if (REG32(MAILBOX_MSGSTATUS_7) == 0) {
                REG32(MAILBOX_MESSAGE_7) = arg;
            }
            Hwi_restore(key);
        }
    }
    else { /* HOSTINT */
        if (Core_getId() == 1) {
            /* DSS to HOST */
            key = Hwi_disable();
            if (REG32(MAILBOX_MSGSTATUS_8) == 0) {
                REG32(MAILBOX_MESSAGE_8) = arg;
            }
            Hwi_restore(key);
        }
        else {
            /* VIDEO to HOST */
            key = Hwi_disable();
            if (REG32(MAILBOX_MSGSTATUS_6) == 0) {
                REG32(MAILBOX_MESSAGE_6) = arg;
            }
            Hwi_restore(key);
        }
    }
}


/*!
 *  ======== InterruptDucati_intPost ========
 *  Simulate an interrupt from a remote processor
 */
Void InterruptDucati_intPost(UInt16 srcProcId, IInterrupt_IntInfo *intInfo,
                             UArg arg)
{
    UInt key;

    if (srcProcId == InterruptDucati_videoProcId ||
        srcProcId == InterruptDucati_dssProcId) {
        if (Core_getId() == 1) {
            /* VIDEO to DSS */
            REG16(INTERRUPT_CORE_1) |= 0x1;
        }
        else {
            /* DSS to VIDEO */
            REG16(INTERRUPT_CORE_0) |= 0x1;
        }
    }
    else if (srcProcId == InterruptDucati_dspProcId) {
        if (Core_getId() == 1) {
            /* DSP to DSS */
            key = Hwi_disable();
            if (REG32(MAILBOX_MSGSTATUS_2) == 0) {
                REG32(MAILBOX_MESSAGE_2) = arg;
            }
            Hwi_restore(key);
        }
        else {
            /* DSP to VIDEO */
            key = Hwi_disable();
            if (REG32(MAILBOX_MSGSTATUS_1) == 0) {
                REG32(MAILBOX_MESSAGE_1) = arg;
            }
            Hwi_restore(key);
        }
    }
    else { /* HOSTINT */
        if (Core_getId() == 1) {
            /* HOST to DSS */
            key = Hwi_disable();
            if (REG32(MAILBOX_MSGSTATUS_5) == 0) {
                REG32(MAILBOX_MESSAGE_5) = arg;
            }
            Hwi_restore(key);
        }
        else {
            /* HOST to VIDEO */
            key = Hwi_disable();
            if (REG32(MAILBOX_MSGSTATUS_4) == 0) {
                REG32(MAILBOX_MESSAGE_4) = arg;
            }
            Hwi_restore(key);
        }
    }
}


/*!
 *  ======== InterruptDucati_intClear ========
 *  Clear interrupt
 */
UInt InterruptDucati_intClear(UInt16 remoteProcId, IInterrupt_IntInfo *intInfo)
{
    UInt arg;

    if (remoteProcId == InterruptDucati_videoProcId ||
        remoteProcId == InterruptDucati_dssProcId) {
        arg = REG32(InterruptDucati_ducatiCtrlBaseAddr);

        /* Look at BIOS's ducati Core id */
        if (Core_getId() == 0) {
            if ((REG16(INTERRUPT_CORE_0) & 0x1) == 0x1) {
                /* DSS to VIDEO */
                REG16(INTERRUPT_CORE_0) &= ~(0x1);
            }
        }
        else {
            if ((REG16(INTERRUPT_CORE_1) & 0x1) == 0x1) {
                /* VIDEO to DSS */
                REG16(INTERRUPT_CORE_1) &= ~(0x1);
            }
        }
    }
    else if (Core_getId() == 0) {
        if (remoteProcId == InterruptDucati_hostProcId) {
            /* HOST to VIDEO */
            arg = REG32(MAILBOX_MESSAGE_4);
            REG32(MAILBOX_IRQSTATUS_CLR_0) = 0x100;
        }
        else {
            /* DSP to VIDEO */
            arg = REG32(MAILBOX_MESSAGE_1);
            REG32(MAILBOX_IRQSTATUS_CLR_0) = 0x4;
        }
    }
    else { /* M3DSSINT */
        if (remoteProcId == InterruptDucati_hostProcId) {
            /* HOST to DSS */
            arg = REG32(MAILBOX_MESSAGE_5);
            REG32(MAILBOX_IRQSTATUS_CLR_1) = 0x400;
        }
        else {
            /* DSP to DSS */
            arg = REG32(MAILBOX_MESSAGE_2);
            REG32(MAILBOX_IRQSTATUS_CLR_1) = 0x10;
        }
    }

    return (arg);
}

/*
 *************************************************************************
 *                      Internals functions
 *************************************************************************
 */

/*!
 *  ======== InterruptDucati_intShmDucatiStub ========
 */
Void InterruptDucati_intShmDucatiStub(UArg arg)
{
    InterruptDucati_FxnTable *table;

    table = &(InterruptDucati_module->fxnTable[2]);
    (table->func)(table->arg);
}

/*!
 *  ======== InterruptDucati_intShmMbxStub ========
 */
Void InterruptDucati_intShmMbxStub(UArg arg)
{
    InterruptDucati_FxnTable *table;

    if (Core_getId() == 0) {
        if ((REG32(MAILBOX_IRQENABLE_SET_0) & 0x4) && 
            REG32(MAILBOX_MSGSTATUS_1) != 0) { /* DSP to VIDEO */
            table = &(InterruptDucati_module->fxnTable[0]);
            (table->func)(table->arg);
        }
        if ((REG32(MAILBOX_IRQENABLE_SET_0) & 0x100) && 
            REG32(MAILBOX_MSGSTATUS_4) != 0) { /* HOST to VIDEO */
            table = &(InterruptDucati_module->fxnTable[1]);
            (table->func)(table->arg);
        }
    }
    else {
        if ((REG32(MAILBOX_IRQENABLE_SET_1) & 0x10) &&
             REG32(MAILBOX_MSGSTATUS_2) != 0) { /* DSP to DSS */
            table = &(InterruptDucati_module->fxnTable[0]);
            (table->func)(table->arg);
        }
        if ((REG32(MAILBOX_IRQENABLE_SET_1) & 0x400) &&
            REG32(MAILBOX_MSGSTATUS_5) != 0) { /* HOST to DSS */
            table = &(InterruptDucati_module->fxnTable[1]);
            (table->func)(table->arg);
        }
    }
}

/*
 */
/*
 *  @(#) ti.sdo.ipc.family.dm740; 1, 0, 0, 0,116; 4-9-2010 18:38:06; /db/vtree/library/trees/ipc/ipc-d23x/src/
 */

