/* 
 * Copyright (c) 2010
 * Texas Instruments
 *
 *  All rights reserved.  Property of Texas Instruments
 *  Restricted rights to use, duplicate or disclose this code are
 *  granted through contract.
 * 
 * */
/*
 *  ======== InterruptDsp.c ========
 *  Mailbox based interrupt manager
 */

#include <xdc/std.h>
#include <xdc/runtime/Assert.h>

#include <ti/sysbios/family/c64p/Hwi.h>
#include <ti/sysbios/family/c64p/omap3430/Wugen.h>

#include <ti/sdo/ipc/notifyDrivers/IInterrupt.h>

#include <ti/sdo/ipc/_Ipc.h>
#include <ti/sdo/utils/_MultiProc.h>

#include "package/internal/InterruptDsp.xdc.h"

/* Register access method. */
#define REG32(A)   (*(volatile UInt32 *) (A))

/* Mailbox management values */
#define MAILBOX_BASE           0x48094000
#define MAILBOX_MESSAGE_0      MAILBOX_BASE + 0x40
#define MAILBOX_MESSAGE_1      MAILBOX_BASE + 0x44
#define MAILBOX_MSGSTATUS_0    MAILBOX_BASE + 0xC0
#define MAILBOX_MSGSTATUS_1    MAILBOX_BASE + 0xC4
#define MAILBOX_IRQSTATUS_0    MAILBOX_BASE + 0x100
#define MAILBOX_IRQSTATUS_1    MAILBOX_BASE + 0x108
#define MAILBOX_IRQENABLE_0    MAILBOX_BASE + 0x104
#define MAILBOX_IRQENABLE_1    MAILBOX_BASE + 0x10C

#define DSPINT  55
#define ARMINT  26


/*
 *************************************************************************
 *                      Module functions
 *************************************************************************
 */

/*!
 *  ======== InterruptDsp_intEnable ========
 */     
Void InterruptDsp_intEnable(UInt16 remoteProcId, IInterrupt_IntInfo *intInfo)
{
    Hwi_enableInterrupt(intInfo->intVectorId);
}
        
/*!
 *  ======== InterruptDsp_intDisable ========
 */     
Void InterruptDsp_intDisable(UInt16 remoteProcId, IInterrupt_IntInfo *intInfo)
{
    Hwi_disableInterrupt(intInfo->intVectorId);
}

/*!
 *  ======== InterruptDsp_intRegister ======== 
 */
Void InterruptDsp_intRegister(UInt16 remoteProcId, IInterrupt_IntInfo *intInfo,
                              Fxn func, UArg arg)
{
    Hwi_Params  hwiAttrs;
    UInt        key;
    
    /* Make sure that we're trying to talk to the HOST */
    Assert_isTrue(remoteProcId == MultiProc_getId("HOST"), Ipc_A_invArgument);
    
    /* Disable global interrupts */
    key = Hwi_disable();

    InterruptDsp_intClear(remoteProcId, intInfo);

    /* Register interrupt for communication between ARM and DSP */
    Hwi_Params_init(&hwiAttrs);
    hwiAttrs.maskSetting = Hwi_MaskingOption_SELF;
    hwiAttrs.arg         = arg;
    hwiAttrs.eventId     = DSPINT;
    Hwi_create(intInfo->intVectorId,
               (Hwi_FuncPtr)func,
               &hwiAttrs,
               NULL);
               
    /* Enable the mailbox interrupt to the DSP */
    REG32(MAILBOX_IRQENABLE_1) = 0x4;

    /* Restore global interrupts */
    Hwi_restore(key);

    /* Unmask IVA_2_IRQ[10] to allow interrupt to come into DSP */
    Wugen_enableEvent(DSPINT);
    Hwi_enableInterrupt(intInfo->intVectorId);
}

/*!
 *  ======== InterruptDsp_intUnregister ======== 
 */
Void InterruptDsp_intUnregister(UInt16 remoteProcId, 
                                IInterrupt_IntInfo *intInfo)
{
    Hwi_Handle  hwiHandle;

    /* Delete the Hwi (and disable the corresponding interrupt) */
    hwiHandle = Hwi_getHandle(intInfo->intVectorId);
    Hwi_delete(&hwiHandle);
}

/*!
 *  ======== InterruptDsp_intSend ========
 *  Send interrupt to the remote processor
 */
Void InterruptDsp_intSend(UInt16 remoteProcId, IInterrupt_IntInfo *intInfo, 
                          UArg arg)
{
    /* Put into the GPP's mailbox to generate the interrupt. */
    REG32(MAILBOX_MESSAGE_0) = arg;
}

/*!
 *  ======== InterruptDsp_intPost ========
 */
Void InterruptDsp_intPost(UInt16 srcProcId, IInterrupt_IntInfo *intInfo, 
                          UArg arg)
{
    /* Put into the GPP's mailbox to generate the interrupt. */
    REG32(MAILBOX_MESSAGE_1) = arg;
}

/*!
 *  ======== Interrupt_intClear ========
 *  Clear interrupt 
 */
UInt InterruptDsp_intClear(UInt16 remoteProcId, IInterrupt_IntInfo *intInfo)
{
    UInt arg;

    /* Read into val. */
    arg = REG32(MAILBOX_MESSAGE_1);

    /* 
     * Clear the IRQ status.
     * If there are more in the mailbox FIFO, it will re-assert.
     */
    REG32(MAILBOX_IRQSTATUS_1) = 0x4;

    return (arg);
}

/*
 */
/*
 *  @(#) ti.sdo.ipc.family.omap3530; 1, 0, 0, 0,108; 4-9-2010 18:38:10; /db/vtree/library/trees/ipc/ipc-d23x/src/
 */

