/*
 *  Copyright 2010 by Texas Instruments Incorporated.
 *
 */

/*
 *  Copyright 2009 by Texas Instruments Incorporated.
 *
 */

/**
 *   @file  test_osal.c
 *
 *   @brief   
 *      This is the OS abstraction layer and is used by the the QMSS LLD.
 *
 *  \par
 *  ============================================================================
 *  @n   (C) Copyright 2009, Texas Instruments, Inc.
 *  @n   Use of this software is controlled by the terms and conditions found 
 *  @n   in the license agreement under which this software has been supplied.
 *  ============================================================================
 *  \par
 */
#include <xdc/std.h>
#include <xdc/runtime/IHeap.h>
#include <xdc/runtime/System.h>
#include <xdc/runtime/Memory.h>
#include <xdc/runtime/Error.h>

#include <ti/sysbios/hal/Hwi.h>
#include <ti/ipc/GateMP.h>

#include <xdc/cfg/global.h>


/**********************************************************************
 ************************** Global Variables **************************
 **********************************************************************/
UInt32 qmssMallocCounter = 0;
UInt32 qmssFreeCounter   = 0;

/**********************************************************************
 *************************** OSAL Functions **************************
 **********************************************************************/

/**
 *  @b Description
 *  @n  
 *      The function is used to allocate a memory block of the specified size.
 *
 *  @param[in]  num_bytes
 *      Number of bytes to be allocated.
 *
 *  @retval
 *      Allocated block address
 */
Ptr Osal_biosMalloc (UInt32 num_bytes)
{
    Error_Block eb;

    Error_init(&eb);

    /* Increment the allocation counter. */
    qmssMallocCounter++;        

    /* Allocate memory. */
    return Memory_alloc(NULL, num_bytes, 0, &eb);
}

/**
 *  @b Description
 *  @n  
 *      The function is used to free a memory block of the specified size.
 *
 *  @param[in]  ptr
 *      Pointer to the memory block to be cleaned up.
 *
 *  @param[in]  size
 *      Size of the memory block to be cleaned up.
 *
 *  @retval
 *      Not Applicable
 */
Void Osal_biosFree (Ptr ptr, UInt32 size)
{
    /* Increment the free counter. */
    qmssFreeCounter++;  
    
    Memory_free(NULL, ptr, size);
}

/**
 *  @b Description
 *  @n  
 *      The function is used to enter a critical section.
 *      Function protects against 
 *      
 *      access from multiple cores 
 *      and 
 *      access from multiple threads on single core
 *
 *  @retval
 *      Handle used to lock critical section
 */
Ptr Osal_qmssCsEnter (Void)
{
    IArg key = 0;
    
    if (GateMP_getDefaultRemote()) {
        key = GateMP_enter(GateMP_getDefaultRemote());
    }

    return ((Ptr)key);

}

/**
 *  @b Description
 *  @n  
 *      The function is used to exit a critical section 
 *      protected using Osal_qmssCsEnter() API.
 *
 *  @param[in]  CsHandle
 *      Handle for unlocking critical section.
 *
 *  @retval
 *      Not Applicable
 */
Void Osal_qmssCsExit (Ptr CsHandle)
{
    if (GateMP_getDefaultRemote()) {
        GateMP_leave(GateMP_getDefaultRemote(), (IArg)CsHandle);
    }

    return;
}
/**
 *  @b Description
 *  @n  
 *      The function is used to enter a critical section.
 *      Function protects against 
 *      access from multiple threads on single core
 *
 *  @retval
 *      Handle used to lock critical section
 */
Ptr Osal_qmssMtCsEnter (Void)
{
    IArg key;
    
    key = Hwi_disable();

    return ((Ptr)key);
}

/**
 *  @b Description
 *  @n  
 *      The function is used to exit a critical section
 *      protected using Osal_qmssCsEnter() API.
 *
 *  @param[in]  CsHandle
 *      Handle for unlocking critical section.
 *
 *  @retval
 *      Not Applicable
 */
Void Osal_qmssMtCsExit (Ptr CsHandle)
{
    Hwi_restore((UInt)CsHandle);

    return;
}
/*
 *  @(#) ti.sdo.ipc.family.tci6498.osal; 1, 0, 0, 0,71; 4-9-2010 18:38:14; /db/vtree/library/trees/ipc/ipc-d23x/src/
 */

