/* 
 * Copyright (c) 2010
 * Texas Instruments
 *
 *  All rights reserved.  Property of Texas Instruments
 *  Restricted rights to use, duplicate or disclose this code are
 *  granted through contract.
 * 
 * */
/*
 *  ======== GateAAMonitor.c ========
 */

#include <xdc/std.h>
#include <xdc/runtime/Assert.h>
#include <xdc/runtime/Error.h>
#include <xdc/runtime/IGateProvider.h>
#include <xdc/runtime/Gate.h>
#include <xdc/runtime/Log.h>
#include <ti/sysbios/family/c64p/Cache.h>

#include <ti/sdo/ipc/interfaces/IGateMPSupport.h>

#include "package/internal/GateAAMonitor.xdc.h"  

#include <ti/sdo/ipc/_Ipc.h>
#include <ti/sdo/ipc/_GateMP.h>
#include <ti/sdo/ipc/_SharedRegion.h>

/*
 *************************************************************************
 *                       Instance functions
 *************************************************************************
 */

/*
 *  ======== GateAAMonitor_Instance_init ========
 */
Int GateAAMonitor_Instance_init(GateAAMonitor_Object *obj,
                                IGateMPSupport_LocalProtect localProtect,
                                const GateAAMonitor_Params *params,
                                Error_Block *eb)
{    
    /* Assert that params->sharedAddr is valid */
    Assert_isTrue(
        (UInt32)params->sharedAddr >= GateAAMonitor_SL2_RANGE_BASE &&
        (UInt32)params->sharedAddr <  GateAAMonitor_SL2_RANGE_MAX,
        GateAAMonitor_A_invSharedAddr);

    /* Assert that params->sharedAddr is cache aligned */
    Assert_isTrue((UInt32)params->sharedAddr % 
                  GateAAMonitor_CACHELINE_SIZE == 0,
                  Ipc_A_addrNotCacheAligned);

    /* Create the local gate */
    obj->localGate = 
        GateMP_createLocal((ti_sdo_ipc_GateMP_LocalProtect)localProtect);
    obj->sharedAddr = params->sharedAddr;
    obj->nested     = 0;
               
    if (!params->openFlag) {
        /* Creating */
        GateAAMonitor_postInit(obj);    
    }
    else {
        /* Opening. */
        Cache_inv((Ptr)obj->sharedAddr, GateAAMonitor_CACHELINE_SIZE, 
                  Cache_Type_ALL, TRUE);
    }
    
    return (0);
}

/*
 *  ======== GateAAMonitor_enter ========
 */
IArg GateAAMonitor_enter(GateAAMonitor_Object *obj)
{
    IArg key;
        
    key = IGateProvider_enter(obj->localGate);
    
    /* If the gate object has already been entered, return the nested value */
    obj->nested++;
    if (obj->nested > 1) {
        return (key);
    }
    
    /* Enter the monitor */
    GateAAMonitor_getLock((Ptr)obj->sharedAddr);
    
    return (key);
}

/*
 *  ======== GateAAMonitor_leave ========
 *  The sharedAddr must come from SL2 (shared L2 memory) so we only
 *  have to freeze L1D.
 */
Void GateAAMonitor_leave(GateAAMonitor_Object *obj, IArg key)
{
    Cache_Mode mode;

    obj->nested--;
    
    if (obj->nested == 0) {
        mode = Cache_setMode(Cache_Type_L1D, Cache_Mode_FREEZE);

        /* Leave the critical region by setting address value to zero */
        *(obj->sharedAddr) = 0;

        Cache_setMode(Cache_Type_L1D, mode);
    }
 
    IGateProvider_leave(obj->localGate, key);
}

/*
 *************************************************************************
 *                       Module functions
 *************************************************************************
 */

/*
 *  ======== GatePeterson_sharedMemReq ========
 */
SizeT GateAAMonitor_sharedMemReq(const IGateMPSupport_Params *params)
{
    SizeT memReq;

    memReq = (SizeT)_Ipc_roundup(GateAAMonitor_CACHELINE_SIZE,
                                 SharedRegion_cacheLineSize);

    return(memReq);
}

/*
 *  ======== GateAAMonitor_query ========
 */
Bool GateAAMonitor_query(Int qual)
{
    Bool rc;

    switch (qual) {
        case IGateProvider_Q_BLOCKING:
            /* Depends on gate proxy? */
            rc = FALSE;
            break;

        case IGateProvider_Q_PREEMPTING:
            /* Depends on gate proxy? */
            rc = TRUE;
            break;
        default:
            rc = FALSE;
            break;
    }
    
    return (rc);
}

/*
 *************************************************************************
 *                       Internal functions
 *************************************************************************
 */
/*
 *  ======== GateAAMonitor_postInit ========
 *  Function to be called during
 *  1. module startup to complete the initialization of all static instances
 *  2. instance_init to complete the initialization of a dynamic instance
 *
 *  Main purpose is to initialize Atomic Access Monitor
 */
Void GateAAMonitor_postInit(GateAAMonitor_Object *obj)
{
    /*
     *  The processor that inits the AAM initializes the value 
     *  to zero (e.g. no one is using it). The other processors 
     *  must invalidate the memory in case it is in cache.
     */
    *(obj->sharedAddr) = 0;
    Cache_wbInv((Ptr)obj->sharedAddr, GateAAMonitor_CACHELINE_SIZE, 
                Cache_Type_ALL, TRUE);
}

/*
 */
/*
 *  @(#) ti.sdo.ipc.gates; 1, 0, 0, 0,142; 4-9-2010 18:38:19; /db/vtree/library/trees/ipc/ipc-d23x/src/
 */

