/* 
 * Copyright (c) 2010
 * Texas Instruments
 *
 *  All rights reserved.  Property of Texas Instruments
 *  Restricted rights to use, duplicate or disclose this code are
 *  granted through contract.
 * 
 * */
/*
 *  ======== GateHWSem.c ========
 */

#include <xdc/std.h>
#include <xdc/runtime/Error.h>
#include <xdc/runtime/Assert.h>
#include <xdc/runtime/IGateProvider.h>
#include <xdc/runtime/Log.h>
#include <xdc/runtime/Startup.h>

#include <ti/sdo/ipc/interfaces/IGateMPSupport.h>

#include <ti/sdo/ipc/_Ipc.h>
#include <ti/sdo/ipc/_GateMP.h>
#include "package/internal/GateHWSem.xdc.h"  


/*
 *************************************************************************
 *                       Module functions
 *************************************************************************
 */

/*
 *  ======== GateHWSem_Module_startup ========
 *  release the HW semaphore if the core owns it.
 */
Int GateHWSem_Module_startup(Int phase)
{
    Int i;
    volatile UInt32 *baseAddr = (volatile UInt32 *)GateHWSem_baseAddr;

    /* releases the HW semaphore if the core owns it otherwise its a nop */
    for (i = 0; i < GateHWSem_numSems; i++) {
        baseAddr[i] = 1;
    }

    return (Startup_DONE);
}

/*
 *************************************************************************
 *                       Instance functions
 *************************************************************************
 */

/*
 *  ======== GateHWSem_Instance_init ========
 */
Int GateHWSem_Instance_init(GateHWSem_Object *obj,
                            IGateMPSupport_LocalProtect localProtect,
                            const GateHWSem_Params *params,
                            Error_Block *eb)
{   
    /* Assert that params->resourceId is valid */
    Assert_isTrue(params->resourceId < GateHWSem_numSems,
                  GateHWSem_A_invSemNum);

   /* Create the local gate */
    obj->localGate  = 
        GateMP_createLocal((ti_sdo_ipc_GateMP_LocalProtect)localProtect); 
    obj->semNum     = params->resourceId;
    obj->nested     = 0;    
    
    /* Check for dynamic open */           
    if (!params->openFlag) {        
        /* Initialize the HW Semaphore */
        GateHWSem_postInit(obj);
    }    
    
    return (0);
}

/*
 *  ======== GateHWSem_Instance_finalize ========
 */
Void GateHWSem_Instance_finalize(GateHWSem_Object *obj, Int status)
{      
    /* Reset HW Semaphore */
    GateHWSem_postInit(obj);
}

/*
 *  ======== GateHWSem_enter ========
 */
IArg GateHWSem_enter(GateHWSem_Object *obj)
{
    IArg key;
    volatile UInt32 *baseAddr = (volatile UInt32 *)GateHWSem_baseAddr;
    
    key = IGateProvider_enter(obj->localGate);
    
    /* If the gate object has already been entered, return the nested value */
    obj->nested++;
    if (obj->nested > 1) {
        return (key);
    }
    
    /* Enter the hardware lock */
    while (baseAddr[obj->semNum] != 1) {
    }

    return (key);
}

/*
 *  ======== GateHWSem_leave ========
 */
Void GateHWSem_leave(GateHWSem_Object *obj, IArg key)
{
    volatile UInt32 *baseAddr = (volatile UInt32 *)GateHWSem_baseAddr;
    
    obj->nested--;
    
    if (obj->nested == 0) {
        /* Leave the critical region by setting address value to 1 */
        baseAddr[obj->semNum] = 1;
    }
    
    IGateProvider_leave(obj->localGate, key);
}

/*
 *  ======== GateHWSem_getResourceId ========
 */
Bits32 GateHWSem_getResourceId(GateHWSem_Object *obj)
{    
    return (obj->semNum);
}

/*
 *************************************************************************
 *                       Module functions
 *************************************************************************
 */
 
/*
 *  ======== GateHWSem_query ========
 */
Bool GateHWSem_query(Int qual)
{
    Bool rc;

    switch (qual) {
        case IGateProvider_Q_BLOCKING:
            /* Depends on gate proxy? */
            rc = TRUE;
            break;

        case IGateProvider_Q_PREEMPTING:
            /* Depends on gate proxy? */
            rc = TRUE;
            break;
        default:
            rc = FALSE;
            break;
    }
    return (rc);
}

/*
 *  ======== GateHWSem_sharedMemReq ========
 */
SizeT GateHWSem_sharedMemReq(const IGateMPSupport_Params *params)
{
    /* No shared memory needed */
    return (0);
}


/*
 *************************************************************************
 *                       Internal functions
 *************************************************************************
 */
/*
 *  ======== GateHWSem_postInit ========
 *  Function to be called during
 *  1. module startup to complete the initialization of all static instances
 *  2. instance_init to complete the initialization of a dynamic instance
 *
 *  Main purpose is to initialize hardware semaphroe
 */
Void GateHWSem_postInit(GateHWSem_Object *obj)
{
    volatile UInt32 *baseAddr = (volatile UInt32 *)GateHWSem_baseAddr;
    
    /* Reset the hardware semaphore */
    baseAddr[obj->semNum] = 1;
}
/*
 */
/*
 *  @(#) ti.sdo.ipc.gates; 1, 0, 0, 0,142; 4-9-2010 18:38:19; /db/vtree/library/trees/ipc/ipc-d23x/src/
 */

