/* 
 * Copyright (c) 2010
 * Texas Instruments
 *
 *  All rights reserved.  Property of Texas Instruments
 *  Restricted rights to use, duplicate or disclose this code are
 *  granted through contract.
 * 
 * */
/*
 *  ======== GateHWSpinlock.c ========
 */

#include <xdc/std.h>
#include <xdc/runtime/Error.h>
#include <xdc/runtime/Assert.h>
#include <xdc/runtime/IGateProvider.h>
#include <xdc/runtime/Log.h>

#include <ti/sdo/ipc/Ipc.h>
#include <ti/sdo/ipc/interfaces/IGateMPSupport.h>

#include "package/internal/GateHWSpinlock.xdc.h"  

#include <ti/sdo/ipc/_GateMP.h>
/*
 *************************************************************************
 *                       Instance functions
 *************************************************************************
 */

/*
 *  ======== GateHWSpinlock_Instance_init ========
 */
Int GateHWSpinlock_Instance_init(GateHWSpinlock_Object *obj,
                            IGateMPSupport_LocalProtect localProtect,
                            const GateHWSpinlock_Params *params,
                            Error_Block *eb)
{   
    /* Assert that params->resourceId is valid */
    Assert_isTrue(params->resourceId < GateHWSpinlock_numLocks,
                  GateHWSpinlock_A_invSpinLockNum);
    
    /* Create the local gate */
    obj->localGate = 
        GateMP_createLocal((ti_sdo_ipc_GateMP_LocalProtect)localProtect);
    if (obj->localGate == NULL) {
        return (1);
    }    
    
    obj->lockNum = params->resourceId;
    obj->nested   = 0;    
    
    /* Check for dynamic open */           
    if (params->openFlag) {        
        Log_write1(GateHWSpinlock_LM_open, (UArg)obj->lockNum);
    }
    else {        
        Log_write1(GateHWSpinlock_LM_create, (UArg)obj->lockNum);
    }    
    
    return (0);
}

/*
 *  ======== GateHWSpinlock_enter ========
 */
IArg GateHWSpinlock_enter(GateHWSpinlock_Object *obj)
{
    volatile UInt32 *baseAddr = (volatile UInt32 *)GateHWSpinlock_baseAddr;
    IArg key;
    
    key = IGateProvider_enter(obj->localGate);
    
    /* If the gate object has already been entered, return the nested value */
    obj->nested++;
    if (obj->nested > 1) {
        return (key);
    }
    
    /* Enter the spinlock */
    while (1) {
        if (baseAddr[obj->lockNum] == 0) {
            break;
        }
    }
    
    Log_write2(GateHWSpinlock_LM_enter,(UArg)obj->lockNum, key);

    return (key);
}

/*
 *  ======== GateHWSpinlock_leave ========
 */
Void GateHWSpinlock_leave(GateHWSpinlock_Object *obj, IArg key)
{
    volatile UInt32 *baseAddr = (volatile UInt32 *)GateHWSpinlock_baseAddr;

    obj->nested--;
    
    /* Leave the spinlock if the leave() is not nested */
    if (obj->nested == 0) {
        baseAddr[obj->lockNum] = 0;
    }
    
    IGateProvider_leave(obj->localGate, key);
    
    Log_write2(GateHWSpinlock_LM_leave, (UArg)obj->lockNum, key);
}

/*
 *  ======== GateHWSpinlock_getResourceId ========
 */
Bits32 GateHWSpinlock_getResourceId(GateHWSpinlock_Object *obj)
{    
    return (obj->lockNum);
}

/*
 *  ======== GateHWSpinlock_sharedMemReq ========
 */
SizeT GateHWSpinlock_sharedMemReq(const IGateMPSupport_Params *params)
{
    return (0);
}

/*
 *************************************************************************
 *                       Module functions
 *************************************************************************
 */
 
/*
 *  ======== GateHWSpinlock_query ========
 */
Bool GateHWSpinlock_query(Int qual)
{
    Bool rc;

    switch (qual) {
        case IGateProvider_Q_BLOCKING:
            /* Depends on gate proxy? */
            rc = FALSE;
            break;

        case IGateProvider_Q_PREEMPTING:
            /* Depends on gate proxy? */
            rc = TRUE;
            break;
        default:
            rc = FALSE;
            break;
    }
    return (rc);
}

/*
 */
/*
 *  @(#) ti.sdo.ipc.gates; 1, 0, 0, 0,142; 4-9-2010 18:38:19; /db/vtree/library/trees/ipc/ipc-d23x/src/
 */

