/* 
 * Copyright (c) 2010
 * Texas Instruments
 *
 *  All rights reserved.  Property of Texas Instruments
 *  Restricted rights to use, duplicate or disclose this code are
 *  granted through contract.
 * 
 * */
/*
 *  ======== INotifyDriver.xdc ========
 *  
 */

/*!
 *  ======== INotifyDriver ========
 *  Notify driver interface
 *
 *  Interface implemented by all drivers for the notify module.  Modules that
 *  implement this interface expect the eventId arguments to be valid.
 */

interface INotifyDriver {
    
instance:

    /*!
     *  ======== registerEvent ========
     *  Register a callback to an event
     *
     *  This driver function is called by the Notify_registerEvent function 
     *  within the Notify module gate. Refer to its documentation for more 
     *  details.  
     *
     *  @param(eventId)      Number of event that is being registered
     */
    Void registerEvent(UInt32 eventId);

    /*!
     *  ======== unregisterEvent ========
     *  Remove an event listener from an event
     *
     *  This driver function is called by the Notify_unregisterEvent function
     *  within the Notify module gate. Refer to it for more details.
     *
     *  @param(eventId)      Number of event that is being unregistered
     */
    Void unregisterEvent(UInt32 eventId);

    /*!
     *  ======== sendEvent ========
     *  Send a signal to an event
     *
     *  This interface function is called by the Notify_sendEvent function.
     *  Notify_sendEvent does not provide any context protection for
     *  INotifyDriver_sendEvent, so appropriate measures must be taken within
     *  the driver itself.
     *
     *  @param(eventId)      Number of event to signal
     *  @param(payload)      Payload (optional) to pass to callback function
     *  @param(waitClear)    If TRUE, have the NotifyDriver wait for
     *                       acknowledgement back from the destination
     *                       processor.
     *
     *  @b(returns)          Notify status
     */
    Int sendEvent(UInt32 eventId, UInt32 payload, Bool waitClear);

    /*!
     *  ======== disable ========
     *  Disable a NotifyDriver instance
     *
     *  Disables the ability of a Notify driver to receive events for a given
     *  processor. This interface function is called by the Notify_disable
     *  function. Refer to its documentation for more details.
     */
    Void disable();

    /*!
     *  ======== enable ========
     *  Enable a NotifyDriver instance
     *
     *  Enables the ability of a Notify driver to receive events for a given
     *  processor. This interface function is called by the Notify_restore 
     *  function. Refer to its documentation for more details.
     */
    Void enable();

    /*!
     *  ======== disableEvent ========
     *  Disable an event
     *
     *  This interface function is called by the Notify_disableEvent function. 
     *  Refer to its documentation for more details.
     *
     *  The Notify module does validation of the eventId.  The Notify module
     *  enters calls this function within the Notify module gate.
     *
     *  @param(eventId)      Number of event to disable
     */
    Void disableEvent(UInt32 eventId);

    /*!
     *  ======== enableEvent ========
     *  Enable an event
     *
     *  This interface function is called by the Notify_disableEvent function. 
     *  Refer to its documentation for more details.
     *
     *  The Notify module does validation of the eventId.  The Notify module
     *  enters calls this function within the Notify module gate.
     *
     *  @param(eventId)      Number of event to enable
     */
    Void enableEvent(UInt32 eventId);
    
    /*! @_nodoc
     *  ======== setNotifyHandle ========
     *  Called during Notify instance creation to 'send' its handle to its
     *  corresponding notify driver instance
     */
    Void setNotifyHandle(Ptr driverHandle);
    
}
/*
 *  @(#) ti.sdo.ipc.interfaces; 1, 0, 0, 0,298; 4-9-2010 18:38:25; /db/vtree/library/trees/ipc/ipc-d23x/src/
 */

