/* 
 * Copyright (c) 2010
 * Texas Instruments
 *
 *  All rights reserved.  Property of Texas Instruments
 *  Restricted rights to use, duplicate or disclose this code are
 *  granted through contract.
 * 
 * */
/*
 *  ======== TransportShm.c ========
 */

#include <xdc/std.h>
#include <xdc/runtime/Assert.h>
#include <xdc/runtime/Error.h>
#include <xdc/runtime/Gate.h>
#include <xdc/runtime/Memory.h>

#include <ti/sysbios/hal/Cache.h>
#include <ti/sysbios/knl/Swi.h>

#include "package/internal/TransportShm.xdc.h"

#include <ti/sdo/ipc/_Ipc.h>
#include <ti/sdo/utils/_MultiProc.h>
#include <ti/sdo/ipc/_SharedRegion.h>
#include <ti/sdo/ipc/_Notify.h>
#include <ti/sdo/ipc/_GateMP.h>
#include <ti/sdo/ipc/_ListMP.h>
#include <ti/sdo/ipc/_MessageQ.h>

/*
 *  ======== TransportShm_notifyFxn ========
 */
Void TransportShm_notifyFxn(UInt16 procId,
                            UInt16 lineId,
                            UInt32 eventId, 
                            UArg arg, 
                            UInt32 payload)
{
    Swi_Handle swiHandle;

    /* Swi_Handle was passed as arg in register */
    swiHandle = (Swi_Handle)arg;

    /* post the Swi */
    Swi_post(swiHandle);
}

/*
 *  ======== TransportShm_swiFxn ========
 */
Void TransportShm_swiFxn(UArg arg)
{
    UInt32 queueId;
    TransportShm_Object *obj = (TransportShm_Object *)arg;
    MessageQ_Msg msg = NULL;
    
    /*
     *  While there are messages, get them out and send them to
     *  their final destination.
     */
    msg = (MessageQ_Msg)ListMP_getHead(obj->localList);
    while (msg != NULL) {
        /* Get the destination message queue Id */
        queueId = MessageQ_getDstQueue(msg);
        
        /* put the message to the destination queue */
        MessageQ_put(queueId, msg);
        
        /* check to see if there are more messages */
        msg = (MessageQ_Msg)ListMP_getHead(obj->localList);
    }
}

/*
 *************************************************************************
 *                       Instance functions
 *************************************************************************
 */

/*
 *  ======== TransportShm_Instance_init ========
 */
Int TransportShm_Instance_init(TransportShm_Object *obj,
                                   UInt16 procId,
                                   const TransportShm_Params *params,
                                   Error_Block *eb)
{
    Int            localIndex;
    Int            remoteIndex;    
    Int            status;
    Bool           flag;
    UInt32         minAlign;
    ListMP_Params  listMPParams[2];
    Swi_Handle     swiHandle;
    Swi_Params     swiParams;    
    Ptr            localAddr;

    swiHandle = TransportShm_Instance_State_swiObj(obj);
    
    /* 
     *  Determine who gets the '0' slot in shared memory and who gets
     *  the '1' slot. The '0' slot is given to the lower MultiProc id.
     */
    if (MultiProc_self() < procId) {
        localIndex  = 0;
        remoteIndex = 1;
    }
    else {
        localIndex  = 1;
        remoteIndex = 0;
    }
    
    if (params->openFlag) {
        /* Open by sharedAddr */
        obj->objType = Ipc_ObjType_OPENDYNAMIC;
        obj->self = (TransportShm_Attrs *)params->sharedAddr;
        obj->regionId = SharedRegion_getId(params->sharedAddr);
        obj->cacheEnabled = SharedRegion_isCacheEnabled(obj->regionId);
    
        localAddr = SharedRegion_getPtr(obj->self->gateMPAddr);
        status = GateMP_openByAddr(localAddr, (GateMP_Handle *)&obj->gate);
        if (status != GateMP_S_SUCCESS) {
            return(1);
        }
    }
    else {
        /* init the gate for ListMP create below */
        if (params->gate != NULL) {
            obj->gate = params->gate;
        }
        else {
            obj->gate = (ti_sdo_ipc_GateMP_Handle)GateMP_getDefaultRemote();
        }
    
        /* Creating using sharedAddr */
        obj->regionId = SharedRegion_getId(params->sharedAddr);
            
        /* Assert that the buffer is in a valid shared region */
        Assert_isTrue(obj->regionId != SharedRegion_INVALIDREGIONID, 
                      Ipc_A_addrNotInSharedRegion);
                        
        /* Assert that sharedAddr is cache aligned */
        Assert_isTrue(((UInt32)params->sharedAddr % 
                      SharedRegion_getCacheLineSize(obj->regionId) == 0),
                      Ipc_A_addrNotCacheAligned);
                
        /* set object's cacheEnabled, type, self */
        obj->cacheEnabled = SharedRegion_isCacheEnabled(obj->regionId);
        obj->objType = Ipc_ObjType_CREATEDYNAMIC;
        obj->self = (TransportShm_Attrs *)params->sharedAddr;
    }
    
    /* determine the minimum alignment to align to */
    minAlign = Memory_getMaxDefaultTypeAlign();
    if (SharedRegion_getCacheLineSize(obj->regionId) > minAlign) {
        minAlign = SharedRegion_getCacheLineSize(obj->regionId);
    }
    
    /* 
     *  Carve up the shared memory.
     *  If cache is enabled, these need to be on separate cache lines.
     *  This is done with minAlign and _Ipc_roundup function.
     */
    obj->other = (TransportShm_Attrs *)((UInt32)(obj->self) +
        (_Ipc_roundup(sizeof(TransportShm_Attrs), minAlign)));
                                                   
    ListMP_Params_init(&(listMPParams[0]));
    listMPParams[0].gate = obj->gate;
    listMPParams[0].sharedAddr = (UInt32 *)((UInt32)(obj->other) +
        (_Ipc_roundup(sizeof(TransportShm_Attrs), minAlign)));
    
    ListMP_Params_init(&listMPParams[1]);
    listMPParams[1].gate = obj->gate;
    listMPParams[1].sharedAddr = (UInt32 *)((UInt32)(listMPParams[0].sharedAddr)
        + ListMP_sharedMemReq(&listMPParams[0]));
    
    obj->notifyEventId = params->notifyEventId;
    obj->priority      = params->priority;
    obj->remoteProcId  = procId;
    
    Swi_Params_init(&swiParams);
    swiParams.arg0 = (UArg)obj;
    Swi_construct(Swi_struct(swiHandle),
                 (Swi_FuncPtr)TransportShm_swiFxn,
                 &swiParams, eb);
             
    if (params->openFlag == FALSE) {   
        obj->localList = ListMP_create(&(listMPParams[localIndex]));
        if (obj->localList == NULL) {
            return (2);
        }
        
        obj->remoteList = ListMP_create(&(listMPParams[remoteIndex]));
        if (obj->localList == NULL) {
            return (2);
        }
    }
    else {
        /* Open the local ListMP instance */
        status = ListMP_openByAddr(listMPParams[localIndex].sharedAddr,
                                   &(obj->localList));

        if (status != ListMP_S_SUCCESS) {
            return (2);
        }

        /* Open the remote ListMP instance */
        status = ListMP_openByAddr(listMPParams[remoteIndex].sharedAddr,
                                   &(obj->remoteList));
        
        if (status != ListMP_S_SUCCESS) {
            return (2);
        }
    }

    /* register the event with Notify */
    status = Notify_registerEventSingle(
                 procId,                /* remoteProcId */
                 0,                     /* lineId */
                 params->notifyEventId, /* notify event id */
                 (Notify_FnNotifyCbck)TransportShm_notifyFxn,
                 (UArg)swiHandle);
    
    if (status < 0) {
        return (3);
    }

    /* Register the transport with MessageQ */
    flag = MessageQ_registerTransport(TransportShm_Handle_upCast(obj),
                                      procId, params->priority);
    if (flag == FALSE) {
        return (4);
    }
    
    if (params->openFlag == FALSE) {
        obj->self->creatorProcId = MultiProc_self();
        obj->self->notifyEventId = obj->notifyEventId;
        obj->self->priority = obj->priority;
        
        /* Store the GateMP sharedAddr in the Attrs */
        obj->self->gateMPAddr =
            GateMP_getSharedAddr(obj->gate);
        obj->self->flag = TransportShm_UP;
        
        if (obj->cacheEnabled) {
            Cache_wbInv(obj->self, sizeof(TransportShm_Attrs),
                     Cache_Type_ALL, TRUE);
        }
    }
    else {
        obj->other->flag = TransportShm_UP;
        if (obj->cacheEnabled) {
            Cache_wbInv(&(obj->other->flag), minAlign, Cache_Type_ALL, TRUE);
        } 
    }
    
    obj->status = TransportShm_UP;

    return (0);
}

/*
 *  ======== TransportShm_Instance_finalize ========
 */
Void TransportShm_Instance_finalize(TransportShm_Object* obj, Int status)
{
    Swi_Handle     swiHandle;
    
    if (obj->objType == Ipc_ObjType_CREATEDYNAMIC) {
        /* clear the self flag */
        obj->self->flag = 0;
        
        if (obj->cacheEnabled) {
            Cache_wbInv(&(obj->self->flag),
                SharedRegion_getCacheLineSize(obj->regionId), Cache_Type_ALL, 
                    TRUE);
        }
        
        if (obj->localList != NULL) {
            ListMP_delete(&(obj->localList));
        }
        
        if (obj->remoteList != NULL) {
            ListMP_delete(&(obj->remoteList));
        }
    }
    else {
        /* other flag was set by remote proc */
        obj->other->flag = 0;
    
        if (obj->cacheEnabled) {
            Cache_wbInv(&(obj->other->flag),
                SharedRegion_getCacheLineSize(obj->regionId), Cache_Type_ALL, 
                    TRUE);
        }
        
        if (obj->gate != NULL) {
            GateMP_close((GateMP_Handle *)&(obj->gate));
        }
        
        if (obj->localList != NULL) {
            ListMP_close(&(obj->localList));
        }
        
        if (obj->remoteList != NULL) {
            ListMP_close(&(obj->remoteList));
        }
    }
    
    switch(status) {
        case 0:
            /* MessageQ_registerTransport succeeded */
            MessageQ_unregisterTransport(obj->remoteProcId, obj->priority);
        case 1: /* GateMP open failed */
        case 2: /* ListMP create/open failed */
        case 3: /* Notify_registerEventSingle failed */
        case 4: /* MessageQ_registerTransport failed */
            Notify_unregisterEventSingle(
                obj->remoteProcId,
                0,
                obj->notifyEventId);
            break;
    }
    
    /* Destruct the swi */
    swiHandle = TransportShm_Instance_State_swiObj(obj);
    if (swiHandle != NULL) {
        Swi_destruct(Swi_struct(swiHandle));
    }
            
}

/*
 *  ======== TransportShm_put ========
 *  Assuming MessageQ_put is making sure that the arguments are ok
 */
Bool TransportShm_put(TransportShm_Object *obj, Ptr msg)
{
    Int32 status;
    Bool retval = TRUE;
    IArg key;

    /* writeback invalidate the message */
    if (obj->cacheEnabled) {
        Cache_wbInv((Ptr)msg, ((MessageQ_Msg)(msg))->msgSize, Cache_Type_ALL,
                    TRUE);
    }
    
    /* make sure ListMP_put and sendEvent are done before remote executes */
    key = GateMP_enter(obj->gate);

    /* Put the message on the remoteList */
    ListMP_putTail(obj->remoteList, (ListMP_Elem *)msg);

    /* Notify the remote processor */
    status = Notify_sendEvent(
                 obj->remoteProcId, 
                 0,
                 obj->notifyEventId,
                 0,
                 FALSE);

    /* check the status of the sendEvent */
    if (status < 0) {
        /* remove the message from the List and return 'FALSE' */
        ListMP_remove(obj->remoteList, (ListMP_Elem *)msg);
        retval = FALSE;
    }

    /* leave the gate */
    GateMP_leave(obj->gate, key);

    return (retval);
}

/*
 *  ======== TransportShm_control ======== 
 */
Bool TransportShm_control(TransportShm_Object *obj, UInt cmd, UArg cmdArg) 
{
    return (FALSE);
}

/*
 *  ======== TransportShm_getStatus ======== 
 */
Int TransportShm_getStatus(TransportShm_Object *obj)
{
    return (obj->status);
}

/*
 *************************************************************************
 *                      Module functions
 *************************************************************************
 */

/*
 *  ======== TransportShm_close ========
 */
Void TransportShm_close(TransportShm_Handle *handle)
{
    TransportShm_delete(handle);
}
 
/*
 *  ======== TransportShm_openByAddr ========
 */
Int TransportShm_openByAddr(Ptr sharedAddr,
                      TransportShm_Handle *handlePtr,
                      Error_Block *eb)           
{
    TransportShm_Params params;
    TransportShm_Attrs *attrs;
    Int status;
    UInt16 id;
    
    TransportShm_Params_init(&params);
    
    /* Tell Instance_init() that we're opening */
    params.openFlag = TRUE;
    
    attrs = (TransportShm_Attrs *)sharedAddr;
    id = SharedRegion_getId(sharedAddr);

    /* Assert that the region is valid */
    Assert_isTrue(id != SharedRegion_INVALIDREGIONID,
        Ipc_A_addrNotInSharedRegion);
    
    /* invalidate the attrs before using it */
    if (SharedRegion_isCacheEnabled(id)) {
        Cache_inv(attrs, sizeof(TransportShm_Attrs), Cache_Type_ALL, TRUE);
    }
    
    /* set params field */
    params.sharedAddr    = sharedAddr;
    params.notifyEventId = attrs->notifyEventId;
    params.priority      = attrs->priority;
    
    if (attrs->flag != TransportShm_UP) {
        /* make sure transport is up */
        *handlePtr = NULL;
        status = MessageQ_E_NOTFOUND;
    }
    else {
        /* Create the object */
        *handlePtr = TransportShm_create(attrs->creatorProcId, &params, eb);
        if (*handlePtr == NULL) {
            status = MessageQ_E_FAIL;
        }
        else {
            status = MessageQ_S_SUCCESS;
        }
    }
    
    return (status);
}

/*
 *  ======== TransportShm_sharedMemReq ========
 */
SizeT TransportShm_sharedMemReq(const TransportShm_Params *params)
{
    SizeT memReq, minAlign;
    UInt16 regionId;
    ListMP_Params listMPParams;

    regionId = SharedRegion_getId(params->sharedAddr);
    
    minAlign = Memory_getMaxDefaultTypeAlign();
    if (SharedRegion_getCacheLineSize(regionId) > minAlign) {
        minAlign = SharedRegion_getCacheLineSize(regionId);
    }

    /* for the Attrs structure */
    memReq = _Ipc_roundup(sizeof(TransportShm_Attrs), minAlign);
    
    /* for the second Attrs structure */
    memReq += _Ipc_roundup(sizeof(TransportShm_Attrs), minAlign);
    
    ListMP_Params_init(&listMPParams);
    listMPParams.regionId = regionId;
    
    /* for localListMP */
    memReq += ListMP_sharedMemReq(&listMPParams);
    
    /* for remoteListMP */
    memReq += ListMP_sharedMemReq(&listMPParams);
    
    return(memReq);
}

/*
 *  ======== TransportShm_setErrFxn ======== 
 */
Void TransportShm_setErrFxn(TransportShm_ErrFxn errFxn)
{
    UArg key;
    
    key = Gate_enterSystem();
    
    TransportShm_module->errFxn = errFxn;
    
    Gate_leaveSystem(key);
}

/*
 */
/*
 *  @(#) ti.sdo.ipc.transports; 1, 0, 0, 0,109; 4-9-2010 18:38:35; /db/vtree/library/trees/ipc/ipc-d23x/src/
 */

