/* 
 * Copyright (c) 2010
 * Texas Instruments
 *
 *  All rights reserved.  Property of Texas Instruments
 *  Restricted rights to use, duplicate or disclose this code are
 *  granted through contract.
 * 
 * */
/*
 *  ======== TransportShm.xdc ========
 *
 */
 
import xdc.runtime.Error;
import ti.sdo.ipc.Ipc;
import ti.sdo.ipc.GateMP;
import ti.sdo.ipc.ListMP;
import ti.sdo.ipc.SharedRegion;
import ti.sysbios.knl.Swi;

/*!
 *  ======== TransportShm ========
 *  Transport for MessageQ that acts on shared memory.
 *
 */
@InstanceFinalize
@InstanceInitError

module TransportShm inherits ti.sdo.ipc.interfaces.IMessageQTransport 
{   
    /*! @_nodoc
     *  ======== openByAddr ========
     *  Open a created TransportShm instance by address
     *
     *  Just like {@link #open}, openByAddr returns a handle to a created
     *  TransportShm instance.  This function is used to open a
     *  TransportShm using a shared address instead of a name.  
     *  While {@link #open} should generally be used to open transport
     *  instances that have been either locally or remotely created, openByAddr
     *  may be used to bypass a NameServer query that would typically be  
     *  required of an {@link #open} call.
     *  
     *  Opening by address requires that the created instance was created
     *  by supplying a {@link #sharedAddr} parameter rather than a
     *  {@link #regionId} parameter.
     *
     *  A status value of Status_SUCCESS is returned if the instance
     *  is successfully opened.  Status_FAIL indicates that the instance
     *  is not yet ready to be opened.  Status_ERROR indicates that
     *  an error was raised in the error block.
     *
     *  Call {@link #close} when the opened instance is no longer needed.
     *
     *  @param(sharedAddr)  Shared address for the instance
     *  @param(handlePtr)   Pointer to handle to be opened
     *  @param(eb)          Pointer to error block
     *
     *  @a(returns)         TransportShm status
     */
    Int openByAddr(Ptr sharedAddr, Handle *handlePtr, Error.Block *eb);

    /*!
     *  ======== close ========
     *  Close an opened instance
     *
     *  Closing an instance will free local memory consumed by the opened
     *  instance.  Instances that are opened should be closed before the
     *  instance is deleted.
     *
     *  @param(handle)  handle that is returned from an {@link #openByAddr}
     */
    Void close(Handle *handle);

    /*! @_nodoc
     *  ======== sharedMemReq ========
     *  Amount of shared memory required for creation of each instance
     *
     *  Can be used to make sure the {link #sharedAddr} buffer is large 
     *  enough before calling create.
     *
     *  The {@link #sharedAddr} needs to be
     *  supplied because the cache alignment settings for the region
     *  may affect the total amount of shared memory required.
     *
     *  @param(params)      Pointer to the parameters that will be used in
     *                      the create.
     *
     *  @a(returns)         Number of MAUs needed to create the instance.
     */
    SizeT sharedMemReq(const Params *params);
    

instance:

    /*!
     *  ======== gate ========
     *  GateMP used for critical region management of the shared memory
     */
    config GateMP.Handle gate = null;
    
    /*! @_nodoc
     *  ======== openFlag ========
     *  Set to 'true' by the open() call. No one else should touch this!
     */
    config Bool openFlag = false;
    
    /*!
     *  ======== sharedAddr ========
     *  Physical address of the shared memory
     *
     *  The creator must supply the shared memory that is used to maintain
     *  shared state information.
     */
    config Ptr sharedAddr = null;
    
    /*!
     *  ======== notifyEventId ========
     *  Notify event ID for transport.
     */
    config UInt32 notifyEventId = 8;


internal:

    /*! 
     *  Constants that all delegate writers need.
     */    
    const UInt32 UP = 0xBADC0FFE;    

    /*!
     *  ======== swiFxn ========
     *  This function takes the messages from the transport ListMP and
     *  calls MessageQ_put to send them to their destination queue.
     *  This function is posted by the NotifyFxn.
     *
     *  @param(arg)     argument for the function
     */
    Void swiFxn(UArg arg);
    
    /*!
     *  ======== notifyFxn ========
     *  This is a callback function registered with Notify.  It is called
     *  when a remote processor does a Notify_sendEvent().  It is executed
     *  at ISR level.  It posts the instance Swi object to execute swiFxn.
     *
     *  @param(eventId) Notify event id
     *  @param(arg)     argument for the function
     *  @param(payload) 32-bit payload value.
     */
    Void notifyFxn(UInt16 procId, UInt16 lineId, UInt32 eventId, UArg arg, 
                   UInt32 payload);

    /* Structure of attributes in shared memory */
    struct Attrs {
        Bits32              flag;
        Bits32              creatorProcId;
        Bits32              notifyEventId;
        Bits16              priority;
        SharedRegion.SRPtr  gateMPAddr;
    };
    
    /* Instance State object */
    struct Instance_State {
        Attrs           *self;         /* Attrs in shared memory        */
        Attrs           *other;        /* Only flag field is used       */
        ListMP.Handle   localList;     /* ListMP to my processor        */
        ListMP.Handle   remoteList;    /* ListMP to remote processor    */
        Swi.Object      swiObj;        /* Each instance has a swi       */
        Int             status;        /* Current status                */
        Ipc.ObjType     objType;       /* Static/Dynamic? open/creator? */
        SizeT           allocSize;     /* Shared memory allocated       */
        Bool            cacheEnabled;  /* Whether to do cache calls     */
        UInt32          notifyEventId; /* which notify event to use     */
        UInt16          regionId;      /* the shared region id          */
        UInt16          remoteProcId;  /* dst proc id                   */
        UInt16          priority;      /* priority to register          */
        GateMP.Handle   gate;          /* Gate for critical regions     */
    };    
    
    /* Module State object */
    struct Module_State {
        ErrFxn         errFxn;
    };    
} 
/*
 *  @(#) ti.sdo.ipc.transports; 1, 0, 0, 0,109; 4-9-2010 18:38:35; /db/vtree/library/trees/ipc/ipc-d23x/src/
 */

