/* 
 * Copyright (c) 2010
 * Texas Instruments
 *
 *  All rights reserved.  Property of Texas Instruments
 *  Restricted rights to use, duplicate or disclose this code are
 *  granted through contract.
 * 
 * */
/*
 *  ======== List.c ========
 *  Implementation of functions specified in List.xdc.
 */

#include <xdc/std.h>
#include <xdc/runtime/Gate.h>

#include "package/internal/List.xdc.h"

/*
 *************************************************************************
 *                       Instance functions
 *************************************************************************
 */
    
/*
 *  ======== List_Instance_init ========
 */
Void List_Instance_init(List_Object *obj, const List_Params *params)
{
    obj->elem.next = obj->elem.prev = &(obj->elem);
}

/*
 *  ======== List_empty ========
 *  No thread-safety is needed since this API does not modify
 *  any of the contents. 
 */
Bool List_empty(List_Object *obj)
{
    return (obj->elem.next == &(obj->elem));
}

/*
 *  ======== List_get ========
 */
Ptr List_get(List_Object *obj)
{
    List_Elem *elem;
    UInt key;

    key = Gate_enterSystem();

    elem = List_dequeue(obj);
    
    Gate_leaveSystem(key);

    return (elem);
}

/*
 *  ======== List_put ========
 */
Void List_put(List_Object *obj, List_Elem *elem)
{
    UInt key;

    key = Gate_enterSystem();

    List_enqueue(obj, elem);

    Gate_leaveSystem(key);
}

/*
 *  ======== List_putHead ========
 */
Void List_putHead(List_Object *obj, List_Elem *elem)
{
    UInt key;

    key = Gate_enterSystem();

    List_enqueueHead(obj, elem);

    Gate_leaveSystem(key);
}

/*
 *  ======== List_insert ========
 */
Void List_insert(List_Object *obj, List_Elem *newElem, List_Elem *curElem)
{    
    List_enqueue((List_Object *)curElem, newElem);
}

/*
 *  ======== List_next ========
 */
Ptr List_next(List_Object *obj, List_Elem *elem)
{
    List_Elem *retElem;  /* returned elem */
    
    /* elem == NULL -> start at the head */
    if (elem == NULL) {
        retElem = obj->elem.next;   
    }
    else {
        retElem = elem->next;   
    }
    
    if (retElem == (List_Elem *)obj) {
        retElem = NULL;
    }
    
    return (retElem);
}

/*
 *  ======== List_prev ========
 */
Ptr List_prev(List_Object *obj, List_Elem *elem)
{
    List_Elem *retElem;  /* returned elem */
    
    /* elem == NULL -> start at the tail */
    if (elem == NULL) {
        retElem = obj->elem.prev;   
    }
    else {
        retElem = elem->prev;   
    }
    
    if (retElem == (List_Elem *)obj) {
        retElem = NULL;
    }
    
    return (retElem);
}
    
/*
 *  ======== List_remove ========
 */
Void List_remove(List_Object *obj, List_Elem *elem) 
{
    elem->prev->next = elem->next;
    elem->next->prev = elem->prev;
}

/*
 *  ======== List_dequeue ========
 */
Ptr List_dequeue(List_Object *obj)
{
    List_Elem *elem;

    elem = obj->elem.next;

    /* See if the List was empty */
    if (elem == (List_Elem *)obj) {
        elem = NULL;
    }
    else {
        obj->elem.next   = elem->next;
        elem->next->prev = &(obj->elem);
    }

    return (elem);
}

/*
 *  ======== List_enqueue ========
 */
Void List_enqueue(List_Object *obj, List_Elem *elem)
{
    elem->next           = &(obj->elem);
    elem->prev           = obj->elem.prev;
    obj->elem.prev->next = elem;
    obj->elem.prev       = elem;
}

/*
 *  ======== List_enqueueHead ========
 */
Void List_enqueueHead(List_Object *obj, List_Elem *elem)
{
    elem->next           = obj->elem.next;
    elem->prev           = &(obj->elem);
    obj->elem.next->prev = elem;
    obj->elem.next       = elem;    
}

/*
 *************************************************************************
 *                      Module functions
 *************************************************************************
 */

/*
 *  ======== List_elemClear ========
 */

Void List_elemClear(List_Elem *elem)
{
    elem->next = elem->prev = elem;
}

/*
 */
/*
 *  @(#) ti.sdo.utils; 1, 0, 0, 0,372; 4-9-2010 18:38:38; /db/vtree/library/trees/ipc/ipc-d23x/src/
 */

