/* 
 * Copyright (c) 2010
 * Texas Instruments
 *
 *  All rights reserved.  Property of Texas Instruments
 *  Restricted rights to use, duplicate or disclose this code are
 *  granted through contract.
 * 
 * */
/*
 *  ======== MultiProc.xdc ========
 *
 */

import xdc.rov.ViewInfo;

import xdc.runtime.Assert;

/*!
 *  ======== MultiProc ========
 *  Processor Id Module Manager
 *
 *  Many IPC modules require the ability to uniquely specify and identify
 *  processors in a multi-processor environment. The MultiProc module
 *  centeralizes processor id management into one module.  Since this 
 *  configuration is almost always universally required, most IPC applications 
 *  require supplying configuration of this module.
 *
 *  Each processor in the MultiProc module may be uniquely identified by
 *  either a name string or an integer ranging from 0 to MAXPROCESSORS - 1.
 *  Configuration is supplied using the {@link #setConfig} meta function.
 *
 *  The setConfig function tells the MultiProc module:
 *  @p(blist)
 *      - The specific processor for which the application is being built
 *      - Which processors out of a set of possible processors on a device are
 *        being used by the multi-processor application
 *  @p
 *
 *  Using the information supplied using the {@link #setConfig} meta function,
 *  The {@link #numProcessors} module configuration and the processor IDs are
 *  internally set.  Please refer to the documentation for {@link #setConfig}
 *  for more details.
 *
 *  At runtime, the {@link #getId} call returns the MultiProc id
 *  for any processor. At config-time, the {@link #getIdMeta} call returns the
 *  the same value.  At both run time and static time, the 
 *  {@link #numProcessors} module config is equal to the length of the 
 *  procNames array supplied in {@link #setConfig}.
 *
 */

module MultiProc
{
    metaonly struct ModuleView {
        UInt16       id;                /* Own ID                           */
        UInt16       numProcessors;     /* # of processors                  */
        String       nameList[];        /* Proc names ordered by procId     */
    }

    @Facet
    metaonly config ViewInfo.Instance rovViewInfo =
        ViewInfo.create({
            viewMap: [
            [
                'Module',
                {
                    type: ViewInfo.MODULE,
                    viewInitFxn: 'viewInitModule',
                    structName: 'ModuleView'
                }
            ],
            ]
        });

    /*!
     *  Assert raised when an invalid processor id is used
     */
    config Assert.Id A_invalidMultiProcId  = {
        msg: "A_invalidMultiProcId: Invalid MultiProc id"
    };

    /*!
     *  Invalid processor id constant
     *
     *  This constant denotes that the processor id is not valid.
     */
    const UInt16 INVALIDID = 0xFFFF;
    
    /*! @_nodoc
     *  ======== nameList ========
     *  Unique name for the each processor used in a multi-processor app
     *
     *  This array should never be set or read directly by the MultiProc user.
     *  The nameList is used to store names configuration supplied via the
     *  {@link #setConfig} static function.  
     *
     *  At runtime, the {@link #getName} function may be used to retrieve
     *  the name of any processor given it's MultiProc id.
     */
    config String nameList[];

    /*! @_nodoc
     *  ======== id ========
     *  Unique software id number for the processor
     *
     *  This value should never be set or read directly by the MultiProc user.
     *  Instead, the {@link #getId}, {@link #getIdMeta}, and 
     *  {@link #setLocalId} calls should be used to respectively retrieve any
     *  processors' MultiProc ids or set the local processor's MultiProc id.
     */
    config UInt16 id = 0;

    /*!
     *  ======== numProcessors ========
     *  Number of processors in the system
     *
     *  This configuration should only be read from and should never be set:
     *  numProcessors is internally set by the {@link #setConfig} meta function.
     *  setConfig statically sets the value of this configuration to the length
     *  of the supplied nameList array. After {@link #setConfig} has been 
     *  called, it is possible to retrive the maximum # of processors by 
     *  reading this module config either at run-time or at config time.
     */
    config UInt16 numProcessors = 1;

    /*!
     *  ======== getIdMeta ========
     *  Meta version of {@link #getId}
     *
     *  Statically returns the internally set ID based on configuration 
     *  supplied via {@link #setConfig}.
     */
    metaonly UInt16 getIdMeta(String name);

    /*! @_nodoc
     *  ======== getName$view ========
     *  ROV-time version of {@link #getName}
     */
    metaonly String getName$view(UInt id);

    /*! @_nodoc
     *  ======== getSlot ========
     *  Determines the offset for any two processors.
     *
     *  @param(remoteProcId)    remote MultiProc id.
     *  @b(returns)             offset for creating an instance.
     */
    UInt getSlot(UInt16 remoteProcId);

    /*!
     *  ======== setConfig ========
     *  Configure the MultiProc module
     *
     *  Configuration of the MultiProc module is primarily accomplished using
     *  the setConfig API at config time.  The setConfig API allows the
     *  MultiProc module to identify:
     *  @p(blist)
     *      - Which is the local processor
     *      - Which processors are being used
     *  @p
     *  The second of these two pieces of information is supplied via the 
     *  nameList argument.  The nameList is a non-empty set of distinct
     *  processors valid for the particular device.  For a list of valid 
     *  processor names for a given device, please refer to the :
     *  {@link ./../ipc/family/doc-files/procNames.html Table of     
     *   Valid Names for Each Device}.
     *
     *  The local processor is identified by using a single name from 
     *  nameList.  A MultiProc id is internally set to the index of
     *  'name' in the supplied 'nameList'.  If the local processor is 
     *  not known at static time, it is possible to supply a null name.
     *  MultiProc will set the local id to {@link #INVALIDID} until it
     *  is set at runtime using {@link #setLocalId}.
     *  
     */
    metaonly Void setConfig(String name, String nameList[]);

internal:

    /* id is in Module_State to support the changing of it via setLocalId */
    struct Module_State {
        UInt16 id;
    };
}

/*
 *  @(#) ti.sdo.utils; 1, 0, 0, 0,372; 4-9-2010 18:38:38; /db/vtree/library/trees/ipc/ipc-d23x/src/
 */

