/* 
 * Copyright (c) 2010
 * Texas Instruments
 *
 *  All rights reserved.  Property of Texas Instruments
 *  Restricted rights to use, duplicate or disclose this code are
 *  granted through contract.
 * 
 * */
/*
 *  ======== list.c ========
 *  This program demonstrates the following:
 *  Use Case 1: Treating a List as a FIFO Queue
 *  Use Case 2: Treating a List as a LIFO Stack
 *  Use Case 3: Traversal of a List
 *  Use Case 4: Finding and removing an element from a List
 */

#include <xdc/std.h>
#include <xdc/runtime/System.h>
#include <xdc/runtime/Gate.h>
#include <ti/sdo/utils/List.h>

/* 
 * This structure can be added to a List because the first
 * field is a List_Elem. 
 */
typedef struct Rec {
    List_Elem elem;
    Int data;
} Rec;

/*
 *  ======== main ========
 */
Int main(Int argc, Char* argv[])
{
    UInt key;
    List_Handle myList;    
    Rec r1, r2;
    Rec* rp;
    
    r1.data = 100;
    r2.data = 200;
    
    /* 
     *  No parameters are needed to create a List. Using a NULL Error_Block.
     *  This List could have been created statically via the List.create function
     *  in the List.cfg file.
     */
    myList = List_create(NULL, NULL);

    /* 
     *  =================================================================
     *  Use Case #1: Treat the List as a FIFO Queue
     *     
     *  The code prints
     *  ---------------
     *  rec: 100
     *  rec: 200
     *  =================================================================
     */    
    System_printf("Use Case #1\n");
    List_put(myList, &(r1.elem));
    List_put(myList, &(r2.elem));
 
    /* get the records and print their data */
    while ((rp = List_get(myList)) != NULL) {
        System_printf("rec: %d\n", rp->data);
    }
    
    /* 
     *  =================================================================
     *  Use Case #2: Treat the List as a LIFO Stack 
     *
     *  The code prints
     *  ---------------
     *  rec: 200
     *  rec: 100
     *  =================================================================
     */    
    System_printf("\nUse Case #2\n");
    List_putHead(myList, &(r1.elem));
    List_putHead(myList, &(r2.elem));
 
    /* get the records and print their data */
    while ((rp = List_get(myList)) != NULL) {
        System_printf("rec: %d\n", rp->data);
    }
    
    /* 
     *  =================================================================
     *  Use Case #3: Show traversal of a List
     *
     *  The code prints
     *  ---------------
     *  rec: 100
     *  rec: 200
     *  =================================================================
     */
    System_printf("\nUse Case #3\n");
    List_put(myList, &(r1.elem));
    List_put(myList, &(r2.elem));
    
    /* To start the search at the beginning of the List */
    rp = NULL;
    
    /* 
     *  Enter critical region to protect the List during traversal.
     *  Not really needed since this is in the context of main(), but
     *  shown for completeness.
     */
    key = Gate_enterSystem();
    
    while ((rp = List_next(myList, (List_Elem *)rp)) != NULL) {
        System_printf("rec: %d\n", rp->data);
    }
    
    /* Done acting on the List, leave critical region */
    Gate_leaveSystem(key);
    
    /* 
     *  =================================================================
     *  Use Case #4: Show traversal and removal of an element from a List
     *
     *  Note: Two elems are still present on List from Sample 3.
     *
     *  The code prints
     *  ---------------
     *  Found and removed element
     *  =================================================================
     */
    System_printf("\nUse Case #4\n");
    
    /* To start the search at the beginning of the List */
    rp = NULL;
    
    /* 
     *  Enter critical region to protect the List during traversal.
     *  Not really needed since this is in the context of main(), but
     *  shown for completeness.
     */
    key = Gate_enterSystem();
    
    while ((rp = List_next(myList, (List_Elem *)rp)) != NULL) {
        if (rp->data == 200) {
            List_remove(myList, (List_Elem *)rp);
            System_printf("Found and removed element\n");
        }
    }
    
    /* Done acting on the List, leave critical region */
    Gate_leaveSystem(key);

    return (0);
}

/*
 */
/*
 *  @(#) ti.sdo.utils.examples; 1, 0, 0, 0,106; 4-9-2010 18:39:57; /db/vtree/library/trees/ipc/ipc-d23x/src/
 */

